#ifndef _SYSTERMIOSH
#define _SYSTERMIOSH


typedef unsigned char cc_t;
typedef unsigned int speed_t;
typedef unsigned int tcflag_t;

#define NCC 8
struct termio {
	unsigned short c_iflag;		/* input mode flags */
	unsigned short c_oflag;		/* output mode flags */
	unsigned short c_cflag;		/* control mode flags */
	unsigned short c_lflag;		/* local mode flags */
	unsigned char c_line;		/* line discipline */
	unsigned char c_cc[NCC];	/* control characters */
};

#define NCCS 20
struct termios {
        tcflag_t        c_iflag;        /* input flags */
        tcflag_t        c_oflag;        /* output flags */
        tcflag_t        c_cflag;        /* control flags */
        tcflag_t        c_lflag;        /* local flags */
        cc_t            c_cc[NCCS];     /* control chars */
        int             c_ispeed;       /* input speed */
        int             c_ospeed;       /* output speed */
};


#define B0      0
#define B50     50
#define B75     75
#define B110    110
#define B134    134
#define B150    150
#define B200    200
#define B300    300
#define B600    600
#define B1200   1200
#define B1800   1800
#define B2400   2400
#define B4800   4800
#define B9600   9600
#define B19200  19200
#define B38400  38400



/* c_cflag */
#define CLOCAL          0x00008000      /* ignore modem status lines */
#define CS5             0x00000000      /* 5 bits (pseudo) */
#define CS6             0x00000100      /* 6 bits */
#define CS7             0x00000200      /* 7 bits */
#define CS8             0x00000300      /* 8 bits */
#define CSIZE           0x00000300      /* character size mask */
#define CSTOPB          0x00000400      /* send 2 stop bits */
#define HUPCL           0x00004000      /* hang up on last close */
#define PARENB          0x00001000      /* parity enable */
#define PARODD          0x00002000      /* odd parity, else even */

/* c_iflag bits */
#define BRKINT          0x00000002      /* map BREAK to SIGINTR */
#define ICRNL           0x00000100      /* map CR to NL (ala CRMOD) */
#define IGNBRK          0x00000001      /* ignore BREAK condition */
#define IGNCR           0x00000080      /* ignore CR */
#define IGNPAR          0x00000004      /* ignore (discard) parity errors */
#define INLCR           0x00000040      /* map NL into CR */
#define INPCK           0x00000010      /* disable checking of parity errors */
#define ISTRIP          0x00000020      /* strip 8th bit off chars */
#define IXOFF           0x00000400      /* enable input flow control */
#define IXON            0x00000200      /* enable output flow control */
#define PARMRK          0x00000008      /* mark parity and framing errors */

/* c_lflag bits */
#define ECHO            0x00000008      /* enable echoing */
#define ECHOE           0x00000002      /* visually erase chars */
#define ECHOK           0x00000004      /* echo NL after line kill */
#define ECHONL          0x00000010      /* echo NL even if ECHO is off */
#define ICANON          0x00000100      /* canonicalize input lines */
#define IEXTEN          0x00000400      /* enable FLUSHO and LNEXT */
#define ISIG            0x00000080      /* enable signals INTR, QUIT, [D]SUSP */
#define NOFLSH          0x80000000      /* don't flush after interrupt */
#define TOSTOP          0x00400000      /* stop background jobs from output */

/* c_oflag bits */
#define OPOST           0x00000001      /* enable following output processing */

/* c_cc characters */
#define VEOF            0       /* ICANON */
#define VEOL            1       /* ICANON */
#define VERASE          3       /* ICANON */
#define VINTR           8       /* ISIG */
#define VKILL           5       /* ICANON */
#define VMIN            16      /* !ICANON */
#define VQUIT           9       /* ISIG */
#define VSTART          12      /* IXON, IXOFF */
#define VSTOP           13      /* IXON, IXOFF */
#define VSUSP           10      /* ISIG */
#define VTIME           17      /* !ICANON */

/* c_cc characters for 4.3+BSD compatibility */
#define VDISCARD        15      /* IEXTEN */
#define VDSUSP          11      /* ISIG */
#define VEOL2           2       /* ICANON */
#define VLNEXT          14      /* IEXTEN */
#define VREPRINT        6       /* ICANON */
#define VSTATUS         18      /* ISIG */
#define VWERASE         4       /* ICANON */

/* tcflow() and TCXONC use these */
#define	TCOOFF		0      /* suspend output */
#define	TCOON		1      /* restart suspended output */
#define	TCIOFF		2      /* suspend input */
#define	TCION		3      /* restart suspended input */

/* tcflush() and TCFLSH use these */
#define	TCIFLUSH	0     /* flush data received but not read */
#define	TCOFLUSH	1     /* flush data written but not transmitted */
#define	TCIOFLUSH	2     /* flush both data both input and output queues */

/* tcsetattr uses these */
#define	TCSANOW		0
#define	TCSADRAIN	1
#define	TCSAFLUSH	2


#endif













