/* stdlib.c
 *
 *	(C) Copyright Apr 15 1995, Edmond J. Breen.
 *		   ALL RIGHTS RESERVED.
 * This code may be copied for personal, non-profit use only.
 *
 */


#include "stdio.h"
#include "stdlib.h"
#include "ctype.h"
#include "limits.h"

#include "string.h"
#include "math.h"
#include "float.h"
#include "unistd.h"
#include "fcntl.h"


#include "eic.h"

#include "stdliblocal.h"

#include "xalloc.h"
#include "assert.h"

#include "signal.h"

#include "sys/types.h"
#include "unistd.h"
#include "sys/stat.h"


#include "errno.h"

#include <stdarg.h>




#define getptrarg(x,y)  y = arg(x,getargs(),ptr_t)
#define checkEp(P,s)  if((char*)P.p + (s) > (char *)P.ep)\
                             Error("Access beyond array");

static void Error(char * fmt, ...)
{
    va_list args;
    va_start(args,fmt);
    vfprintf(stderr,fmt,args);
    va_end(args);
    raise(SIGSEGV);
}



#define setEp(P,sz)  (P).ep = (char*)(P).p + (sz)


int _eicUprintf(int (*output)(), void *arg, char *fmt, arg_list ap)
{
    /* copy of _UPRINTF that works with EiC's run time stack.
     * Orginally developed from James Hendrix's Small C tools.
     *
     * (c) Edmond J.Breen (March, 1995).
     *           ALL RIGHTS RESERVED.
     *
     * Purpose: To perform formatted printing to a function.
     *          It performs a superset of printf type
     *          operations.
     *
     * 		  _UPRINTF   RECOGNISES:
     * FLAGS:
     *         -, +, 0 and #.
     *    plus ! and |  (see below).
     *
     * CONVERSION CHARACTERS:
     *         d, i, o, x, X, u, c, s,
     *         f, e, E, g, G, p, n, %
     *    plus b   (see below).
     *
     * LENGTH MODIFIES:
     *          l, L, and h.
     *
     * The WIDTH and or the PRECISION can be set indirectly
     *  via * as specified by K&R.
     *
     *        _UPRINTF EXTRA FEATURES.
     * NEW FLAGS:
     *        FLAG     WHAT IT SPECIFIES
     *         |      Centre justification.
     *         !      Used with floating point numbers.
     *                   It specifies, if possible, that the
     *                   number will be centred with respect
     *                   to the decimal point.
     *                   If used with non floating point
     *                   numbers or the floating point number
     *                   does not contain a decimal point,
     *                   the ! flag is equivalent to |.
     * NEW CONVERSION CHARACTER:
     *      CHARACTER	WHAT IT SPECIFIES
     *         b      int, unsigned binary notation.
     */
  
    int v;

    static char str[128];
    char *sptr, lseen,Lseen;
  
    int  left, right, centre, maxchr,
    pad, len, width, sign, dprec,
    mod, prec,dot,gotfloat;
  
    v = 0;

    while(*fmt) {
	if(*fmt != '%') {
	    (*output)(*fmt,arg);
	    ++fmt;++v;
	    continue;
	} else if(*++fmt =='%') {
	    (*output)(*fmt++,arg);
	    ++v;
	    continue;
	}
	pad = ' ';
	centre = len = right = left = 0;
	gotfloat = dot = sign = mod = 0;
	Lseen = lseen  = dprec = 0;
	while(*fmt)  {		/* collect in any order */
	    if(*fmt == '-')  		left = 1;
	    else if(*fmt == '0')  	pad  = '0';
	    else if(*fmt == '+') 	sign = 1;
	    else if(*fmt == '#') 	mod = 1;
	    else if(*fmt == '|')        centre = 1;
	    else if(*fmt == '!')        centre = 1, dot = 1;
	    else
		break;
	    ++fmt;
	}
	if(isdigit(*fmt)) {
	    width = atoi(fmt);
	    while(isdigit(*fmt)) ++fmt;
	} else  if(*fmt == '*') {
	    width = nextarg(ap,int);
	    fmt++;
	} else
	    width = -1;
	if(*fmt == '.')  {
	    if(*++fmt == '*')  {
		maxchr = nextarg(ap,int);
		fmt++;
	    } else {
		maxchr = atoi(fmt);
		while(isdigit(*fmt)) ++fmt;
	    }
	} else
	    maxchr = -1;
    
	switch(*fmt) {		/* check for length modifier*/
	  case 'h': fmt++;break;
	  case 'l': lseen = 1; fmt++;break;
	  case 'L': Lseen = 1; fmt++;break;
	}
	sptr = str;
	switch(*fmt++) {
	  case 'c': sptr[0] = (char)nextarg(ap,int);
	    sptr[1] = 0;
	    break;
	  case 'b':
	    if(lseen) ultoa(nextarg(ap,long),sptr,2);
	    else utoa(nextarg(ap,int),sptr,2);
	    break;
	  case 'i':
	  case 'd': if(lseen) ltoa(nextarg(ap,long),sptr,10);
	  else itoa(nextarg(ap,int),sptr,10);
	  dprec=1;
	    break;
	  case 'u': if(lseen) ultoa(nextarg(ap,unsigned long),sptr,10);
	  else utoa(nextarg(ap,unsigned),sptr,10);
	  dprec=1;
	    break;
	  case 'o':	if(mod) *sptr = '0';
	    if(lseen)ltoa(nextarg(ap,long),&sptr[mod],8);
	    else itoa(nextarg(ap,int),&sptr[mod],8);
	    dprec = 1;
	    break;
	  case 's': sptr = nextarg(ap,ptr_t).p; break;
	  case 'x': case 'X':
	    if(mod) {
		*sptr = '0', sptr[1] = *(fmt-1);
		mod++;
	    }
	    if(lseen)ultoa(nextarg(ap,long),&sptr[mod],16);
	    else utoa(nextarg(ap,int),&sptr[mod],16);
	    if(*(fmt-1) == 'X') {
		while(*sptr) {
		    *sptr = toupper(*sptr);
		    sptr++;
		}
		sptr = str;
	    }
	    dprec = 1;
	    break;
	    
	  case 'e': case 'E': case 'g': case 'G':
	  case 'f': {
	      int trunc = 0;
	      char type = *(fmt-1),c;
	      double d;
      
	      gotfloat = 1;
	      if(maxchr < 0) prec = 6;
	      else prec = maxchr,maxchr = -1;

	      if(Lseen)
		  d = nextarg(ap,double);
	      else
		 d = nextarg(ap,double);
      
	      if(type == 'g' || type == 'G') {
		  double ex;
		  if(d !=0)
		      ex = log10(d < 0? -d:d);
		  else
		      ex = 1;
		  if(ex < -4 || ex >= prec)
		      c = type - 2;
		  else
		      c = 'f';
		  trunc = 1;
	      } else
		  c = type;
	      if(mod) {
		  if(!prec) prec = 1;
		  trunc = 0;
	      }
	      sptr = fftoa(d,str, prec, c,trunc);
	  }
	    break;
	  case 'n': *(int*)nextarg(ap,ptr_t).p = v; continue;
	  case 'p':
	      ultoa((long)nextarg(ap,ptr_t).p,sptr,10);
	    break;
	  default: (*output)(*(fmt-1),arg);
	    while(*fmt) {
		(*output)(*fmt++,arg);
		++v;
	    }
	    return v;
	}
    
	if(!len) len = strlen(sptr);
	if(sign) len += (*sptr == '-' ? 0: 1);

	if(dprec && maxchr >=0) {
	    dprec = maxchr;
	    pad = '0';
	    
	}else
	    dprec = 0;
	    
	
	if(maxchr > -1 && maxchr<len)
	    len = maxchr;
	if(width>len)
	    width = width -len;
	else if(dprec > len)
	    width = dprec - len;
	else
	    width = 0;
    
	if(centre) {
	    left = (right = width >> 1) + (width%2);
	    if(dot && gotfloat) {
		int d = 0,c;
		while(sptr[d] && sptr[d] != '.')
		    d++;
		if(sptr[d] == '.') {
		    c = (width + len)/2;
		    if(sign && sptr[0] != '-')
			d++;
		    if(c-d > 0) right = c-d;
		    else right = 0;
		    if(width - right > 0)
			left = width - right;
		    else
			left = 0;
		}
	    }
	} else 
	    if(!left)
		right = width;
	    else
		left = width;
    
	if(sign && !left && pad == '0') {
	    if(*sptr != '-') {
		(*output)('+',arg);
		++v;len--;
	    }
	    sign = 0;
	}
	while(right--) {(*output)(pad,arg); ++v;}
	if(sign && *sptr != '-') {(*output)('+',arg);len--;++v;}
	while(len--) {(*output)(*sptr++,arg);++v;}
	while(left--) {(*output)(pad,arg);++v;}
    }
    return v;
}



int _eicUscanf(int (*input)(), int (*uget)(),
	    void *arg, const char *fmt, arg_list ap)
{
    /*_USCANF
     *
     * (C) May 10 1995 Edmond J.Breen.
     *           ALL RIGHTS RESERVED.
     *
     * Purpose: To perform formatted reading from a function.
     *          _Uscanf performs a superset of scanf type
     *          operations.
     *
     * 		  _USCANF   RECOGNISES:
     *    
     *     %[*] [width] [h|l|L] [b|d|i|o|x|X|u|c|s|f|e|E|g|G|p|n|%] 
     *     %[*] [width] [scan-set]
     *
     * CONVERSION CHARACTERS:
     *         d, i, o, x, X, u, c, s
     *         f, e, E, g, G, p, n, %
     *    plus b   (see below).
     *
     * LENGTH MODIFIES:
     *          l, L, and h.
     *
     *     *    optional assignment-suppression character
     * WIDTH    optional integer width of field specifier
     *
     * SCAN-SET
     *      [...]   matches the longest non-emtpy string
     *              of input from the set between the brackects.
     *      [^...]  matches the longest non-emtpy string
     *              of input NOT from the set between the brackects.
     *      []...] and [^]...]
     *                  includes ] as part of the scan-set.
     *
     * NEW CONVERSION CHARACTER:
     *      CHARACTER	WHAT IT SPECIFIES
     *         b      int, unsigned binary notation.
     *
     * STANDARD C AUXILIARY FUNCTIONS REQUIRED 
     *    strtol, strtod, strtoul
     *
     * LIMITATIONS:
     *       (1) long double handled as double
     *       (2) error checking could be improved
     */

    static char field[256], *p;
    
    char *carg, sbuf;
    int type;
    int  wast, width,cwidth, ch,  base,cnt;
    void *vptr;
    int v;

    
    cnt = v = 0;
    
    while(*fmt != '\0') {
	if(isspace(*fmt)) {  /* match white space */
	    while(*fmt && isspace(*fmt)) ++fmt;
	    if(*fmt) {
		while(isspace((ch=(*input)(arg))))
		    ++cnt;
		(*uget)(ch,arg);
	    }
	    continue;
	}
	if(*fmt != '%') { /* match literal text */
	    while(*fmt && !isspace(*fmt) && *fmt != '%')
		if((ch=(*input)(arg)) == *fmt) 
		    ++fmt,++cnt;
		else
		    return v;
	    continue;
	}
	    
	if(*++fmt == '*') {
	    vptr = &sbuf;
	    wast = 1;
	    ++fmt;
	} else if(*fmt == '%') {
	    cnt++;
	    fmt++;
	    continue;
	} else  {
	    wast = 0;
	    vptr = nextarg(ap,ptr_t).p;
	}

	for(width = 0;isdigit(*fmt);fmt++)
	    width = *fmt - '0' + width * 10;
	if(!width)
	    width = INT_MAX, cwidth = 1;
	else
	    cwidth = width;

	if(*fmt != 'c' && *fmt != '[' && *fmt != 'n') {
	    /* strip leading white space */
	    while(isspace(ch = (*input)(arg)))
		++cnt;
	    if(ch == EOF) {
		if(v)
		    break;
		else
		    return EOF;
	    }
	    (*uget)(ch,arg);
	}

	
	switch(*fmt) {
	  case 'h': type = 'h'; ++fmt; break;
	  case 'l': type = 'l'; ++fmt; break;
	  case 'L': type = 'L'; ++fmt; break;
	  default:  type = 0;
	}
	    
	switch(*fmt) {
	  case 'c': carg = vptr;
	    while(cwidth--) {
		*carg = (*input)(arg);
		++cnt;
		if(carg != &sbuf)
		    ++carg;
	    }
	    break;
	  case 's':
	    carg = vptr;
	    while(width--) {
		if((*carg = (*input)(arg)) == EOF)
		    break;
		++cnt;
		if(isspace(*carg)) {
		    (*uget)(*carg,arg);
		    --cnt;
		    break;
		} if(carg != &sbuf)
		    ++carg;
	    }
	    *carg = '\0';
	    break;
	  case '[':
	    ++fmt;
	    
	    if(*fmt == '^') {++fmt; type = 0;}
	    else type = 1;
	    
	    p = (char*)fmt;
	    if(*p == ']') ++p;
	    
	    while(*p && *p != ']') ++p;
	    cwidth = p - fmt;
	    if(cwidth == 0) return EOF;
	    carg = vptr;
	    while((ch = (*input)(arg)) != EOF) 
		if(type) {
		    if(memchr(fmt,ch,cwidth)) {
			if(!wast)
			    *carg++ = ch;
			++cnt;
		    }else {
			(*uget)(ch,arg);
			break;
		    }
		} else 
		    if(!memchr(fmt,ch,cwidth)) {
			if(!wast)
			    *carg++ = ch;
			++cnt;
		    }else {
			(*uget)(ch,arg);
			break;
		    }
	    *carg = '\0';
	    fmt += cwidth;
	    break;	    
	  case 'e': case 'f': case 'g':
	  case 'E': case 'G':
	    p = field;
	    ch = (*input)(arg);
	    if(ch == '-' || ch == '+') {
		++cnt;
		*p++ = ch;
	    } else
		(*uget)(ch,arg);
	    while(width-- && isdigit(ch=(*input)(arg))) {
		++cnt; *p++ = ch;
	    }
	    if(ch == '.' && width-- > 0) {
		*p++ = '.';
		++cnt;
		while(isdigit(ch=(*input)(arg)) && width--) {
		    ++cnt;
		    *p++ = ch;
		}
	    }
	    if((ch == 'e' || ch == 'E') && width--) {
		*p++ = ch;
		ch = (*input)(arg);
		if(ch != '+' && ch != '-' && !isdigit(ch) &&
		   ch != EOF && !isspace(ch)) {
		    (*uget)(ch,arg);
		    ch = *--p;
		} else
		    cnt++;
		    
	    }
	    if((ch == '+' || ch == '-') && width--) {
		*p++ = ch;
		++cnt;
		ch = (*input)(arg);
	    }		
	    while(isdigit(ch) && width--) {
		*p++ = ch;
		++cnt;
		ch = (*input)(arg);
	    }
	    (*uget)(ch,arg);
	    if(p == field)
		return v;
	    *p = '\0';
	    if(!wast){
		double d = strtod(field,NULL);
		if(!type || type == 'h')
		    *(float*)vptr = (float)d;
		else if(type == 'l' || type == 'L')
		    *(double*)vptr = d;
	    }
	    break;
	  case 'n':
	    if(type == 'l')
		*(long*)vptr = cnt;
	    else if(type == 'h')
		*(short*)vptr = cnt;
	    else
		*(int*)vptr = cnt;
	    v--;
	    break;	    
	  default:
	    switch(*fmt) {
	    case 'b': base = 2; break;
	    case 'i':
	    case 'd':
	    case 'u':
		ch = (*input)(arg);
		if(ch == '0') base = 8;
		else  base = 10;

		(*uget)(ch,arg);
		break;
	    case 'o': base = 8;  break;
	    case 'x':
	    case 'X': base = 16; break;
	    case 'p': base = 0; type = 'l'; break;
	    default:
		return v;
	    }
	    p = field;
	    while(width-- && !isspace(ch=(*input)(arg)) && ch != EOF) {
		*p++ = ch;
		++cnt;
	    }
	    if(width > 0)
		(*uget)(ch,arg);	
	    if(wast)
		break;
	    if(p == field)
		return v;
	    *p = '\0';
	{
	    char *endptr;
	    if(*fmt == 'd' || *fmt == 'i') { /* signed conversion */
		long lval = strtol(field,&endptr,base);
		if(type) {
		    if(type == 'h')
			*(short *)vptr =(short) lval;
		    else 
			*(long *)vptr = lval;
		} else
		    *(int *)vptr = (int) lval;
			
	    } else {
		unsigned long ulval = strtoul(field,&endptr,base);
		if(type) {
		    if(type == 'h')
			*(unsigned short *)vptr =(unsigned short) ulval;
		    else
			*(unsigned long *)vptr = ulval;
		} else
		    *(unsigned *)vptr = (unsigned)ulval;
	    }
	    while(--p >= endptr)
		(*uget)(*p,arg);
	}
	    break;
	}
	++fmt;
	++v;
    }
    return v;
}


val_t eic_popen()
{
    val_t v;
    arg_list ap = getargs();

    v.p.sp = v.p.p = popen(arg(0,ap,ptr_t).p,arg(1,ap,ptr_t).p);
    setEp(v.p,sizeof(FILE) -1);
    
    return  v;
}
val_t eic_pclose()
{
    val_t v;
    arg_list ap = getargs();

    v.ival = pclose(arg(0,ap,ptr_t).p);
    return  v;
}

/*
   STRING.H STUFF
   ------------------------------*/

						      
val_t eic_memcpy(void)
{
    val_t v;
    ptr_t  p2;
    size_t s;

    getptrarg(0,v.p);
    getptrarg(1,p2);

    s = arg(2,getargs(),size_t);

    checkEp(v.p,s-1);
    checkEp(p2,s-1);
    
    memcpy(v.p.p,p2.p,s);
    
    return v;
}
val_t eic_memmove(void)
{
    val_t v;
    ptr_t  p2;
    size_t s;

    getptrarg(0,v.p);
    getptrarg(1,p2);

    s = arg(2,getargs(),size_t);


    checkEp(v.p,s-1);
    checkEp(p2,s-1);
    
    
#ifdef _SUNOS
    bcopy(p2.p,v.p.p,s);
#else
    memmove(v.p.p,p2.p,s);
#endif
    return v;
}
val_t eic_strcpy(void)     /* unsafe */
{
    val_t v;

    getptrarg(0,v.p);
    checkEp(v.p,strlen(arg(1,getargs(), ptr_t).p) - 1);

    strcpy(v.p.p,
	   arg(1,getargs(), ptr_t).p);    
    return v;
}
val_t eic_strncpy(void)
{
    val_t v;
    ptr_t  p2;
    size_t s;

    getptrarg(0,v.p);
    getptrarg(1,p2);

    s = arg(2,getargs(),size_t);

    
    checkEp(v.p,s-1);
    
    strncpy(v.p.p,p2.p,s);


    return v;
}
val_t eic_strcat(void)  /* unsafe */
{
    val_t v;
    ptr_t  p2;

    getptrarg(0,v.p);
    getptrarg(1,p2);

    checkEp(v.p,strlen(v.p.p)+strlen(p2.p));
    
    strcat(v.p.p,p2.p);
    return v;
}

val_t eic_strncat(void) /* unsafe */
{
    val_t v;
    ptr_t  p2;
    size_t s, s1;
    
    getptrarg(0,v.p);
    getptrarg(1,p2);

    s = arg(2,getargs(),size_t);
    s1 = strlen(p2.p);

    checkEp(v.p,strlen(v.p.p) +
	    (s < s1 ? s : s1));
    
    strncat(v.p.p,p2.p,s);

    return v;
}
val_t eic_memcmp(void)
{
    val_t v;
    v.ival = memcmp(arg(0,getargs(),ptr_t).p,
		 arg(1,getargs(),ptr_t).p,
		 arg(2,getargs(),size_t));
    return v;
}
val_t eic_strcmp(void)
{
    val_t v;
    v.ival  = strcmp(arg(0,getargs(),ptr_t).p,
		  arg(1,getargs(),ptr_t).p);
    return v;
}
val_t eic_strcoll(void)
{
    val_t v;
    v.ival = strcoll(arg(0,getargs(),ptr_t).p,
		  arg(1,getargs(),ptr_t).p);
    return v;
}
val_t eic_strncmp(void)
{
    val_t v;
    v.ival  = strncmp(arg(0,getargs(),ptr_t).p,
		   arg(1,getargs(),ptr_t).p,
		   arg(2,getargs(),size_t));
    return v;
}
val_t eic_strxfrm(void)
{
    val_t v;
    v.szval = strxfrm(arg(0,getargs(),ptr_t).p,
		     arg(1,getargs(),ptr_t).p,
		     arg(2,getargs(),size_t));
    return v;
}
val_t eic_memchr(void)
{
    val_t v;

    v.p = arg(0,getargs(),ptr_t);
    
    v.p.p = memchr(v.p.p,
		   arg(1,getargs(),int),
		   arg(2,getargs(),size_t));
    return v;
}
val_t eic_strchr(void)
{
    val_t v;
    v.p = arg(0,getargs(),ptr_t);
    v.p.p = strchr(v.p.p,
	   arg(1,getargs(),int));
    return v;
}
val_t eic_strcspn(void)
{
    val_t v;
    v.szval = strcspn(arg(0,getargs(),ptr_t).p,
		     arg(1,getargs(),ptr_t).p);
    return v;
}
val_t eic_strpbrk(void)
{
    val_t v;

    getptrarg(0,v.p);
    v.p.p = strpbrk(v.p.p,
		    arg(1,getargs(),ptr_t).p);
    return v;
}
val_t eic_strrchr(void)
{
    val_t v;
    getptrarg(0,v.p);
    v.p.p = strrchr(v.p.p,
		    arg(1,getargs(),int));
    return v;
}
val_t eic_strspn(void)
{
    val_t v;
    v.szval = strspn(arg(0,getargs(),ptr_t).p,
		    arg(1,getargs(),ptr_t).p);
    return v;
}
val_t eic_strstr(void)
{
    val_t v;
    getptrarg(0,v.p);
    
    v.p.p = strstr(v.p.p,
		   arg(1,getargs(),ptr_t).p);
    return v;
}
val_t eic_strtok(void)
{
    val_t v;

    getptrarg(0,v.p);
    
    v.p.p = strtok(v.p.p,
		   arg(1,getargs(),ptr_t).p);
    return v;
}
val_t eic_memset(void)
{
    val_t v;
    
    getptrarg(0,v.p);
    memset(v.p.p,
	   arg(1,getargs(),int),
	   arg(2,getargs(),size_t));
    return v;
}
val_t eic_strerror(void)
{
    val_t v;
    
    v.p.p = strerror(arg(0,getargs(),int));

    v.p.sp = v.p.p;
    
    setEp(v.p,strlen(v.p.p));
    
    return v;
}
val_t eic_strlen(void)
{
    val_t v;
    v.szval = strlen(arg(0,getargs(),ptr_t).p);
    return v;
}


val_t eic_strrev(void)
{
    val_t v;
    getptrarg(0,v.p);
    v.p.p = strrev(v.p.p);
    return  v;
}

val_t eic_fftoa(void)
{
    val_t v;
    arg_list ap = getargs();

    getptrarg(1,v.p);
    
    fftoa(arg(0,ap,float),   /* value */
	  v.p.p,             /* string */
	  arg(2,ap,int),     /* precision */
	  arg(3,ap,char),    /* 'f','e' or 'E' */
	  arg(4,ap,int));    /* truncation, i.e. 0 or 1 */
    return v;
}


/*
   STDLIB.H  STUFF
   ----------------------------------*/

val_t eic_malloc(void)
{
    val_t v;

    v.p.sp = v.p.p = xmalloc(arg(0,getargs(),size_t));
    setEp(v.p,arg(0,getargs(),size_t) -1);
    
    return v;
}

val_t eic_calloc(void)
{
    val_t v;
    arg_list ap = getargs();
    size_t s1,s2;
    
    s1 = arg(0,ap,size_t);
    s2 = arg(1,ap,size_t);
    
    v.p.sp = v.p.p = xcalloc(s1,s2);
    
    setEp(v.p,s1 * s2 - 1);  
    return v;
}

val_t eic_realloc(void)
{
    val_t v;
    arg_list ap = getargs();

    v.p.sp = v.p.p = xrealloc(arg(0,ap,ptr_t).p,
			      arg(1,ap,size_t));
    setEp(v.p,arg(1,ap,size_t) - 1);
    return v;
}

val_t eic_free(void)
{
    val_t v;
    ptr_t *p;
    
    
    p = arg(0,getargs(),ptr_t*);
    
    xfree(p->p);
    p->p = p->sp = p->ep = NULL; 
        
    return v;
}


val_t eic_strtod(void)
{
    val_t v;
    arg_list ap = getargs();
    ptr_t p1, *p2;
    void * endptr;

    p1 = arg(0,ap,ptr_t);
    
    if(arg(1,ap,ptr_t).p) {
	p2 = arg(1,ap,ptr_t).p;
	p2->sp = p1.sp;
	p2->ep = p1.ep;
	endptr= &p2->p;
    } else
        endptr = NULL;

    v.dval = (float)strtod(p1.p,
			   endptr);
    return v;
}

val_t eic_strtol(void)
{
    val_t v;
    arg_list ap = getargs();
    ptr_t p1, *p2;
    void * endptr;

    p1 = arg(0,ap,ptr_t);
    
    if(arg(1,ap,ptr_t).p) {
	p2 = arg(1,ap,ptr_t).p;
	p2->sp = p1.sp;
	p2->ep = p1.ep;
	endptr= &p2->p;
    } else
        endptr = NULL;
    
    
    v.lval = strtol(p1.p,
		    endptr,
		    arg(2,ap,int));
    return v;
}

val_t eic_strtoul(void)
{
    val_t v;
    arg_list ap = getargs();
    ptr_t p1, *p2;
    void * endptr;

    p1 = arg(0,ap,ptr_t);
    
    if(arg(1,ap,ptr_t).p) {
	p2 = arg(1,ap,ptr_t).p;
	p2->sp = p1.sp;
	p2->ep = p1.ep;
	endptr= &p2->p;
    } else
        endptr = NULL;
    
  
    v.ulval = strtoul(p1.p,
		      endptr,
		      arg(2,ap,int));
    return v;
}

val_t eic_system(void)
{
    val_t v;
    v.ival = system(nextarg(getargs(),ptr_t).p);
    return v;
}



val_t eic_itoa(void)
{
    arg_list ap = getargs();
    val_t v;

    getptrarg(1,v.p);
    
    switch(arg(3,ap,int)) {
      case 1:
	v.p.sp = v.p.p = itoa(arg(0,ap,int),
			      v.p.p,
			      arg(2,ap,int));
	break;
      case 2:
	v.p.sp = v.p.p = utoa(arg(0,ap,unsigned),
		      v.p.p,
		      arg(2,ap,int));
    }
    setEp(v.p,strlen(v.p.p));
    
    return v;
}    

val_t eic_ltoa(void)
{
    arg_list ap = getargs();
    val_t v;

    getptrarg(1,v.p);

    switch(arg(3,ap,int)) {
      case 1:
	v.p.sp = v.p.p = ltoa(arg(0,ap,long),
		      v.p.p,
		      arg(2,ap,int));
	break;
      case 2:
	v.p.sp = v.p.p = ultoa(arg(0,ap,unsigned long),
		       v.p.p,
		       arg(2,ap,int));
	break;
    }

    setEp(v.p,strlen(v.p.p));

    
    return  v;
}

static unsigned long _Rseed = 1;
int rand(void)
{
    _Rseed = _Rseed * 1103515245 + 12345;
    return (unsigned)(_Rseed >> 16) & RAND_MAX;
}
void srand(unsigned int x)
{
    _Rseed = x;
}

val_t eic_rand(void)
{
    val_t v;
    v.ival = rand();
    return v;
}
    
val_t eic_srand(void)
{
    val_t v;
    srand(nextarg(getargs(),unsigned int));
    return v;
}

val_t eic_atoi(void)
{
    val_t v;
    v.ival = atoi(nextarg(getargs(),ptr_t).p);
    return v;
}

val_t eic_atol(void)
{
    val_t v;
    v.lval = atol(nextarg(getargs(),ptr_t).p);
    return v;
}

val_t eic_atof(void)
{
    val_t v;
    v.dval = atof(nextarg(getargs(),ptr_t).p);
    return v;
}

val_t eic_getenv(void)
{
    val_t v;
    v.p.sp = v.p.p = getenv(nextarg(getargs(),ptr_t).p);
    if(v.p.p)
	setEp(v.p,strlen(v.p.p));
    else
	v.p.ep = v.p.p;
    return v;
}

val_t eic_exit(void)
{
    extern int interActive; /* defined in starteic.c */    
    val_t v;

    if(!interActive) {
	v.ival = arg(0,getargs(),int);
	exit(v.ival);
    }
    raise(SIGUSR1);
    return v;
}




/* FCNTL.H STUFF */

val_t eic_fcntl(void)
{
    val_t v;

    if(getargc() == 3) {
	v.ival = fcntl(arg(0,getargs(),int),
		      arg(1,getargs(),int),
		      arg(2,getargs(),int));
    } else
	v.ival = fcntl(arg(0,getargs(),int),
		      arg(1,getargs(),int));
    return v;
}

val_t eic_open(void)
{
    val_t v;
    if(getargc() == 3) {
	v.ival = open(arg(0,getargs(),ptr_t).p,
		      arg(1,getargs(),int),
		      arg(2,getargs(),mode_t));
    } else
	v.ival = open(arg(0,getargs(),ptr_t).p,
		      arg(1,getargs(),int));
    return v;
}

val_t eic_creat(void)
{
    val_t v;
    arg_list ap = getargs();
    v.ival = creat(arg(0,ap,ptr_t).p,
		   arg(1,ap,int));
    return v;
}


/* CTYPE.C STUFF */

val_t _get_ctype(void)
{
    val_t v;
    v.p.p = _CtYpE;
    v.p.sp = (char *)v.p.p - sizeof(unsigned short);
    setEp(v.p,255*sizeof(unsigned short));  /* see ctype.c */
    return v;
}




/* STDIO.C STUFF */
val_t _get_IoB(void)
{
    val_t v;
    /*
     * This function exists so that EiC can get the address 
     * _IoB.
     */
    v.p.sp = v.p.p = _IoB;
    setEp(v.p,sizeof(_IoB) - 1);
    return v;
}


val_t eic_ftell(void)
{
    val_t v;
    v.lval = ftell(arg(0,getargs(),ptr_t).p);
    return v;
}

val_t eic_fseek(void)
{
    val_t v;
    v.ival = fseek(arg(0,getargs(),ptr_t).p,
		   arg(1,getargs(),long),
		   arg(2,getargs(),int));
    return v;
}


val_t eic_printf(void)
{
    val_t v;
    arg_list ap = getargs();
    char *fmt;
    
    fmt = nextarg(ap,ptr_t).p;
    v.ival = _eicUprintf(fputc,stdout,fmt,ap);
    return v;
}

val_t eic_fprintf(void)
{
    val_t v;
    arg_list ap = getargs();
    FILE *fp;
    char *fmt;
    
    fp = nextarg(ap,ptr_t).p;
    fmt = nextarg(ap,ptr_t).p;
    v.ival = _eicUprintf(fputc,fp,fmt,ap);
    return v;
}

val_t eic_vfprintf(void)
{
    val_t v;
    arg_list ap = getargs();
    arg_list ags;
    FILE *fp;
    char *fmt;
    
    fp = nextarg(ap,ptr_t).p;
    fmt = nextarg(ap,ptr_t).p;
    ags = nextarg(ap,arg_list);
    v.ival = _eicUprintf(fputc,fp,fmt,ags);
    return v;
}




val_t eic_sprintf(void)
{
    val_t v;
    arg_list ap = getargs();
    char *fmt, *str; 
    str = nextarg(ap,ptr_t).p;
    fmt = nextarg(ap,ptr_t).p;    
    v.ival = _eicUprintf(charout_,&str,fmt,ap);
    return v;
}


val_t eic_vsprintf(void)
{
    val_t v;
    arg_list ap = getargs();
    arg_list ags;
    char * str;
    char *fmt;
    
    str = nextarg(ap,ptr_t).p;
    fmt = nextarg(ap,ptr_t).p;
    ags = nextarg(ap,arg_list);
    v.ival = _eicUprintf(charout_,&str,fmt,ags);
    return v;
}


val_t eic_freopen(void)
{
    val_t v;
    arg_list ap = getargs();
    v.p.sp = v.p.p = freopen(arg(0,ap,ptr_t).p,
			     arg(1,ap,ptr_t).p,
			     arg(2,ap,ptr_t).p);
    
    setEp(v.p,sizeof(FILE) - 1);
    return v;
}

val_t eic_fopen(void)
{
    val_t v;
    arg_list ap = getargs();
    v.p.sp = v.p.p = fopen(arg(0,ap,ptr_t).p,
			   arg(1,ap,ptr_t).p);
    
    setEp(v.p,sizeof(FILE) - 1);
    return v;
}





val_t eic_ungetc(void)
{
    val_t v;
    arg_list ap = getargs();
    v.ival = ungetc(arg(0,ap,int),
		    arg(1,ap,ptr_t).p);
    return v;
}
    
val_t eic_fgetc(void)
{
    val_t v;
    v.ival = fgetc(nextarg(getargs(),ptr_t).p);
    return v;
}
val_t eic_fclose(void)
{
    val_t v;
    v.ival = fclose(nextarg(getargs(),ptr_t).p);
    return v;
}
val_t eic_fflush(void)
{
    val_t v;
    v.ival = fflush(nextarg(getargs(),ptr_t).p);
    return v;
}

val_t eic_fputc(void)
{
    val_t v;
    arg_list ap = getargs();
    v.ival = fputc(arg(0,ap,int),
		   arg(1,ap,ptr_t).p);
    return v;
}

val_t eic_fputs(void)
{
    val_t v;
    arg_list ap = getargs();
    v.ival = fputs(arg(0,ap,ptr_t).p,
		   arg(1,ap,ptr_t).p);
    return v;
}

val_t eic_fgets(void)
{
    val_t v;
    arg_list ap = getargs();
    getptrarg(0,v.p);
    v.p.p = fgets(v.p.p,
		  arg(1,ap,int),
		  arg(2,ap,ptr_t).p);
    return v;
}

val_t eic_gets(void)
{
    val_t v;
    getptrarg(0,v.p);
    v.p.p = gets(v.p.p);
    return v;
}

val_t eic_fread(void)
{
    val_t v;
    arg_list ap = getargs();
    v.szval = fread(arg(0,ap,ptr_t).p,
		   arg(1,ap,size_t),
		   arg(2,ap,size_t),
		   arg(3,ap,ptr_t).p);
    return v;
}

val_t eic_fwrite(void)
{
    val_t v;
    arg_list ap;
    ap = getargs();
    v.szval = fwrite(arg(0,ap,ptr_t).p,
		   arg(1,ap,size_t),
		   arg(2,ap,size_t),
		   arg(3,ap,ptr_t).p);
    return v;
}

val_t eic_fscanf(void)
{
    val_t v;
    arg_list ap = getargs();
    v.ival = _eicUscanf(fgetc,ungetc,
		     arg(0,ap,ptr_t).p,
		     arg(1,ap,ptr_t).p,ap-2);
    return v;
}

val_t eic_scanf(void)
{
    val_t v;
    arg_list ap = getargs();
    v.ival = _eicUscanf(fgetc,ungetc,
		     stdin,
		     arg(0,ap,ptr_t).p,ap-1);
    return v;
}

val_t eic_sscanf(void)
{
    val_t v;
    char * str;
    arg_list ap = getargs();
    str = arg(0,ap,ptr_t).p;
    v.ival = _eicUscanf(charin_,charback_,
		     &str,
		     arg(1,ap,ptr_t).p,ap-2);
    return v;
}

val_t eic_setvbuf(void)
{
    val_t v;
    arg_list ap = getargs();
    v.ival = setvbuf(arg(0,ap,ptr_t).p,
		     arg(1,ap,ptr_t).p,
		     arg(2,ap,int),
		     arg(3,ap,size_t));
    return v;
}
val_t eic_tmpnam(void)
{
    val_t v;
    v.p.sp = v.p.p = tmpnam(arg(0,getargs(),ptr_t).p);
    setEp(v.p,strlen(v.p.p));
    return v;
}
val_t eic_tmpfile(void)
{
    val_t v;
    v.p.sp = v.p.p = tmpfile();
    setEp(v.p,strlen(v.p.p));
    return v;
}

val_t eic_rename(void)
{
    val_t v;
    v.ival = link(arg(0,getargs(),ptr_t).p,
		  arg(1,getargs(),ptr_t).p);
    if(v.ival == 0)
	v.ival =  unlink(arg(0,getargs(),ptr_t).p);
    return v;
}


val_t eic_clearerr(void)
{
    val_t v;
    clearerr(arg(0,getargs(),ptr_t).p);
    return v;
}

val_t eic_perror(void)
{
    val_t v;
    perror(arg(0,getargs(),ptr_t).p);
    return v;
}


/* MATH.H STUFF */

val_t eic_div(void)
{
    /*  rem: the return type is a structure;
     *  therefore, the first slot is used for
     *  a hidden parameter
     */

    val_t v;
    static div_t d; /* create some memory */

    /* skip argument -1 & start at -2 */

    int num = arg(1,getargs(),int);
    int dem = arg(2,getargs(),int);

    d.quot = num/dem;
    d.rem = num - dem * d.quot;
    if(d.quot < 0 && 0 < d.rem) {
	d.quot += 1;
	d.rem -= dem;
    }

    /* set safe */
    v.p.sp = v.p.p = &d;
    v.p.ep = (char*)&d + sizeof(div_t);
    
    return v;
}

val_t eic_ldiv(void)
{
    val_t v;
    static ldiv_t d; /* create some memory */

    /* skip argument -1 & start at -2 */

    long num = arg(1,getargs(),long);
    long dem = arg(2,getargs(),long);

    d.quot = num/dem;
    d.rem = num - dem * d.quot;
    if(d.quot < 0 && 0 < d.rem) {
	d.quot += 1;
	d.rem -= dem;
    }

    /* set safe */
    v.p.sp = v.p.p = &d;
    v.p.ep = (char*)&d + sizeof(ldiv_t);

    return v;
}

val_t eic_acos(void)
{
    val_t v;
    v.dval = acos(arg(0,getargs(),double));
    return v;
}

val_t eic_asin(void)
{
    val_t v;
    v.dval = asin(arg(0,getargs(),double));
    return v;
}
val_t eic_atan(void)
{
    val_t v;
    v.dval = atan(arg(0,getargs(),double));
    return v;
}
val_t eic_atan2(void)
{
    val_t v;
    v.dval = atan2(arg(0,getargs(),double), arg(1,getargs(),double));
    return v;
}
val_t eic_cos(void)
{
    val_t v;
    v.dval = cos(arg(0,getargs(),double));
    return v;
}
val_t eic_sin(void)
{
    val_t v;
    v.dval = sin(arg(0,getargs(),double));
    return v;
}
val_t eic_tan(void)
{
    val_t v;
    v.dval = tan(arg(0,getargs(),double));
    return v;
}
val_t eic_cosh(void)
{
    val_t v;
    v.dval = cosh(arg(0,getargs(),double));
    return v;
}
val_t eic_sinh(void)
{
    val_t v;
    v.dval = sinh(arg(0,getargs(),double));
    return v;
}
val_t eic_tanh(void)
{
    val_t v;
    v.dval = tanh(arg(0,getargs(),double));
    return v;
}
val_t eic_exp(void)
{
    val_t v;
    v.dval = exp(arg(0,getargs(),double));
    return v;
}
val_t eic_frexp(void)
{
    val_t v;
    v.dval = frexp(arg(0,getargs(),double), arg(1,getargs(),ptr_t).p);
    return v;
}
val_t eic_ldexp(void)
{
    val_t v;
    v.dval = ldexp(arg(0,getargs(),double), arg(1,getargs(),int));
    return v;
}
val_t eic_log(void)
{
    val_t v;
    v.dval = log(arg(0,getargs(),double));
    return v;
}
val_t eic_log10(void)
{
    val_t v;
    v.dval = log10(arg(0,getargs(),double));
    return v;
}
val_t eic_modf(void)
{
    val_t v;
    v.dval = modf(arg(0,getargs(),double), arg(1,getargs(),ptr_t).p);
    return v;
}
val_t eic_pow(void)
{
    val_t v;
    v.dval = pow(arg(0,getargs(),double), arg(1,getargs(),double));
    return v;
}
val_t eic_sqrt(void)
{
    val_t v;
    v.dval = sqrt(arg(0,getargs(),double));
    return v;
}
val_t eic_ceil(void)
{
    val_t v;
    v.dval = ceil(arg(0,getargs(),double));
    return v;
}
val_t eic_fabs(void)
{
    val_t v;
    v.dval = fabs(arg(0,getargs(),double));
    return v;
}
val_t eic_floor(void)
{
    val_t v;
    v.dval = floor(arg(0,getargs(),double));
    return v;
}
val_t eic_fmod(void)
{
    val_t v;
    v.dval = fmod(arg(0,getargs(),double), arg(1,getargs(),double));
    return v;
}


/* DIRENT.H STUFF */

#include "dirent.h"

off_t telldir(DIR *dir);
void seekdir(DIR *dir, off_t offset);

val_t eic_opendir(void)
{
    val_t v;
    v.p.sp = v.p.p = opendir(arg(0,getargs(),ptr_t).p);
    setEp(v.p,sizeof(DIR) - 1);
    return v;
}

val_t eic_readdir(void)
{
    val_t v;
    v.p.sp = v.p.p = readdir(arg(0,getargs(),ptr_t).p);
    setEp(v.p,sizeof(struct dirent) - 1);
    return v;
}

val_t eic_telldir(void)
{
    val_t v;
    v.lval = telldir(arg(0,getargs(),ptr_t).p);
    return v;
}

val_t eic_seekdir(void)
{
    val_t v;
    seekdir(arg(0,getargs(),ptr_t).p,
	    arg(1,getargs(),long));
    return v;
}


val_t eic_rewinddir(void)
{
    val_t v;
    rewinddir(arg(0,getargs(),ptr_t).p);
    return v;
}
val_t eic_closedir(void)
{
    val_t v;
    v.ival = closedir(arg(0,getargs(),ptr_t).p);
    return v;
}


/* STAT.H STUFF */
val_t eic_chmod(void)
{
    val_t v;

    v.ival = chmod(arg(0,getargs(),ptr_t).p,
		    arg(1,getargs(),mode_t ));
    return v;
}

val_t eic_fstat(void)
{
    val_t v;

    v.ival = fstat(arg(0,getargs(),int ),
		   arg(1,getargs(),ptr_t).p);

    return v;
}

val_t eic_mkdir(void)
{
    val_t v;

    v.ival = mkdir(arg(0,getargs(),ptr_t).p,
		    arg(1,getargs(),mode_t));
    return v;
}
val_t eic_mkfifo(void)
{
    val_t v;

    v.ival = mkfifo(arg(0,getargs(),ptr_t).p,
		    arg(1,getargs(),mode_t));
    return v;
}

val_t eic_stat(void)
{
    val_t v;

    v.ival = stat(arg(0,getargs(),ptr_t).p,
		  arg(1,getargs(),ptr_t).p);

    return v;
}

val_t eic_umask(void)
{
    val_t v;
    v.mval = umask(arg(0,getargs(),mode_t));    
    return v;
}


/* ASSERT.H STUFF */
val_t eic_assert()
{
    val_t v;
    fputs(arg(0,getargs(),ptr_t).p, stderr);
    fputs(" -- assert failed\n",stderr);
    return v;
}


/* TIME.H STUFF */


#include "time.h"
val_t eic_asctime()
{
    val_t v;
    v.p.p = v.p.sp = asctime(arg(0,getargs(),ptr_t).p);
    setEp(v.p,strlen((char*)v.p.p));
    return v;
}
val_t eic_clock()
{
    val_t v;
    v.lval = clock();
    return v;
}
val_t eic_ctime()
{
    val_t v;
    v.p.p = v.p.sp = ctime(arg(0,getargs(),ptr_t).p);
    setEp(v.p,strlen((char*)v.p.p));
    return v;
}

val_t eic_difftime()
{
    val_t v;
    time_t t0 = arg(0,getargs(),time_t);
    time_t t1 = arg(1,getargs(),time_t);
    v.dval = t0 > t1 ? t0 - t1: -t0 - t1;
    return v;
}


val_t eic_localtime()
{
    val_t v;
    v.p.p = v.p.sp =  localtime(arg(0,getargs(),ptr_t).p);
    setEp(v.p,sizeof(struct tm)-1);
    return v;
}

val_t eic_gmtime()
{
    val_t v;
    v.p.p = v.p.sp =  gmtime(arg(0,getargs(),ptr_t).p);
    setEp(v.p,sizeof(struct tm)-1);
    return v;
}


val_t eic_mktime()
{
    val_t v;
#ifndef _SUNOS
    v.lval = mktime(arg(0,getargs(),ptr_t).p);
#endif
    return v;
}


val_t eic_strftime()
{
    val_t v;
    v.szval = strftime(arg(0,getargs(),ptr_t).p,
		      arg(1,getargs(),size_t),
		      arg(2,getargs(),ptr_t).p,
		      arg(3,getargs(),ptr_t).p);
    return v;
}

val_t eic_time()
{
    val_t v;
    v.lval = time(arg(0,getargs(),ptr_t).p);
    return v;
}



/* UNISTD.H STUFF */

val_t eic_access()
{
    val_t v;
    v.ival = access(arg(0,getargs(),ptr_t).p,
		     arg(1,getargs(),int ));
    return v;
}


val_t eic_sleep()
{
    val_t v;
    v.uival = sleep(arg(0,getargs(),unsigned ));
    return v;
}

val_t eic_pipe(int *a)
{
     val_t v;
     v.ival = pipe(arg(0,getargs(),ptr_t).p);
     return v;
}
val_t eic_getpid(void)
{
    val_t v;
    v.ival = getpid();
    return v;
}
val_t eic_dup(int fd)
{
    val_t v;
    v.ival = dup(arg(0,getargs(),int ));
    return v;
}
val_t eic_dup2(int oldfd, int newfd)
{
    val_t v;
    v.ival = dup2(arg(0,getargs(),int),
		   arg(1,getargs(),int));
    return v;
    
}

val_t eic_fork(void)
{
    val_t v;
    v.pid = fork();
    return v;
}

val_t eic_unlink(void)
{
    val_t v;
    v.ival = unlink(arg(0,getargs(),ptr_t).p);
    return v;
}

val_t eic_link(void)
{
    val_t v;
    v.ival = link(arg(0,getargs(),ptr_t).p,
		  arg(1,getargs(),ptr_t).p);
    return v;
}


val_t eic_close(void)
{
    val_t v;
    arg_list ap = getargs();
    v.ival = close(arg(0,ap,int));
    return v;
}

val_t eic_read(void)
{
    val_t v;
    arg_list ap = getargs();
    v.sszval = read(arg(0,ap,int),
		    arg(1,ap,ptr_t).p,
		    arg(2,ap,unsigned));
    return v;
}
val_t eic_write(void)
{
    val_t v;
    arg_list ap = getargs();
    v.sszval = write(arg(0,ap,int),
		     arg(1,ap,ptr_t).p,
		     arg(2,ap,unsigned));
    return v;
}

val_t eic_lseek(void)
{
    val_t v;
    arg_list ap  = getargs();
    v.offval = lseek(arg(0,ap,int),
		   arg(1,ap,long),
		   arg(2,ap,int));
    return v;
}

val_t eic_rmdir(void)
{
    val_t v;
    v.ival = rmdir(arg(0,getargs(),ptr_t).p);
    return v;
}


/* ERRNO.H STUFF */
val_t _get_errno(void)
{
    val_t v;
    /*
     * This function exists so that EiC can get the address 
     * of the errno;
     */
    
    errno = 0;
    v.p.ep = v.p.sp = v.p.p = &errno;
    return v;
}


/*  STDARG.H STUFF */

val_t _StArT_Va(void)
{
    val_t v;
    getptrarg(0,v.p);
    v.p.ep = v.p.p;
    v.p.sp = (char*)v.p.p - 256 * sizeof(val_t);
    return v;
}

val_t _get_AR_t_size(void)
{
    val_t v;
    v.ival = sizeof(AR_t);
    return v;
}









