/* cdecl.c
 *
 *	(C) Copyright Apr 15 1995, Edmond J. Breen.
 *		   ALL RIGHTS RESERVED.
 * This code may be copied for personal, non-profit use only.
 *
 */
#include "stdio.h"
#include "stdlib.h"
#include "string.h"

#include "MachSet.h"
#include "lexer.h"
#include "global.h"
#include "typemod.h"
#include "func.h"
#include "xalloc.h"
#include "cdecl.h"
#include "preproc.h"
#include "error.h"
#include "typesets.h"
#include "parser.h"
#include "symbol.h"

extern token_t token;
extern int S_LEVEL;
extern char LEXEM[];

int work_tab;
static int LSP, RESET = 0;
int RETURNON = 0;
int INFUNC = 0;
int ABSDECL = 0;     /* constant for abstract declorator */
int INPARMLIST = 0;
token_t * INSTUN = 0;
type_expr *PRAMHANDLE;
token_t *RETURNEXPR;


void reset_env_pointers(token_t *e1,int bp);


/** PROTOTYPES from cdecl.c **/
int isredec(token_t * e1);
void establish_id(token_t * e1);
void new_var(token_t * e1);
void make_func(token_t * e1);
void add_func_parm(func_t * f, type_expr ** type, char * name);
void add_func_str(func_t * f, char *s);
void add_func_static(func_t * f, int n);
void addreturn(token_t *,code_t * c);
void free_func_inf(func_t * f);
int ext_decl(token_t * e1);
void f_ext_decl(token_t * e1);
void ff_ext_decl(token_t * e1);
void fff_ext_decl(token_t * e1);
void semi_colon(token_t * e1);
void store_class(token_t * e1);
void decl_spec(token_t * e1);
void type_spec2(token_t * e1, int t);
void type_spec(token_t * e1, int t);
void init_decl_list(token_t * e1);
void init_decl(token_t * e1, int t);
void initialiser(token_t * e1);
void decl(token_t * e1, int t);
type_expr * pointer(void);
void dir_decl(token_t * e1, int t);
void f_dir_decl(token_t * e1);
void ff_dir_decl(token_t * e1);
void parm_type_list(func_t * f);
void f_parm_type_list(func_t * f);
void parm_decl(token_t * e1);
void enum_spec(token_t * e1);
void f_enum_spec(token_t * e1);
void enum_list(token_t * e1);
void array_decl(token_t * e1);
void st_un_spec(token_t * e1, int t);
void f_st_un_spec(token_t * e1);
void s_decl_list(token_t * e1);
void st_decl(token_t * e1, int t);
void spec_qual_list(token_t * e1, int t);
void r_spec_qual_list(token_t * e1);
void free_un_mem(type_expr * e);
void addst_un_tag(symentry_t *sym);
void addst_un_mem(token_t * e1, token_t * e2);
void spec_declor_list(token_t * e1, token_t * e2);
void st_declor(token_t * e1, int t);
void f_st_declor();
int type_name(token_t * e1);
void abs_decl(token_t * e1, int t);
void f_abs_decl(token_t * e1);
void dir_abs_decl(token_t * e1, int t);
void f1_dir_abs(token_t * e1);
void f2_dir_abs(token_t * e1);


/* RoundUp returns x rounded to the next multiple
 * of n, which must be a power of two.
 */
#define RoundUp(x,n) (((x)+((n)-1))&(~((n)-1)))
#define init_ident(e,t)  (e)->val = token.val; (e)->tab = (t)

int isredec(token_t * e1)
{
    /* test for re declaration. */
    return sametypes(e1->type, e1->val.sym->type);
}

#if 0

#define setBoundaryLimits(x)

#else
void setBoundaryLimits(token_t *e1)
{
    int t = gettype(e1->type);
    if(e1->sclass == c_typedef) /* watch out for typedefs */
	return;
    if((!INFUNC || GI(e1) == 0) && t > t_pointer &&  
       !(t == t_func || t == t_funcdec || t == t_builtin)) {
	ptr_t *p;
	int s = get_sizeof(e1->type);
	
	if(!isconst(e1->type)) 
	    p = &ENV->AR[e1->sym->val.ival].v.p;
	else 
	    p  = &e1->sym->val.p;
	
	p->sp = p->p;
	    p->ep = (char*)p->p + s -1;
	    
    }
    /* automatic aggregate
     * types are done on the fly.
     */

}
#endif
   

int checklevel(token_t *e1,symentry_t * sym, int level)
{
    /*
     * Checks to see if there is a true difference between
     * the current scope and the declaration scope.
     */
    
    if(sym->level != level) {
	int t = gettype(e1->type);
	if((t == t_func || t == t_funcdec) && !INPARMLIST) {
	    return 0;
	} else 
	    return 1;
    } else
#if 0    
	if(level == 1 && sym->id != CurrentFileName()) {
	    if (e1->sclass & c_static)
		return 1;
	}
#endif
    return 0;
}

void setscope(symentry_t * sym,int level, int t)
{
    void UpdateSymPos(symentry_t * sym);
    if((t == t_func || t == t_funcdec) && !INPARMLIST) {
	sym->level = 1;
    } else if(sym->sclass == c_extern)
	sym->level = 1;
    else
	sym->level = level;
    if(sym->level < level ||
       (sym->next && sym->next->level > sym->level))
	UpdateSymPos(sym);
}

void freeArray(token_t *e1)
{
    if(get_sizeof(e1->val.sym->type))
	xfree(ENV->AR[e1->val.sym->val.ival].v.p.p);
    ENV->AR[e1->val.sym->val.ival].v.p.p = NULL;
}

void check_decl(token_t * e1)
{
    int t;
    type_expr *ty = e1->type;
    for(;ty;ty=nextType(ty))
	if((t=gettype(ty)) == t_funcdec || t == t_func) {
	    t = gettype(nextType(ty));
	    if(t == t_array || t == t_funcdec)
		error("Illegal return type for %s",
		      e1->val.sym->id);
	}
}

void doBuiltin(token_t * e1)
{
    void UpdateEntry(symentry_t * sym);
    if(token.tok == ';') { /* must be a prototype */
	UpdateEntry(e1->val.sym);
	if(nextType(e1->val.sym->type)) {
	    warningerror("2nd prototype for builtin -> %s",
			 e1->val.sym->id);
	}
	freetype(e1->val.sym->type);
	exchtype(t_builtin,e1->type);
	e1->val.sym->type = e1->type;
	e1->sym = e1->val.sym;
    }else
	error("Illegal redefinition of builtin function %s",
	      e1->val.sym->id);
}

void showRedec(token_t *e1)
{
   char *  getClashedfname(char nspace,char *id);
   extern int Pclash;
   char *fn;

   error(" Redeclaration of parameter `%s'",e1->val.sym->id);
   if(Pclash)
       fn = getClashedfname(work_tab,e1->val.sym->id);
   else
       fn = e1->val.sym->fname;
   
   fprintf(stderr,"Previously declared in: %s\n",fn);

   /* generate a dummy entry */
   e1->val.sym = insertLUT(e1->tab,
			   e1->val.sym->id, ID);
}

void establish_id(token_t * e1)
{
    extern int Pclash;
    int level,t;
    void UpdateEntry(symentry_t * sym);
    
    if ((t=gettype(e1->val.sym->type)) == ID) {
	/* variable not declared  previously,
	 * but check for possible clashes with
	 * previously declared static variables
	 */
	if(Pclash && !(e1->sclass & c_static))
	    showRedec(e1);
    } else if(e1->sclass != c_extern && 
	    ( checklevel(e1,e1->val.sym,S_LEVEL) 
	    || e1->tab != e1->val.sym->nspace))
	e1->val.sym = insertLUT(e1->tab, e1->val.sym->id, ID);
    else if (isredec(e1)) {
	if(INFUNC && S_LEVEL == 2 && e1->val.sym->val.ival < 0)
	    showRedec(e1);
	/* catch declaration after definition */
	if(t == t_func) {
	    /* Swap  Parmameter Lists */

	    if(token.tok == '{') { /* is definition */
		if(e1->val.sym->fname == CurrentFileName()) {
		    func_t *f2 = (func_t *)getInf(e1->type);
		    if(hasPrototype(f2)) 
			swapFPLists(getInf(e1->val.sym->type),f2);
		    UpdateEntry(e1->val.sym);
		} else
		    showRedec(e1);
	    }
	    
	    freetype(e1->type);
	    e1->type = e1->val.sym->type ;
	    e1->sym = e1->val.sym;
	    return;
	    
	} else if(t == t_builtin) {
	    doBuiltin(e1);
	    return;
	} else if(CurrentFileName() != e1->val.sym->fname &&
		  (e1->sclass & c_static)) {
	    e1->val.sym = insertLUT(e1->tab, e1->val.sym->id, ID);
	} else if(CurrentFileName() != e1->val.sym->fname &&
		  e1->sclass != c_extern &&
		  e1->val.sym->sclass != c_extern &&
		  t != t_funcdec) {
	    showRedec(e1);
	} else {
	    UpdateEntry(e1->val.sym);
	    if((t=gettype(e1->type)) == t_array) {
		if(e1->sclass != c_extern)
		    freeArray(e1);
	    } else if(t == t_struct || (t == t_union)) {
		/* use original copy */
		freetype(e1->type);
		e1->type = e1->val.sym->type ;
	    }
	    e1->sym = e1->val.sym;
	    return;
	}
    } else if(t == t_builtin) {
	doBuiltin(e1);
	return;
    } else {
	if(e1->val.sym->level == S_LEVEL &&
	   !(e1->sclass & c_static && 
	   e1->val.sym->fname != CurrentFileName()))	    
	    showRedec(e1);
	else
	    /* generate space in lookup table */
	    e1->val.sym = insertLUT(e1->tab, e1->val.sym->id, ID);
    }
    e1->val.sym->sclass = e1->sclass;
    setscope(e1->val.sym,S_LEVEL,gettype(e1->type));
    e1->val.sym->nspace = e1->tab;
    e1->sym = e1->val.sym;
    if (e1->sclass == c_static) {
	level = 1;
	if(S_LEVEL == 1) {
	    /* mark as private */
	    e1->sym->sclass |= c_private;
	}
	    
    } else
	level = e1->val.sym->level; /*S_LEVEL;*/
    /*
     * N.B. if changes are made to the condition 
     * for stacking, make sure that the free_sym 
     * function remains consistent.
     */
    if (!isconst(e1->type) &&
	e1->tab == stand_tab && e1->sclass != c_typedef &&
	e1->val.sym->val.ival == -1) 
	stackit(e1->val.sym, level);
}

#if 1


static size_t TempSz = 0;

void clearTempories()
{
    void updateLocals(void);
    extern unsigned  CurTemp,ASPOT;
    if(CurTemp) {
        updateLocals();
	ENV->lsp = ENV->lsp > CurTemp ?  ENV->lsp - CurTemp: 0;
        ASPOT = ASPOT > TempSz ? ASPOT - TempSz : 0;
        TempSz = CurTemp = 0;
    }
}
 
void newSlot(token_t * E, size_t sz, val_t *v, int align)
{
    extern unsigned ASPOT;
    /* Non static locals */
    static val_t v2;

    if(IsTemp(E->type))
        TempSz += (sz + RoundUp(ASPOT,align) - ASPOT);
    
    if (ASPOT != 0)
	ASPOT = RoundUp(ASPOT,align);

    v2.ival = -1;

    /* the lda instruction relies on stoptr
     * being the next instruction, so don't change
     * unless made compatible with 'lda' usage 
     * in interpret.c
     */
#if 1
    generate(&E->code,lda,&v2,ASPOT); 
    generate(&E->code, stoptr, v, 1);
    ASPOT += sz;
#else
    ASPOT += sz;
    generate(&E->code,lda,&v2,ASPOT-1); 
    generate(&E->code, stoptr, v, 1);
#endif    
    
}
#endif

void new_var(token_t * e1)
{
    int t = gettype(e1->type);
  
    e1->type = revtype(e1->type);
    establish_id(e1);
    check_decl(e1);
    if (e1->sym) {
	if(gettype(e1->type) == t_enum && e1->tab != tag_tab)
	{ /* hack for enumeration variables, catch and cast them
	     to integers */
	    e1->type = addtype(t_int, freetype(e1->type));
	    setAsBaseType(e1->type);
	}
	newsymtype(e1->sym, e1->type);    
	if (!(e1->sym->sclass == c_typedef) && !INPARMLIST)
	    if (e1->tab == stand_tab &&
		((t = gettype(e1->type)) == t_array ||
		 t == t_struct || t == t_union)) {
		val_t v;
		token_t *E;
		v.ival = get_sizeof(e1->sym->type);
		/*
		 * Here, we must consider 3 types
		 * of aggregate data: (1) local,
		 * (2) local but static and
		 * (3) global.
		 * Global and local static data
		 * get placed on the global stack.
		 * Local data goes on the local stack.
		 */
		if (INFUNC && (e1->sym->sclass & c_static)) { 
		    E = RETURNEXPR;
		    add_func_static(getInf(RETURNEXPR->type),
		   		    e1->sym->val.ival);
		} else
		    E = e1;
	
		if (/*IsTemp(E->type) ||*/ (E->sym->level > 1 && E != RETURNEXPR)) {

#if 1
		    newSlot(E,v.ival,&e1->sym->val,get_align(e1->type));
#else 		    
		    extern unsigned ASPOT,SlotNo;
		    /* Non static locals */
		    val_t v2;		   
		    if(SlotNo == -1)
			SlotNo = nextstackitem(S_LEVEL);
		   
		    v2.ival = SlotNo;
		    generate(&E->code, rvalptr, &v2, 1);
		    if (ASPOT != 0) {
			ASPOT = RoundUp(ASPOT,get_align(e1->type));
			v2.ival = 1;
			generate(&E->code, bump, &v2, 0);
			v2.ival = ASPOT;
			generate(&E->code, pushint, &v2, 0);
			generate(&E->code, addptr2int, &v2, 0);
		    } 
			
		    
		    generate(&E->code, stoptr, &e1->sym->val, 1);
		    ASPOT += v.ival;
#endif		    
		} else {
		    /*
		     * Globals and static local arrays/structs
		     * are made on the fly. However, if not
		     * NULL, assume memory has already been allocated.
		     */
		    int sz;
		    sz = v.ival > 1? v.ival:1;
		    
		    if(isconst(e1->type)) {
			if(e1->sym->val.ival == -1) /* not very safe ! */
			    e1->sym->val.p.p = (void*)
				xcalloc(1,sz);
		    } else {
			if(!ENV->AR[e1->sym->val.ival].v.p.p)
			    ENV->AR[e1->sym->val.ival].v.p.p
				= (void*)xcalloc(1,sz);
		    }
		}
	    }
    }
}

void addreturn(token_t * e1, code_t * c)
{
    val_t v;
    int  i, n, lcode, rtn;
    int err = 1; /* expect an error */
    int analyseCode(code_t *c);

    if(ParseError)
	return;

    n = nextinst(c) - 1;
    lcode = opcode(c,n);
    rtn = analyseCode(c);
    
    if(lcode == eicreturn && rtn <= n)
	return;
	/*printf("rtn = %d possible %s\n",rtn,e1->sym->id);*/
    
    if (lcode == fmem) { /* free memory */
	/* The last instruction is fmem. Thus, force all
	 * return calls within the function
	 * to exit via fmem
	 */
	if(rtn <= n - 1  && opcode(c,n-1) == eicreturn)
	    err=0; /* no possible error */
	for(i = 0; i < n; ++i)
	    if(c->inst[i].opcode == eicreturn) {
		c->inst[i].opcode = jmpu;
		c->inst[i].val.ival = n - i;
	    }
	rtn = n;
    }
    if(rtn >= n) {
    	generate(c, eicreturn, &v, 0);
        if(gettype(nextType(e1->type))  != t_void && err)
	    warningerror("Flow reaches end "
			 "of non-void function `%s'",
			 e1->sym->id);
    }
}

int ext_decl(token_t * e1)
{
    int h;
    e1->pflag = 0;
    switch (lexan()) {
      TYPEQUAL:
      STORECLASS:
      TYPESPEC:
	h = RESET;
	RESET = 0;
	decl_spec(e1);
	f_ext_decl(e1);
	clearTempories();
	RESET = h;
	break;
      default:
	retractlexan();
	return 0;
    }
    return 1;
}

void f_ext_decl(token_t * e1)
{
    switch (lexan()) {
      DECL:
	decl(e1, token.tok);
	ff_ext_decl(e1);
	break;
      case ';':
	break;
      default:
	error("Declaration error");
	match(';', " ; ");
    }
}

token_t *genTemporay(type_expr *type, int level)
{
    token_t *e1 = xcalloc(sizeof(token_t),1);
    extern symentry_t * nxtTemp(int obj, int level);
    symentry_t * sym;

    sym = nxtTemp(ID,level);
    inittoken(e1);
    SetTemp(type);
    e1->type = copytype(type);
    e1->val.sym =  sym;
    e1->tab = sym->nspace;
    new_var(e1);
    e1->val = sym->val;
    return e1;

    
}
    
void ff_ext_decl(token_t * e1)
{

    extern void UpdateEntry(symentry_t * sym);

    int t;
    switch (lexan()) {
      case '{': /* handle funtion definition */

	if (S_LEVEL > 1) {
	  error("Illegal function definition in %s",RETURNEXPR->sym->id);
	  break; 
	}
	new_var(e1);

	if((t = gotMissingNames(getInf(e1->type))))
            error("Missing name for Paramater %d",t);
	
	if(ErrorRecover) /* force recovery */
	    ErrorRecover = 0;
	
	if((t = gettype(e1->type)) == t_funcdec || t == t_func) {
	    token_t e2;
	    code_t *code;

	    inittoken(&e2);

	    if(HasHiddenParm(e1->type)) 
		addoffsettolevel(stand_tab, S_LEVEL+1,2);
	    else
		addoffsettolevel(stand_tab, S_LEVEL+1,1);

	    retractlexan();
	    if (e1->sym->ass) {
		warningerror("Function Re-definition of %s",e1->sym->id);
		freeFuncComments(getInf(e1->type));
       	    }

	    /* Update declaration to definition. */
	    exchtype(t_func, e1->type);
	    INFUNC = 1;
	    RETURNON++;
	    RETURNEXPR = e1;
	    stmt(&e2);

	    RETURNON--;
	    reset_env_pointers(e1, LSP);

	     /* remove reducear,  if present */
	    if(e2.code.nextinst != 0) { 
		if (e2.code.inst[e2.code.nextinst - 1].opcode ==
		    reducear)
		    e2.code.nextinst--;
		    addreturn(e1,&e2.code);
	    } else {
		warningerror("Empty function definition");
		generate(&e2.code, eicreturn, &e2.val, 0);    
	    }
	    if(!ParseError) {
		if(e1->sym->ass)
		    killcode(e1->sym);
		code = (code_t *) xcalloc(1, sizeof(code_t));
		e1->sym->ass = 1;
		*code = e2.code;
		ENV->AR[e1->sym->val.ival].v.p.p = code;
		codeName(code) = CurrentFileName();
	    } else {
		void cleancode(code_t *);
		cleancode(&e2.code);
		freecode(&e2.code);
	    }
		
	    freetype(e2.type);
	    e1->type = NULL;	/* hide new type */
	    INFUNC = 0;
	} else {
	    e1->type = NULL;
	    retractlexan();
	}
	break;
      default:
	if (e1->val.sym) 
	    new_var(e1);
	remlevel(S_LEVEL + 1);
	if(token.tok == ASS) {
	    initialiser(e1);
	    lexan();
	}
	fff_ext_decl(e1);
    }
}

void fff_ext_decl(token_t * e1)
{
    if (RESET) {
	reset_env_pointers(e1, LSP);
	RESET--;
    }
       setBoundaryLimits(e1);

    if (token.tok == ',') {
	init_decl_list(e1);
	e1->type = NULL;	/* hide new type */
	match(';', " ;");
    } else {
	retractlexan();
	semi_colon(e1);
    }
}

void semi_colon(token_t * e1)
{
    e1->type = NULL;	/* hide new type */
    if(!match(';', ";"))
	retractlexan();
}

void decl_spec(token_t * e1)
{
    int sclass = 0;
    void specifier(token_t *e1, int,int *sclass);

    switch (token.tok) {
      STORECLASS:
      TYPEQUAL:
      TYPESPEC:
	specifier(e1, token.tok,&sclass);
	if ((sclass == c_auto || sclass == c_register)
	    && !(INFUNC || S_LEVEL > 1))
	    error("Illegal storage class usage");
	else if(INPARMLIST && sclass && sclass != c_register)
	    error("Illegal storage class for parameter %s",LEXEM);
	else
	    e1->sclass = sclass;
	setAsBaseType(e1->type);
      default:
	retractlexan();
    }
}

void specifier(token_t *e1, int t,int *sclass)
{
    /* This function was modelled from lcc32's
     * specifier function
     */

    int cls, cons, sign, size, type, vol,ty;
    cls = ty = vol = cons = sign = size = type = 0;

    if(sclass == NULL)
	cls = c_auto;
        
     while(1){
	int *p, tt = t;
	switch (t) {
	  case autosym:     tt = c_auto;     p = &cls; break;  
	  case externsym:   tt = c_extern;   p = &cls; break;  
	  case registersym: tt = c_register; p = &cls; break;  
	  case staticsym:   tt = c_static;   p = &cls; break;  
	  case typedefsym:  tt = c_typedef;  p = &cls; break;  
	  case constsym:   p = &cons;       break;
	  case volatilesym:p = &vol;        break;
	  case signedsym:
	  case unsignedsym:p = &sign;       break;
	  case longsym:
	  case shortsym:   p = &size;       break;
	  case voidsym:    ty = t_void;   p = &type;   break;
	  case charsym:    ty = t_char;   p = &type;   break;
	  case intsym:     ty = t_int;    p = &type;   break;
	  case floatsym:   ty = t_float;  p = &type;   break;
	  case doublesym:  ty = t_double; p = &type;   break;
	  case enumsym:     
	  case structsym:
	  case unionsym:   {
	      token_t e2;
	      p = &type;
	      inittoken(&e2);
	      if (t == enumsym)
		  enum_spec(&e2);
	      else
		  st_un_spec(&e2, t);
	      /*
	       *  test tag name.
	       */
	      if (e2.val.sym) {
		  e1->type = copytype(e2.type);
		  e1->tab = stand_tab;
		  if(INSTUN)
		      addst_un_tag(e2.val.sym);
	      } else {
		  int sclass = e1->sclass;
		  *e1 = e2;
		  e1->sclass = sclass;
	      }
	  }
	    break;
	  case TYPENAME:
	    if(type == 0) {
		e1->type = copytype(token.val.sym->type);
		p = &type;
	    } else { /* allow for masking of a typedef name */
		token.tok = ID;
		retractlexan();
		p = NULL;
	    }
	    break;
	    
	  default:
	    retractlexan();
	    p = NULL;
	}
	if (p == NULL)
	    break;
	if (*p)
	    error("invalid specification");
	*p = tt;
	t = lexan();
    }
    if(sclass)
	*sclass = cls;
    if(type == 0) {
	type = intsym;
	ty = t_int;
    }
    if ((size == shortsym && type != intsym)
        ||  (size == longsym  && type != intsym && type != doublesym)
        ||  (sign && type != intsym && type != charsym))
                error("invalid type specification");
    if (type == charsym && sign)
	ty = sign == unsignedsym ? t_uchar : t_char;
    else if (size == shortsym)
	ty = sign == unsignedsym ? t_ushort : t_short;
    else if (size == longsym && type == doublesym)
	ty = t_double;
    else if (size == longsym)
	ty = sign == unsignedsym ? t_ulong : t_long;
    else if (sign == unsignedsym && type == intsym)
	ty = t_uint;
    if(ty)
	e1->type = addtype(ty,e1->type);

    if (cons == constsym) {
	if(INPARMLIST || INFUNC)  /* */
	    setConstp(e1->type);
	else 	    
	    setConst(e1->type);
	/*
	if(INFUNC && !INPARMLIST && sclass) 
	*sclass = c_static;
	*/
    }
    /* ignore volatile for now;
       if (vol  == VOLATILE)
       ty = qual(VOLATILE, ty);
       */
}

void init_decl_list(token_t * e1)
{
    token_t e2;
    do {
	switch (lexan()) {
	  DECL:
	    inittoken(&e2);
	    e2.type = copyBaseType(e1->type);
	    e2.sclass = e1->sclass;
	    
	    init_decl(&e2, token.tok);
	    if (RESET) {
		reset_env_pointers(&e2, LSP);
		RESET--;
	    }
	    concode(&e1->code, &e2.code);
	    remlevel(S_LEVEL+1);
	    setBoundaryLimits(&e2);
	    break;
	    
	  default:
	    error("Init decl list error");
	}
    } while (lexan() == ',');
    retractlexan();
}

void init_decl(token_t * e1, int t)
{
    decl(e1, t);
    new_var(e1);
    if (lexan() == ASS) {
	initialiser(e1);
    } else
	retractlexan();
}

void initglobal(type_expr * type, void **addr,int lev);
void initstruct(type_expr * type, void **addr,int lev);
void do_struct(type_expr *type, void **addr,int lev);
int initarray(type_expr * type, void ** addr, int size, int lev);
void assign_var(type_expr *type, void *addr,int allow);
int do_array(type_expr *type, void ** addr, int size, int lev, int inc);
void SaveGlobalString(ptr_t *s);

void structUnionCode(token_t * e1, token_t *e2)
{
    val_t v;
    v = e1->val;
    e1->val = e1->sym->val;
    output(e1);
    e1->val = v;
    v.ival = 1;
    generate(&e1->code, bump, &v, 0);
    v.ival = get_sizeof(e2->type);
    generate(&e2->code,refmem,&v,0);
}

static token_t * E1;
void initialiser(token_t * e1)
{
    int t;
    int peek = lexan();
    retractlexan();
    
    if(GI(e1) == 0) {  /* global or static local variable */
	E1 = e1;
	if ((t =gettype(e1->type)) <= t_pointer) { /* get scalars */
	    int v = 1;
	    if(e1->sclass & c_static) /* catch statics */
		v = 0;
	    if(!isconst(e1->type))  		
		assign_var(e1->type,&ENV->AR[e1->sym->val.ival].v,v);
	    else
		assign_var(e1->type,&e1->sym->val,v);
	} else {
 	    if(!isconst(e1->type))
		initglobal(e1->type,&ENV->AR[e1->sym->val.ival].v.p.p,0);
	    else 
		initglobal(e1->type,&e1->sym->val.p.p,0);
	}
    } else {	
	if ((t =gettype(e1->type)) <= t_pointer /* local scalar types */
	    || ((t == t_struct || t==t_union) && peek != '{')  
	    ) { 
	    if(isconst(e1->type)) 
		assign_var(e1->type,&e1->sym->val,0);
	    else {
		token_t e2;
		val_t h; 
		inittoken(&e2);
		assign_expr(&e2);
		if (isconst(e2.type)) {
		    castconst(&e2,e1,0);	    
		    h = e1->val;
		    e1->val = e1->sym->val;
		    generate(&e1->code,pushval,&e2.val,0);
		    do_stooutput(e1);
		    e1->val = h;
		    freetoken(&e2);
		} else {
		    output(&e2);
		    castvar(&e2,e1,0);
		    if (t == t_struct || t == t_union)
			structUnionCode(e1,&e2);
		    concode(&e1->code,&e2.code);
		    h = e1->val;
		    e1->val = e1->sym->val;
		    do_stooutput(e1);
		    e1->val = h;
		}
		freetoken(&e2);
	    }
	} else if(INFUNC) {
	    /* code for initialising automatic aggregate types */
	    extern unsigned ASPOT;
	    int s2, s1 = get_sizeof(e1->type);
	    val_t v;
	    E1 = e1;
	    
	    if(s1)
		v.p.p = xcalloc(1,s1);
	    else
		v.p.p = NULL;
	    initglobal(e1->type,&v.p.p,0);
	    if((s2 = get_sizeof(e1->type)) > s1)
		ASPOT += s2 - s1;
	    add_func_initialiser(getInf(RETURNEXPR->type),v.p.p);
	    generate(&e1->code,minit,&v,s2);
	} else	
	    error("Initialisation not "
		  "supported for non scalar local types");	
    }
}

void initglobal(type_expr * type, void **addr, int lev)
{
    switch(gettype(type)) {
      case t_array:
	initarray(type,addr,0,lev); break;
      case t_struct:
	initstruct(type,addr,lev); break;
      case t_union:
    {
	struct_t * S;
	int n;
	S = getInf(type);
	n = S->n;
	S->n = 1;  /* force initialisation of first member only */	
	initstruct(type,addr,lev+1);
	S->n = n;
    }
	break;
      default:
	assign_var(type, *addr,0); break;
    }
}

void initstruct(type_expr * type, void **addr, int lev)
{ 
    if(lexan() == '{' ) {
	do_struct(type,addr,lev);
	if(lev > 1 && lexan() != ',')
		retractlexan();
	match('}', " }");
    } else { 
	retractlexan();
	if(lev > 0) 
	    do_struct(type,addr,lev);
	else /* else intializer must be an expression of the same type */
	    assign_var(type,*addr,1);
    }
}

void do_struct(type_expr *type, void **addr,int lev)
{
    struct_t *S;
    int i;
    S = getInf(type);
    for(i=0;i<S->n;i++) 
	if(lexan() != '}') {
	    retractlexan();
	    *addr = (char*)*addr + S->offset[i];
	    initglobal(S->type[i],addr,lev+1);
	    *addr = (char*)*addr - S->offset[i];
	    if(lexan() != ',')
		retractlexan();
	} else 
	    break;
    retractlexan();

}

int dostrlit(void ** addr,int ln, int sln, char * str)
{    /*
      * parse eg. char a[] = "hello world";
      * or  char a[5] = "hello world";
      */    
    xmark(str,eicgstring); /* mark for garbage collector */
    if(ln) {
	if(sln >= ln)
	    warningerror("Array of chars is too long");
    } else {
	ln = sln+1; /* allow for null at end */
	*addr = xrealloc(*addr,ln);
    }
   if(sln < ln)
	memcpy(*addr,str,sln+1);
   else
       memcpy(*addr,str,ln);
	
    return ln;
}

int initarray(type_expr * type, void ** addr, int size, int lev)
{
    static int INCREMENT;
    
    if(gettype(type) == t_array) {
	int t,tok,s;
	if((tok =lexan()) == '{') {
	    t = get_sizeof(type);
	    if(lev == 0) 
		INCREMENT = t == 0 ? get_sizeof(nextType(type)): 0;
	    s = do_array(nextType(type),addr,t,lev,INCREMENT);
	    if(t == 0) {
		t = s/get_sizeof(nextType(type));
		setNumElems(type,t);
	    }
	    match('}'," }");
	} else if(tok == STR) { /* handle string literals */
	    if(gettype(nextType(type)) == t_char) {
		size_t sln;
		t = get_sizeof(type);
		sln = (char*)token.val.p.ep - (char*)token.val.p.p;
		s = dostrlit(addr,t,sln,token.val.p.p);
		if(lev == 0 && t == 0)
		    setNumElems(type,s);
	    }else
		error("Illegal initialisation");
	} else{
	    retractlexan();
	    if(lev > 0) /* lev indicates a recusive call */
		size = do_array(nextType(type),addr,size,lev,INCREMENT);
	    else 
		error("missing { in initialisation of an array");
	}
    } else
	initglobal(type,addr,lev+1);
    return size;
}

int do_array(type_expr *type, void ** addr, int size, int lev, int inc)
{
    void *memset(void *s, int c, size_t n);
    int n = 0;
    while(lexan() != '}') {
	retractlexan();
	if(n >= size) {
	    if(inc) {
		size += inc;
		*addr = xrealloc(*addr, size);
		memset((char*)*addr + size - inc, 0, inc);
	    } else {
		error("Too many initialisers");
		return 0;
	    }
	}
	*addr = (char*)*addr + n;
	size = initarray(type,addr, size,lev+1);
	*addr = (char*)*addr - n;
	n += get_sizeof(type);
	if (lexan() != ',') 
	    break;
    } 
    retractlexan();
    return size;
}

void assign_var(type_expr *type, void *addr,int allow)
{
    int t;
    token_t e1,e2;
    void *getaddress(token_t *);
    

    inittoken(&e2);
    assign_expr(&e2);

#if 0
    if(IsTemp(e2.type))
	clearTempories();
#endif
    
    t = gettype(type);

    if (isconst(e2.type)) {
	e1.type = type;
	castconst(&e2,&e1,0);
	switch(t) {
	  case t_char:
	  case t_uchar:    *(char*)addr = e2.val.ival; break;
	  case t_short:
	  case t_ushort:  *(short*)addr = e2.val.ival; break;
	  case t_int:
	  case t_uint:     *(int *)addr = e2.val.ival; break;
	  case t_long:
	  case t_ulong:    *(long*)addr = e2.val.lval; break;
	  case t_float:   *(float*)addr = e2.val.dval; break;
	  case t_double: *(double*)addr = e2.val.dval; break;
	  case t_pointer: 
	    if(S_LEVEL == 1 && gettype(e2.type) == t_pointer
	       && gettype(nextType(e2.type)) == t_char) 
		/* got string */
		if(e2.val.p.p)
		    SaveGlobalString(&e2.val.p);
	    *(ptr_t*)addr = e2.val.p;
	    break;
	    
	  default:
	    error("Unknown initialiserXXX");
	}
    } else if(allow) {
	val_t h;
	token_t e3;
	e3.type = type;
	output(&e2);
	castvar(&e2,&e3,0);	
	if (t == t_struct || t == t_union) 
	    structUnionCode(E1,&e2);
	h = E1->val;
	E1->val = E1->sym->val;
	concode(&E1->code,&e2.code);
	do_stooutput(E1);
	E1->val = h;
    } else {
	if(GI(&e2) != 0)   /* global or static addresses only are allowed */
	    error("Illegal initialization: illegal address operation");
	else {
	    if(gettype(type) == t_pointer) {
		ptr_t p;
		p.sp = p.p = getaddress(&e2);
		p.ep = (char*)p.p + get_sizeof(e2.type) - 1;
		if(!sametypes(e2.type,type))
		    warningerror("Suspicious pointer conversion");
		*(ptr_t*)addr = p;
		freecode(&e2.code);
	    } else
		error("Expected constant expression as an"
		      " initialiser");
	}
    }
    freetoken(&e2);
}


void decl(token_t * e1, int t)
{
    type_expr *P = NULL;
    switch (t) {
      case '*':
	P = pointer();
	dir_decl(e1, lexan());
	if(P)
	    e1->type = catTypes(P,e1->type);
	break;
      case ID:
      case '(':
	dir_decl(e1, t);
	break;
      default:
	error("Declarator error");
	retractlexan();
    }

}

type_expr * pointer(void)
{
    type_expr *t = NULL;
    do {
	t = addtype(t_pointer,t);
	if(lexan() == constsym || token.tok == volatilesym) {
	    if(token.tok == constsym) {
#if 0		
		if(INPARMLIST)
		    setConstp(t);
		else
		    setConst(t);
#else
		setConstp(t);
#endif
	    }
	    /*ignore for now volatilesym*/
	    lexan();
	}
    } while (token.tok == '*');
    retractlexan();
    return t;
}

void dir_decl(token_t * e1, int t)
{
    switch (t) {
      case '(':
	decl(e1, lexan());
	match(')', " )");
	f_dir_decl(e1);
	break;
	
      case ID:

	init_ident(e1, work_tab);
	f_dir_decl(e1);
	break;
	
      default:
	error("Direct declarator error");
	break;
    }
}

void f_dir_decl(token_t * e1)
{
    while (1)
	switch (lexan()) {
	  case '[':
	    array_decl(e1);
	    break;
	  case '(':
	    ff_dir_decl(e1);
	    if(INPARMLIST)
		remlevel(S_LEVEL+1);
	    break;
	  default:
	    retractlexan();
	    return;
	}
}

void ff_dir_decl(token_t * e1)
{

    int h;
    switch ((h=lexan())) {
      TYPEQUAL:
      STORECLASS:
      TYPESPEC:
	S_LEVEL++;
	RESET++;
	h = LSP;
	LSP = ENV->lsp;
	retractlexan();
	make_func(e1);
	/*
	 * Use INPARMLIST to inform other
	 * modules that the following declarations
	 * are function parameters.
	 */
	INPARMLIST++;
	PRAMHANDLE = e1->type;

	parm_type_list(getInf(e1->type));
	/*
	 * now pseudo reverse the parameter
	 * order.
	 */
	reset_env_pointers(e1, LSP);
	LSP = h;
	RESET--;

	INPARMLIST--;
	S_LEVEL--;
	
	break;
      case ')':
	/*
	 * This should really be made illegal, it allows
	 * for function declarations with empty
	 * paramater list, such as:
	 * int f(); Therefore, force an implied t_var argument.
	 */
    {
	type_expr * type;
	make_func(e1);
	type = addtype(t_var,NULL);
	add_func_parm(getInf(e1->type), &type, NULL);
	freetype(type);
    }
	return;
      default:
	if(h == ID)
	    error("Unknown type '%s': possible "
		  "old C type declaration", token.val.sym->id);
	else
	    error("Syntax error");
    }
    match(')', " )");
}


void UpDateParmSym(type_expr *ty, symentry_t *sym)
{
    if(gettype(ty) == t_pointer && gettype(nextType(ty)) ==
       t_funcdec)
	if(gettype(sym->type) != t_pointer)
	    sym->type = addtype(t_pointer,sym->type);
}

void parm_type_list(func_t * f)
{
    char * name = NULL;
    token_t e2;
    inittoken(&e2);
    parm_decl(&e2);
    /*
     * Must watch out for void  as a parameter.
     * The void paramater will have no sym entry.
     */
    if(e2.val.sym) {
	new_var(&e2);
#if 1
	{
	    void adjustParam(type_expr **type);
	    adjustParam(&e2.sym->type);
	}
#endif

	name = e2.sym->id;
    } else  /* still must reverse type if needed */
	e2.type = revtype(e2.type);

    add_func_parm(f, &e2.type,name);
    if(!e2.val.sym)
	freetype(e2.type);
    else
	UpDateParmSym(getFPty(f,getFNp(f)-1),e2.val.sym);
    
    if (lexan() == ',')
	f_parm_type_list(f);
    else
	retractlexan();
}

void f_parm_type_list(func_t * f)
{
    if (lexan() == '.') {
	if (lexan() == '.')
	    if (lexan() == '.') {
		type_expr *type;
		type = addtype(t_var, NULL);
		add_func_parm(f, &type,NULL);
		freetype(type);
		return;
	    }
	retractlexan();
	error("Expected ...");
    } else {
	retractlexan();
	parm_type_list(f);
    }
}

void fff_parm_decl(token_t *e1);
void ff_parm_decl(token_t *e1);
void f_parm_decl(token_t * e1);

void parm_decl(token_t * e1)
{
    switch (lexan()) {
      TYPEQUAL:
      STORECLASS:
      TYPESPEC:
	decl_spec(e1);
	f_parm_decl(e1);
	break;
      default:
	if(token.tok == ID)
	    error("Unknown type '%s'",token.val.sym->id);
	else
	    error("Parameter declaration error");
    }
}


void f_parm_decl(token_t * e1)
{
    if(lexan() == '*') {
	type_expr *P = NULL;
	P = pointer();
	ff_parm_decl(e1);
	if(P)
	    e1->type = catTypes(P,e1->type);	
    } else {
	retractlexan();
	ff_parm_decl(e1);
    }
}


void ff_parm_decl(token_t *e1)
{
    switch(lexan()) {
      case ID: init_ident(e1, work_tab);
	f_dir_decl(e1);
	break;
      case '(': fff_parm_decl(e1);
	break;
      case '[': retractlexan();
	f_dir_decl(e1);
	break;
      default: /* null */
	if(token.tok  != ',' && token.tok != ')')
	    error("Parameter declaration error");
	else
	    retractlexan();
	return ;	
    }
}

void fff_parm_decl(token_t *e1)
{
    switch(lexan()) {
      TYPEQUAL:
      STORECLASS:
      TYPESPEC:
	retractlexan();
	ff_dir_decl(e1);
	remlevel(S_LEVEL+1);
	break;
      default:
	retractlexan();
	f_parm_decl(e1);
	match(')', " ) "); 
	break;
    }
    f_dir_decl(e1);
}


void enum_spec(token_t * e1)
{
    e1->type = addtype(t_enum, NULL);
    f_enum_spec(e1);
}

void f_enum_spec(token_t * e1)
{
    int h, t;
    h = work_tab;
    work_tab = tag_tab;
    switch (lexan()) {
      case ID:			/* enumeration tag */
	if ((t = gettype(token.val.sym->type)) == ID) {
	    init_ident(e1, work_tab);

	    new_var(e1);
	    work_tab = h;

	    if (lexan() == '{') {
		enum_list(e1);
		match('}', " }");
		 e1->type = e1->val.sym->type;
	    } else
                retractlexan();

	} else if (t == t_enum) {
	    freetype(e1->type);
	    e1->type = copytype(token.val.sym->type);
	} else
	    error("Enumeration declaration error");
	break;
      case '{':
	work_tab = h;
	enum_list(e1);
	match('}', " }");
	break;

      default:
	error("Enumeration declaration error");
	retractlexan();
    }
    work_tab = h;
}

void enum_list(token_t * e1)
{
    int ENUMVAL = 0;
    token_t e2;
    token_t e3;
    do {
	inittoken(&e2);
	/*
	 *  now do enumerator part
	 */
	if (lexan() == ID) {
	    e2.type = addtype(t_int, NULL);
	    setConst(e2.type);
	    setAsBaseType(e2.type);
	    init_ident(&e2, work_tab);
	    new_var(&e2);
	    if (lexan() == ASS) {
		inittoken(&e3);
		cond_expr(&e3);
		if (isconst(e3.type) && gettype(e3.type) == t_int)
		    ENUMVAL = e3.val.ival;
		else {
		    freetoken(&e3);
		    error("Assignment must be constant");
		    break;
		}
		freetoken(&e3);
	    } else
		retractlexan();
	    e2.sym->val.ival = ENUMVAL++;
	} else {
	    error("Expected identifier in enumeration");
	    break;
	}
    } while (lexan() == ',');
    retractlexan();
}

void array_decl(token_t * e1)
{
    int t;
    token_t e2;
    inittoken(&e2);
    assign_expr(&e2);
    if (isconst(e2.type)) { 
	if ((t=gettype(e2.type)) != t_int && t != t_uint) {
	    if(sizeof(int) != sizeof(long) || t != t_ulong) {
		error("Constant int expression needed");
		e2.val.ival = 0;
	    }
	}
    } else {
	if(!ABSDECL && /*((t == t_void && INFUNC) || */
	   gettype(e2.type) != t_void )
	    error("Constant int expression needed");
	e2.val.ival = 0;
    }
    e1->type = addtype(t_array, e1->type);
    setNumElems(e1->type,e2.val.ival);
    freetoken(&e2);
    match(']', " ] ");
}


void st_un_spec(token_t * e1, int t)
{
    switch (t) {
      case structsym:
	e1->type = addtype(t_struct, e1->type);
	break;
      case unionsym:
	e1->type = addtype(t_union, e1->type);
	break;
    }
    f_st_un_spec(e1);
}

static eicstack_t stun;
void f_st_un_spec(token_t * e1)
{
    val_t val;
    int t,tok=0;
    symentry_t *sym;
    int h = work_tab;
    work_tab = tag_tab;

    val.p.p = INSTUN;
    INSTUN = e1;
    eicpush(&stun,val);
    switch (lexan()) {
      case ID:			/* struct or union tag */
	work_tab = h;
	sym = token.val.sym;
	t = gettype(sym->type);
       	if(t==ID || ((tok=lexan())=='{'
	   && token.val.sym->level<S_LEVEL)) {
	    init_ident(e1, tag_tab);
	    new_var(e1);
	    setInf(e1->type,xcalloc(1, sizeof(struct_t)));
	    if(!tok)
		tok = lexan();
	    if(tok == '{') {
		S_LEVEL++;
		s_decl_list(e1);
		match('}', " }");
		S_LEVEL--;
		if(lexan() == ';' && S_LEVEL == 1)
		    e1->type = NULL;
		retractlexan();
		   
	    } else
		retractlexan();		    
	} else {
	    if (t == t_struct || t == t_union) {
		freetype(e1->type);
		e1->type = copytype(sym->type);
	    } else {
		init_ident(e1, tag_tab);
		new_var(e1);
		setInf(e1->type,xcalloc(1, sizeof(struct_t)));
	    }
	    if(tok == '{') {
		S_LEVEL++;
		s_decl_list(e1);
		match('}', " }");
		S_LEVEL--;
		if(lexan() == ';' && S_LEVEL == 1) {
		    freetype(e1->type);
		    e1->type = NULL;
		}
		retractlexan();
	    } else
		retractlexan();
	}
	break;
      case '{':
	work_tab = h;
	S_LEVEL++;
	s_decl_list(e1);
	match('}', " }");
	S_LEVEL--;
	break;
    }
    work_tab = h;
    eicpop(&stun,&val);
    INSTUN = val.p.p;

}


void s_decl_list(token_t * e1)
{
    int f = 0, bp = ENV->lsp;
    struct_t *S;

    
    if(!getInf(e1->type)) 
	setInf(e1->type,xcalloc(1, sizeof(struct_t)));
    /*
    if(((struct_t*)getInf(e1->type))->cl) {
	error("Illegal use of struct/union");
	return;
    }
    */
    
    while (1) {
	switch (lexan()) {
	  TYPESPEC:
	  TYPEQUAL:
		f = 1;
	    st_decl(e1, token.tok);
	    break;
	  default:
	    if (!f)
		error("Struct/Union list declaration error");
	    S = getInf(e1->type);
	    S->cl = 1;
	    /*
	     * determine structure alignment.
	     */

	    S->tsize = RoundUp(S->tsize,S->align); 
	    retractlexan();
	    reset_env_pointers(e1, bp);
	    return;
	}
    }
}

void st_decl(token_t * e1, int t)
{
    token_t e2;
    inittoken(&e2);
    spec_qual_list(&e2, t);
    setAsBaseType(e2.type);
    spec_declor_list(e1, &e2);
    freetype(e2.type);

    match(';', " ; ");
}

void spec_qual_list(token_t * e1, int t)
{
    /*type_spec(e1, t);*/
    specifier(e1,t,NULL);
    r_spec_qual_list(e1);
}

void r_spec_qual_list(token_t * e1)
{
    switch (lexan()) {
      TYPESPEC:
	    spec_qual_list(e1, token.tok);
	break;
      default:
	retractlexan();
    }
}

void free_un_mem(type_expr * e)
{
    struct_t *S = getInf(e);
    if (S) {
	int i;
	if(S->n) {
	    for (i = 0; i < S->n; i++) {
		xfree(S->id[i]);
		freetype(S->type[i]);
	    }
	    xfree(S->offset);
	    xfree(S->id);
	    xfree(S->type);
	}
	if(S->ntags) {
	    for(i=0;i<S->ntags;++i)
		freetype(S->tag[i]);
	    xfree(S->tag);
	}
	xfree(S);
    }
}

void addst_un_tag(symentry_t *sym)
{
    struct_t *S;
    S = getInf(INSTUN->val.sym->type);
    S->ntags++;
    if(S->ntags == 1)
	S->tag = (type_expr**)xmalloc(sizeof(type_expr*));
    else
	S->tag = (type_expr**)xrealloc(S->tag,
				      sizeof(type_expr*)*S->ntags);
    S->tag[S->ntags - 1] = transtype(sym->type);
    setaliases(sym->type,1);
}
	
void addst_un_mem(token_t * e1, token_t * e2)
{
    struct_t *S;
    int off, s;

    S = getInf(e1->type);
    if (S->n == 0) {
	S->id = (char **) xcalloc(1, sizeof(char *));
	S->offset = (int *) xcalloc(1, sizeof(int));
	S->type = (type_expr **) xcalloc(1, sizeof(type_expr *));
    } else {
	/*
	 * first check for duplication of names
	 */
	int i;
	for(i=0;i<S->n;++i) 
	    if(strcmp(S->id[i], e2->val.sym->id) == 0)
		error("Duplication of identifier in struct/union");

	S->id = (char **) xrealloc(S->id,
				   (S->n + 1) * sizeof(char *));
	S->offset = (int *) xrealloc(S->offset,
				     (S->n + 1) * sizeof(int));
	S->type = (type_expr **) xrealloc(S->type,
					  (S->n + 1) * sizeof(type_expr *));
    }

    if(isconst(e2->type)) {
	unsetConst(e2->type);
	setConstp(e2->type);
    }
    S->id[S->n] = (char *) strsave(e2->val.sym->id);
    S->type[S->n] = transtype(e2->type);
    
    
    if(!(s = get_sizeof(S->type[S->n]))) {
     error("Incomplete data type for struct/union member %s",S->id[S->n]);
    }
    off = get_align(e2->type);
	
    if (gettype(e1->type) == t_struct) {
        
	S->offset[S->n] = RoundUp(S->tsize,off);
	S->tsize += s + S->offset[S->n] - S->tsize;
    } else { /* handle union */	
	S->offset[S->n] = 0;
	S->tsize = S->tsize > s ? S->tsize : s;
    }
    /*
     * structure alignment is equal to the
     * maximum alignment of its members.
     */
    if(S->align < off)
	S->align = off;
    S->n++;
}

void spec_declor_list(token_t * e1, token_t * e2)
{
    int ftime = 1, cl;
    token_t e3;
    cl = ((struct_t*)getInf(e1->type))->cl;
    do {
	switch (lexan()) {
	  case '*':
	  case ID:
	  case '(':
	  case ':':
	    inittoken(&e3);
	    e3.type = transtype(e2->type);
	    st_declor(&e3, token.tok);
	    if (gettype(e3.type) == t_struct ||
		gettype(e3.type) == t_union) {
		struct_t *S;
		S = getInf(e3.type);
		if (!S->cl)	/* check for closure */
		    error("Illegal Struct/Union member");
	    }
	    /* if structure/union is already closed,
	     * then assume structure is being re-entered.
	     */
	    if (e3.val.sym && !cl) {
		e3.type = revtype(e3.type);
		if (ftime)
		    setaliases(e2->type, 1);
		addst_un_mem(e1, &e3);
		if (ftime) {
		    setaliases(e3.type, 1);
		    ftime = 0;
		}
	    }
	    freetype(e3.type);
	    remsym(e3.val.sym);
	    break;
	  default:
	    retractlexan();
	    error("Struct/Union member declaration error");
	}
    } while (lexan() == ',');
    retractlexan();
}

void st_declor(token_t * e1, int t)
{
    switch (t) {
      case '*':
      case ID:
      case '(':
	  decl(e1, t);
	  /*setBoundaryLimits(e1);*/
	  f_st_declor();
	  establish_id(e1);
	  break;
      case ':':
	  error("Bit fields not supported");
	  break;
    }
}

void f_st_declor()
{
    if (lexan() == ':')
	error("Bit fields not supported");
    else
	retractlexan();
}

int type_name(token_t * e1)
{
    switch (lexan()) {
      TYPEQUAL:
      TYPESPEC:
	spec_qual_list(e1, token.tok);
	setAsBaseType(e1->type);
	switch (lexan()) {
	  case '*':
	  case '(':
	  case '[':
	    abs_decl(e1, token.tok);
	    e1->type = revtype(e1->type);
	    break;
	  default:
	    retractlexan();
	}
	break;
      default:
	retractlexan();
	return 0;
    }
    return 1;
}

void abs_decl(token_t * e1, int t)
{
    type_expr *P = NULL;
    ABSDECL++;
    switch (t) {
      case '*':
	P = pointer();
	f_abs_decl(e1);
	if(P)
	    e1->type = catTypes(P,e1->type);
	break;
      case '(':
      case '[':
	dir_abs_decl(e1, t);
	break;
      case ID:
	error("extraneous identifier `%s'",token.val.sym->id);
	break;
    }
    ABSDECL--;
}

void f_abs_decl(token_t * e1)
{
    switch (lexan()) {
      case '(':
      case '[':
	dir_abs_decl(e1, token.tok);
	break;
      case ID:
	error("extraneous identifier `%s'",token.val.sym->id);
	break;
      default:
	retractlexan();
    }
}

void dir_abs_decl(token_t * e1, int t)
{
    switch (t) {
      case '(':
	f1_dir_abs(e1);
	f2_dir_abs(e1);
	break;
      case '[':
	array_decl(e1);
	f2_dir_abs(e1);
	break;

    }
}

void f1_dir_abs(token_t * e1)
{
    switch (lexan()) {
      case ')':
      TYPEQUAL:
      STORECLASS:
      TYPESPEC:
	retractlexan();
	ff_dir_decl(e1);
	remlevel(S_LEVEL+1);
	break;
      case '*':
      case '(':
      case '[':
	abs_decl(e1, token.tok);
	match(')', " )");
	break;

      default:
	error("Abstract declaration error");
    }
}

void f2_dir_abs(token_t * e1)
{
    switch (lexan()) {
      case '[':
	array_decl(e1);
	f2_dir_abs(e1);
	break;
      case '(':
	ff_dir_decl(e1);
	remlevel(S_LEVEL+1);
	f2_dir_abs(e1);
	break;
      default:
	retractlexan();
    }
}
