/* eicmod.c
 *
 *	(C) Copyright Apr 15 1995, Edmond J. Breen.
 *		   ALL RIGHTS RESERVED.
 * This code may be copied for personal, non-profit use only.
 *
 */



#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include "ctype.h"


#include "typemod.h"
#include "lexer.h"
#include "global.h"

#include "func.h"
#include "xalloc.h"
#include "typesets.h"
#include "preproc.h"
#include "cdecl.h"
#include "error.h"
#include "emitter.h"
#include "symbol.h"

extern char LEXEM[];
extern token_t token;

int listcodeON = 0;
int lineNums = 0;
int interpON = 1;
int TIMER = 0;
int SHOWLINE = 0;
int memdumpON = 0;
int memtraceON = 0;
int traceON = 0;
int traceFunc = 0;
int verboseON = 0;
int interfaceON=0;
int interActive=1;
int showIncludes = 0;

char *fftoa(double, char *, int, char, int);
char * itoa(int,char *s, int);
char * utoa(unsigned,char *s, int);
char * ltoa(long,char *s, int);
char * ultoa(unsigned long,char *s, int);


void outputval(int obj, void *v)
{
#if 0
    char str[512], *s;
    s = str;

    switch (obj) {
      case t_char:   s = itoa(*(char *) v, str, 10); break;
      case t_uchar:  s = itoa(*(unsigned char *) v, str, 10); break;
      case t_short:  s = itoa(*(short *) v, str, 10); break;
      case t_ushort: s = utoa(*(unsigned short*) v, str, 10); break;
      case t_int:    s = itoa(*(int *) v, str, 10); break;
      case t_uint:   s = utoa(*(unsigned *) v, str, 10); break;
      case t_long:   s = ltoa(*(long *) v, str, 10); break;
      case t_ulong:  s = ultoa(*(long *) v, str, 10); break;
      case t_float:  s = fftoa(*(float *) v, str, 6, 'f', 1); break;
      case t_double: s = fftoa(*(double *) v, str, 6, 'f', 1); break;
      case t_lval:   
      case t_pointer:s = str; str[0] = '0';
	                 str[1] = 'x'; ultoa(*(long *) v, &str[2], 16); break;
      case t_void:   s = "(void)"; break;
      case t_enum:   s = "(enumeration)"; break; 
      case STR:      s = *(char **) v; break;
      default:       s = "Unknown type\n"; break;
    }
    if(s)
	fprintf(stdout,"%.100s",s);
    else
	fputs("(null)",stdout);
#else
    switch (obj) {
      case t_char:   fprintf(stdout,"%d",*(char *) v);break;
      case t_uchar:  fprintf(stdout,"%d",*(unsigned char *) v); break;
      case t_short:  fprintf(stdout,"%d",*(short *) v); break;
      case t_ushort: fprintf(stdout,"%d",*(unsigned short*) v); break;
      case t_int:    fprintf(stdout,"%d",*(int *) v); break;
      case t_uint:   fprintf(stdout,"%u",*(unsigned *) v); break;
      case t_long:   fprintf(stdout,"%ld",*(long *) v); break;
      case t_ulong:  fprintf(stdout,"%ld",*(long *) v); break;
      case t_float:  fprintf(stdout,"%g",*(float *) v); break;
      case t_double: fprintf(stdout,"%g",*(double *) v); break;
      case t_lval:   
      case t_pointer:fprintf(stdout,"%p",(void*)*(long *) v); break;
      case t_void:   fprintf(stdout,"(void)"); break;
      case t_enum:   fprintf(stdout,"(enumeration)"); break; 
      case STR:      fprintf(stdout,"%100s",*(char **) v); break;
      default:       fprintf(stdout,"Unknown type"); break;
    }

#endif
}


void showstruct(AR_t * arg)
{
    int i, t;
    struct_t *S;
    char *D;
    S = (struct_t *) getInf(arg->type);
    D = arg->v.p.p;
    fputc('{', stdout);
    if (!D)
	fputs("(null struct) ", stdout);
    else
	for (i = 0; i < S->n; i++) {
	    switch ((t = gettype(S->type[i]))) {
	      case t_array:
		fputs("Array", stdout);
		break;
	      case t_union:
		fputs("Union",stdout);
		break;
	      case t_struct:
		fputs("Struct", stdout);
		break;
	      default:
		outputval(t, &D[S->offset[i]]);
		break;
	    }
	    fputc(',', stdout);
	}
    fputs("\b}", stdout);
}

void showvalue(AR_t * arg)
{
    int t;

    if(!interActive)
	return ;

    fputc('\t', stdout);

    t = gettype(arg->type);

    if(isconst(arg->type) && t != t_pointer) {
	outputval(t, &arg->v);
	fputc('\n', stdout);
	return;
    }
    	
    switch ((t = gettype(arg->type))) {
      case t_union:  fputs("(Union)",stdout); break;
      case t_struct: showstruct(arg); break;
      case t_char:   case t_uchar:  case t_short:  
      case t_ushort: outputval(t_int, &arg->v.ival); break;
      case t_float:  outputval(t_double, &arg->v.ival); break;
      case t_array:
      case t_pointer:
	if (gettype(nextType(arg->type)) == t_char)
	    t = STR;
	else
	    t = t_pointer;
      default:
	outputval(t, &arg->v);
	break;
    }
    fputc('\n', stdout);

}

/*********************
void outFunc(token_t *e1)
{
    void showrevtype(type_expr * t, int expand, FILE *fp);
    FILE *fp;
    type_expr * ty = e1->val.sym->type;
    int t = gettype(ty);
    
    if(t == t_funcdec || t == t_func) {
	char buf[100] = "val_t eic_";
	int i;
	func_t * F = (func_t *) getInf(e1->val.sym->type);
	char *s = NULL;
	strcat(buf,e1->val.sym->id);
	fp = stdout;
	fprintf(fp,"%s(void)\n{\n",buf);
	fputs("\tval_t v;\n",fp);
	switch (gettype(nextType(t))) {
	  case t_char:  
	  case t_uchar:  s = "v.cval = "; break;
	  case t_short: 
	  case t_ushort: s = "v.sval = "; break;
	  case t_int:
	  case t_uint:   s = "v.ival = "; break;
	  case t_long:  
	  case t_ulong:  s = "v.lval = "; break;
	  case t_float:  s = "v.fval = "; break;
	  case t_double: s = "v.dval = "; break;
	  case t_pointer: s = "v.pval = "; break;
	  case t_void:    s = "";
	}
	fprintf(fp,"\t%s%s(",s,e1->val.sym->id);
	if (F && F->n) {
	    for (i = 0; i < F->n; i++) {
		fprintf(fp,"arg(%d,getargs(),",i);
		showrevtype(F->type[i], 0,fp);
		if(i+1 != F->n)
		    fputs("),\n\t\t\t", fp);
		else
		    fputs("));\n",fp);
	    }
	}
	fputs("\treturn v;\n}\n",fp);
    }
}
****************/

void showFunc(type_expr **t, FILE *fp)
{
    static int level = 0;
    void showdectype(type_expr * t, int expand, FILE *fp);
    func_t *F;
    int i,k;
    fputs("Func (", fp);
    F = (func_t *) getInf(*t);
    level++;

    if (F && getFNp(F)) {
	for (i = 0; i < getFNp(F); i++) {
	    fputc('\n',fp);
	    for(k=0;k<level;k++)
		fputc('\t',fp);
	    if(getFPname(F,i))
		printf("%s: ",getFPname(F,i));
	    showdectype(getFPty(F,i), 0,fp);
	    if(i < getFNp(F)-1)
		fputs(",", fp);
	}

    }
    fputc('\n',fp);
    for(k=0;k<level;k++)
	fputc('\t',fp);
    fputs(") returning ", fp);
    if(nextType(*t)) {
	*t = nextType(*t);
	showdectype(*t,0,fp);
	if(*t)
	    while(nextType(*t))
		*t = nextType(*t);
    }
    
    if(F && getFComm(F)) 
	fprintf(fp,"\n\n    /* %s/\n", getFComm(F));  /**/
    level--;
    
}    

void showdectype(type_expr * t, int expand, FILE *fp)
{

    struct_t *S;
    int i;
    static int level = 0;

    while (t) {
	if(isconst(t) || isconstp(t))
	    fputs("const ",fp);
	
	switch (gettype(t)) {
	  case t_var:   fputs("...", fp); break;
	  case t_enum:  fputs("enum ", fp); break;
	  case t_char:  fputs("char ", fp); break;
	  case t_uchar: fputs("unsigned char ", fp); break;
	  case t_short: fputs("short ", fp); break;
	  case t_ushort:fputs("unsigned short ", fp); break;
	  case t_int:   fputs("int ", fp); break;
	  case t_uint:  fputs("unsigned ", fp); break;
	  case t_long:  fputs("long int ", fp); break;
	  case t_ulong: fputs("unsigned long int ", fp); break;
	  case t_float: fputs("float ", fp); break;
	  case t_double:fputs("double ", fp); break;
	  case t_pointer: fputs("* ", fp); break;
	  case t_void:  fputs("void ", fp); break;
	  case t_hidden: fputs("hidden ",fp); break;
	    
	  case t_funcdec: fputs("dec_", fp); showFunc(&t,fp); break;
	  case t_func:   showFunc(&t,fp); break;
	  case t_builtin: fputs("Builtin ",fp); showFunc(&t,fp); break; 

	  case t_array:
	    fprintf(fp,"ARY[%d]",(int) getInf(t));
	    break;
	  case t_union:
	  case t_struct:
	    S = getInf(t);
	    if(!S) {
		fputs("Incomplete", fp);
		break;
	    }
	    if (gettype(t) == t_struct)
		fprintf(fp,"struct: size  %u bytes",S->tsize);
	    else
		fprintf(fp,"union: size  %u bytes",S->tsize);
	    if (expand) {
		level++;
		fputc('\n', fp);
		if (level <= 2) {
		    int j;
		    for (i = 0; i < S->n; i++) {
			for (j = 0; j < level; j++)
			    fputc('\t', fp);
			fputs(S->id[i], fp);
			fputs(" -> ", fp);
			showdectype(S->type[i], expand,fp);
			fputc('\n', fp);
		    }
		}
		level--;
	    }
	    break;
	  case ID: fputs("Identifier ", fp); break;
	  default: fputs("Uknown identifier", fp); return;
	}
	t = nextType(t);
    }
}

int IsObject(int t)
{
    if(t != t_func && t !=t_funcdec && t != t_builtin)
	return 1;
    return 0;
}


int IsFunc(int t)
{
    return !IsObject(t);
}

void displaySclass(symentry_t *sym)
{
    /* display storage class */
    if (sym->sclass == c_typedef)
	fputs("typedef ", stdout);
    else if(sym->sclass & c_private)
	fputs("private ", stdout);
    else if (sym->sclass & c_static)
	fputs("static ", stdout);
    if (sym->typequal == q_const)
	fputs("const ", stdout);
    else if (sym->typequal == q_volatile)
	fputs("volatile ", stdout);
}    

void showtab(int tab, int expand, token_t * e1, char *fname, int allowed())
{
    int i, t;
    symentry_t *sym;
    
    for (i = 0; i < HSIZE; i++)
	for (sym = HTAB[i]; sym; sym = sym->next) {

	    if ((t = gettype(sym->type)) == t_eic || sym->nspace != tab)
		continue;

	    if(e1 && e1->type && ! sametypes(e1->type,sym->type))
		continue;

	    if(fname && strcmp(fname,sym->fname) != 0)
		continue;
	    
	    if(allowed && !allowed(t))
		continue;
	    displaySclass(sym);
	    fputs(sym->id, stdout);
	    fputs(" -> ", stdout);
	    showdectype(sym->type, expand,stdout);
	    fputc('\n', stdout);
	}
}


int show(void)
{
    int t;
    extern token_t token;
    int h = work_tab;
    work_tab = stand_tab;
    if ((t = lexan()) == ID || t == TYPENAME) {
	if(!showMacro(token.val.sym->id)) { /* test macros first */
	    t = gettype(token.val.sym->type);
	    if (t == ID) {
		remsym(token.val.sym);
		/*
		 * check for struct/union tag name.
		 */
		if ((token.val.sym = lookup(tag_tab, LEXEM)) == NULL) {
		    error("Unknown identifier %s",LEXEM);
		    return 0;
		}
		fputs("Tag::",stdout);
	    }

	    displaySclass(token.val.sym);
	    fputs(token.val.sym->id, stdout);
	    fputs(" -> ", stdout);
	    showdectype(token.val.sym->type, 1,stdout);
	    fputc('\n',stdout);
	    if (t == t_func && listcodeON) {
		if (ENV->AR[token.val.sym->val.ival].v.p.p)
		    listcode(ENV->AR[token.val.sym->val.ival].v.p.p);
	    }
	} else
	    remsym(token.val.sym);
    } else
	error("Expected identifier");
    work_tab = h;
    return 1;
}

int filesCommand(void)
{
    void showIncludedFiles(void);
    void showFileMacros(char *fname);
    int nextchar(void);

    int IsIncluded(char *fname);
    int c;
    char *p, buf[256];
    
    while(isspace(c = nextchar()));
    p = buf;
    while( c && c != DONE && !isspace(c)) {
	*p++ = c;
	c = nextchar();
    }
    *p = '\0';
    if(p != buf) {
	if(!IsIncluded(buf))
	     warningerror("Failed to find %s",buf); 
	else {
	    showFileMacros(buf);
	    showtab(stand_tab,0,NULL,buf,IsObject);
	    showtab(stand_tab,0,NULL,buf,IsFunc);
	}
    } else 
	showIncludedFiles();
    
    return 1;
}

int clearCommand(void)
{
    void lutClearFileEntries(char *FileName);
    void ClearFileMacros(char *fname);
    void rmIncludeFileName(char *fname);
    
    int nextchar(void);

    int IsIncluded(char *fname);
    int c;
    char *p, buf[256];
    
    while(isspace(c = nextchar()));
    p = buf;
    while( c && c != DONE && !isspace(c)) {
	*p++ = c;
	c = nextchar();
    }
    *p = '\0';
    if(p != buf) {
	if(!IsIncluded(buf))
	     warningerror("Failed to find %s",buf); 
	else {
	    lutClearFileEntries(buf);
	    ClearFileMacros(buf);
	    rmIncludeFileName(buf);
	}
    } else 
	error("Expected file name");
        
    return 1;
}

int setnextinf(void *info)
{
    int i, t;
    symentry_t *sym;
    type_expr *type;
    for (i = 0; i < HSIZE; i++)
	for (sym = HTAB[i]; sym; sym = sym->next)
	    if ((t = gettype(sym->type)) == t_union || t == t_struct)
		for (type = sym->type; type; type = nextType(type))
		    if (type->alias && getInf(type) == info) {
			type->alias = 0;
			return 1;
		    }
    return 0;
}

void checktype(type_expr * type)
{
    int t;
    while (type) {
	if ((t = gettype(type)) == t_struct || t == t_union)
	    if (!type->alias)
		if (setnextinf(getInf(type)))
		    type->alias = 1;
	type = nextType(type);
    }
}


long getmitem(val_t *v, int t)
{
    switch(t) {
      case INT: return v->ival;
      case UINT: return v->uival;	
      case LONG: return v->lval;
      case ULONG: return v->ulval;
    }
    return -1;      
}

int rm(void)
{
    unsigned long mitem;
    
    int t, h = work_tab;
    work_tab = stand_tab;
    do {
	switch(lexan()) {
	  case ID:
	  case TYPENAME:
	    t = gettype(token.val.sym->type);
	    if (t == ID) {
		remsym(token.val.sym);
		/*
		 * check for a tag name.
		 */
		if((token.val.sym = lookup(tag_tab,LEXEM)) == NULL) {
		    error("Unknown identifier %s",LEXEM);
		}else {
		    checktype(token.val.sym->type);
		    remsym(token.val.sym);
		}
	    } else if (t == t_builtin) {
		error("Cannot remove builtin functions");
	    } else {
		checktype(token.val.sym->type);
		remsym(token.val.sym);
	    }
	    break;
	  case INT: 
	  case UINT:
	  case LONG:
	  case ULONG:	    
	    mitem = getmitem(&token.val,token.tok);
	    if(getMemMark(mitem) == MEM_LEAK)
		freeMemItem(mitem);
	    else
		error("Memory item %ld is not a leak\n", mitem);
	    break;
	  default:
	    error("UnExpected input to rm");
	    break;
	    
	}
    } while (lexan() == ',');
    retractlexan();
    work_tab = h;
    return 1;
}


int showhashtab(void)
{
    int tab;
    token_t e1;

    work_tab = tab  = stand_tab;   /* standard name space */
    inittoken(&e1);

    switch(lexan()) {
      TYPESPEC:
      TYPEQUAL:
	retractlexan();
	type_name(&e1);
	break;
      case ID:
	if(strcmp(LEXEM,"tags") == 0) 
	    tab = tag_tab;
	else if(strcmp(LEXEM,"builtin") == 0)
		e1.type = addtype(t_builtin,0);
	else
	    retractlexan();
	break;		
      default:
	retractlexan();
    }
    if(e1.type) {
	showtab(tab, 0,&e1,NULL,NULL);
	freetype(e1.type);
    } else {
	showtab(tab,0,NULL,NULL,IsObject);
	showtab(tab,0,NULL,NULL,IsFunc);
    }
    return 1;
}

int extra_option(char * option)
{
    work_tab = stand_tab;   /* standard name space */
    if (lexan() == ID) {
	if(strcmp(LEXEM,option) == 0)
	    return 1;
	else
	    retractlexan();
    } else
	retractlexan();
    return 0;
}

int exit_EiC()
{
    extern FILE *EiChist;
    fflush(NULL);
    if(verboseON)
	fputs("EiC Bye\n", stdout);
    if(EiChist)
	fclose(EiChist);
    exit(1);
    return 1;
}

int toggle_verbose()
{
    verboseON = !verboseON;
    return verboseON;
}
int toggle_interface()
{
    interfaceON = !interfaceON;
    return interfaceON;
}


int toggle_memdump()
{
    memdumpON = !memdumpON;
    return memdumpON;
}

int toggle_memtrace()
{
    memtraceON = !memtraceON;
    return memtraceON;
}

int toggle_showline()
{
    SHOWLINE = !SHOWLINE;
    return SHOWLINE;
}


int toggle_timer()
{
    TIMER = !TIMER;
    return TIMER;
}

int toggle_interp()
{
    interpON = !interpON;
    return interpON;
}

int toggle_includes()
{
    showIncludes= !showIncludes;
    return showIncludes;
}

int toggle_trace()
{
    if(extra_option("funcs"))
	traceFunc = !traceFunc;
    else
	traceFunc = 0;

    if(traceFunc)
	traceON = 1;
    else
	traceON = !traceON;

    return traceON;
}

int toggle_listcode()
{
    if(extra_option("linenums"))
	lineNums = !lineNums;
    else
	lineNums = 0;

    if(lineNums)
	listcodeON = 1;
    else
	listcodeON = !listcodeON;

    return listcodeON;
}


int showhistory()
{
    void show_history(FILE *fp);
    show_history(stdout);
    return 1;
}

int showHelp()
{
    puts("-----------------------------------------------------------------------------\n"
	 "EiC-COMMAND          SUMMARY DESCRIPTION\n"
	 "-----------------------------------------------------------------------------\n"
	 ":-I path       Append path to the include-file search list.\n"
	 ":-L            List search paths.\n"
	 ":-R path       Remove path from the include-file search list.\n"
	 ":clear fname   Removes the contents of file fname from EiC.\n"
	 ":exit          Terminates an EiC session.\n"
	 ":files         Display the names of all included files.\n"
	 ":files fname   Summarize the contents of the included file `fname'.\n" 
	 ":help          Display summary of EiC commands.\n"
	 ":history       List the history of all input commands.\n"
	 ":includes      Display path of include files when loaded.\n"
	 ":interpreter   Execute input commands. By default it is on.\n"
	 ":listcode      List stack code.\n"
	 ":listcode linenums     List stack code with associated line numbers.\n"
	 ":memdump       Show potential memory leaks.\n"
	 ":rm  dddd      Remove memory item dddd, which is a constant integer value.\n"
	 ":rm   f        Removes f's definition from the symbol tables.\n"
	 ":show f        Shows type or  macro definition of `f'.\n"
	 ":showline      Show input line after macro expansion.\n"
	 ":status        Display the status of the toggle switches.\n"
	 ":timer         Time in seconds of execution.\n"
	 ":trace         Trace function calls and line numbers during code execution.\n"
	 ":trace funcs   Trace function calls only during code execution.\n"
	 ":variables     Display declared variables and interpreter-ed function names.\n"
         ":variables tags       Display the tag identifiers.\n"
         ":variables type-name  Display variables of type `type-name'.\n"
	 ":verbose       Suppresses EiC's copyright and warning messages on start up.\n"
	 "------------------------------------------------------------------------------\n"
	 );
    return 0;
}


int EiCstatus()
{
    struct toggle {
	char * name;
	int *val; }
    toggles[] =  {
        {"includes",&showIncludes},
	{"interpreter", &interpON},
	{"listcode",&listcodeON},
	{"memdump", &memdumpON},
	{"showline",&SHOWLINE},
	{"timer",&TIMER},
	{"trace",&traceON},
	{"verbose",&verboseON},
    };
    int n = sizeof(toggles)/sizeof(struct toggle);

    while(n--)
	printf("\t%-15s %d\n",toggles[n].name, *toggles[n].val);

    return 0;
}


symentry_t *add_eicfunc(char *name, int (*func) ())
{
    symentry_t *sym;
    sym = insertLUT(eic_tab, name, t_eic);
    sym->level = 1;
    sym->val.func = func;
    return sym;
}

symentry_t *add_builtinfunc(char *name, val_t(*vfunc) ())
			
{
    symentry_t *sym;
    sym = insertLUT(stand_tab, name, ID);
    freetype(sym->type);
    sym->level = 1;
    sym->val.vfunc = vfunc;
    sym->type = addtype(t_builtin,NULL);
    return sym;
}


void comm_switch(void)
{
    int insertpath(char *);
    int removepath(char *);
    void listpath(void);
    int nextchar(void);
    
    int c;
    char buf[256], *p;
    extern token_t token;
    if (lexan() == ID) {
	if(gettype(token.val.sym->type) == ID) {
	    remsym(token.val.sym);
	}
	/* collect switch and string */ 
	while(isspace(c = nextchar()));
	p = buf;
	while( c && c != DONE && !isspace(c)) {
	    *p++ = c;
	    c = nextchar();
	}
	*p = '\0';
	    
	switch(LEXEM[0]) {
	  case 'I':
	    if(!*buf || !insertpath(buf))
	     warningerror("Failed to add path"); 
	    break;
	  case 'R':
	    if(!*buf || !removepath(buf))
	     warningerror("Failed to remove path"); 
	    break;
	  case 'L':
	    listpath();break;
	  default:	    
	    error("Unknown command switch");
	}	    	    
	    
    } else 
	error("Expected command switch");

}

/**************
symentry_t *add_builtinfunc(char *name,
			    val_t(*vfunc) (),
			    type_expr * t)
{
    symentry_t *sym;
    sym = insertLUT(stand_tab, name, ID);
    freetype(sym->type);
    sym->level = 1;
    sym->val.vfunc = vfunc;
    sym->type = addtype(t_builtin, t);
    return sym;
}
******************/


void establish_ARs(void)
{
    ENV->LARsize = 1000;
    ENV->ARsize =  500;
    ENV->LAR =(AR_t*)xmalloc(sizeof(AR_t)*ENV->LARsize);
    ENV->AR = (AR_t*)xmalloc(sizeof(AR_t)*ENV->ARsize);
}


void init_EiC(void)
{
    ENV = (environ_t *) xcalloc(1, sizeof(environ_t));
    ENV->CODE.Filename = "::EiC::";
    ENV->level = 1;
    establish_ARs();
    initpp();
    work_tab = stand_tab;
    /*init_hashtab(31);*/
    /* add EiC command line functions */
    add_eicfunc("listcode", toggle_listcode);
    add_eicfunc("interpreter", toggle_interp);
    add_eicfunc("variables", showhashtab);
    add_eicfunc("exit", exit_EiC);
    add_eicfunc("timer", toggle_timer);
    add_eicfunc("showline", toggle_showline);
    add_eicfunc("show", show);
    add_eicfunc("memdump", toggle_memdump);
    add_eicfunc("memtrace", toggle_memtrace);
    add_eicfunc("rm", rm);
    add_eicfunc("trace",toggle_trace);
    add_eicfunc("verbose",toggle_verbose);
    add_eicfunc("interface",toggle_interface);
    add_eicfunc("history",showhistory);
    add_eicfunc("help", showHelp);
    add_eicfunc("status",EiCstatus);
    add_eicfunc("includes",toggle_includes);
    add_eicfunc("files",filesCommand);
    add_eicfunc("clear",clearCommand);
    
}









