/* error.c
 *
 *	(C) Copyright Apr 15 1995, Edmond J. Breen.
 *		   ALL RIGHTS RESERVED.
 * This code may be copied for personal, non-profit use only.
 *
 */

#include "stdio.h"
#include "stdlib.h"

#include "lexer.h"
#include "global.h"
#include "xalloc.h"
#include "preproc.h"
#include "error.h"

#include <stdarg.h>

extern char *lex_buff;
int ParseError, ErrorRecover;

int EiC_errs = 0;


               /*CUT match.cut*/
int match(int t, char *m)
{
    extern token_t token;
    int lookahead;
    char message[80];

    lookahead = lexan();
    if (ErrorRecover) {		/* error recovery method */
	while ((lookahead != t && lookahead != ';') && lookahead != DONE) {
	    if(lookahead == STR) 
		xfree(token.val.p.p);
	    lookahead = lexan();
	}
	if (lookahead == t || lookahead == ';') {
	    ErrorRecover = 0;
	    if (lookahead != t)
		retractlexan();
	    return 1;
	} else
	    return 0;
    }
    if (lookahead != t) {
	sprintf(message, "Expected %s", m);
	error(message);
	return 0;
    }
    return 1;
}
              /*END CUT*/


void clear_err_msgs(void)
{
    EiC_errs = 0;
}

               /*CUT error.cut*/
void error(char *msg, ...)
{
    va_list args;
    
    va_start(args,msg);
    if (!ErrorRecover) {
	EiC_errs++;
	ErrorRecover = ParseError = 1;
	fprintf(stderr,"Error in %s near line %d: ",
		CurrentFileName(),
		CurrentLineNo());
	vfprintf(stderr,msg,args);
        fprintf(stderr,"\n");
    }
    va_end(args);
}
              /*END CUT*/


               /*CUT warningerror.cut*/
void warningerror(char *msg, ...)
{
    va_list args;
    va_start(args,msg);
    EiC_errs++;
    fprintf(stderr, "Warning: in %s near line %d: ",
	    CurrentFileName(),
	    CurrentLineNo());
    vfprintf(stderr,msg,args);
    fprintf(stderr,"\n");
    va_end(args);
}
              /*END CUT*/














