/* parser.c
 *
 *	(C) Copyright Apr 15 1995, Edmond J. Breen.
 *		   ALL RIGHTS RESERVED.
 * This code may be copied for personal, non-profit use only.
 *
 */
#include "stdio.h"
#include "string.h"
#include "stdlib.h"
#include "limits.h"

#include "MachSet.h"
#include "lexer.h"
#include "global.h"
#include "func.h"
#include "typemod.h"
#include "xalloc.h"
#include "typesets.h"
#include "cdecl.h"
#include "error.h"
#include "symbol.h"


extern token_t token;
extern char LEXEM[];
typedef struct {
    val_t match, loc;
} pair_t;

#define out_expr  Outexpr
#define setInst(code,idx,mem,val)  do { code_t *C__C = code;\
					    if(C__C->binst) \
					C__C->inst[idx].mem = val;\
				    } while(0)


void reset_env_pointers(token_t *e1,int bp);

/** PROTOTYPES parser.c **/
void addoffset(eicstack_t * S, int nxt, int offset);
void correctit(code_t * C, eicstack_t * S, int nxt, int loc);
int comppair(pair_t * p1, pair_t * p2);
void addcaseoffset(int offset);
void fixcasestack(token_t * e1, int nxt, int loc);
void initparser();
void parse(environ_t * env);
void block(token_t * e1);
void addinlocals(token_t * e1,int n);
void stmt(token_t * e1);
void label_stmt(token_t * e1);
void testfortype(token_t *e1, int NoVoid);
void genJump(token_t *e1, val_t *v, int t,int T);
void select_stmt(token_t * e1);
void iter_stmt(token_t * e1);
void jump_stmt(token_t * e1);
void Outexpr(token_t * e1);
void expr(token_t * e1);
void assign_expr(token_t * e1);
void cond_expr(token_t * e1);
void log_or_expr(token_t * e1);
void log_and_expr(token_t * e1);
void inc_or_expr(token_t * e1);
void xor_expr(token_t * e1);
void and_expr(token_t * e1);
void equal_expr(token_t * e1);
void rel_expr(token_t * e1);
void shift_expr(token_t * e1);
void add_expr(token_t * e1);
void mult_expr(token_t * e1);
void cast_expr(token_t * e1);
void f_cast_expr(token_t * e1);
void unary_expr(token_t * e1);
void postfix_expr(token_t * e1);
void r_postfix_expr(token_t * e1);
int arg_expr_list(token_t * E1, token_t * e1);
void primary_expr(token_t * e1);
int GI(token_t * e1);
void process_binary(void (*func) (token_t * e),
		    token_t * e1, int op);
void assignop(void (*func) (token_t * e),
	      token_t * e1, int op);
int findmem(type_expr * t, char *id);
int check4constmem(type_expr * t);


int S_LEVEL;

int NoPTR = 0;

eicstack_t breakstack = {0, NULL};
eicstack_t contstack = {0, NULL};
eicstack_t funclinkstack = {0, NULL};
eicstack_t *casestack;

static int BREAK = 0, CONT = 0, CASEON;


    
void addoffset(eicstack_t * S, int nxt, int offset)
{
    unsigned i;
    val_t *v;
    for (i = nxt, v = &S->val[nxt]; i < S->n; ++v, ++i)
	v->ival += offset;
}


void correctit(code_t * C, eicstack_t * S, int nxt, int loc)
{
    val_t addr;
    if(!C->binst)
	return;
    while (S->n > nxt) {
	eicpop(S, &addr);
	C->inst[addr.ival].val.ival =  loc-addr.ival;
    }
}

int comppair(pair_t * p1,pair_t * p2)
{
    return p1->match.ival - p2->match.ival;
}

void addcaseoffset(int offset)
{
    unsigned i;
    val_t *v;
    v = casestack->val;
    if (v->ival != 0)		/* check for default */
	v->ival += offset;
    for (i = 2; i < casestack->n; i += 2)
	v[i].ival += offset;
}


void fixcasestack(token_t * e1, int nxt, int loc)
{
    void qsort(void *base, size_t nm, size_t sz,
	       int (*f)());
    int i;
    if (casestack->n > 0) {
	correctit(&e1->code, &breakstack, nxt, e1->code.nextinst);
	if (casestack->val[0].ival == 0)
	    casestack->val[0].ival = e1->code.nextinst - loc;
	else
	    casestack->val[0].ival -= loc;
	for (i = 2; i < casestack->n; i += 2)
	    casestack->val[i].ival -= loc;
	qsort(&casestack->val[1],
	      (casestack->n - 1) >> 1,
	      sizeof(val_t) << 1,
	      comppair);
    } else
	error("Illegal switch statement");
}

void initparser()
{
    extern int ErrorRecover;
    work_tab = stand_tab;
    ErrorRecover = 0;
}

void parse(environ_t * env)
{
    void comm_switch(void);
    token_t e1;

    val_t v;
    int t;
    S_LEVEL = 1;
    inittoken(&e1);
    while ((t = lexan()) != DONE) {
	freetoken(&e1);
	
	if (t == ':') {		/* get an EiC command */
	    int h;
	    h = work_tab;
	    work_tab = eic_tab;
	    if (lexan() == ID) {
		if (gettype(token.val.sym->type) == t_eic)
		    (*token.val.sym->val.func) ();
		else {
		    remsym(token.val.sym);
		    error("Unknown EiC command");
		}
	    } else if(token.tok == '-')
		comm_switch();
	    else
		error("Expected an EiC command");
	    work_tab = h;
	    continue;
	} else if(t == '?') {
	  printf("\t If you want help, enter :help\n\n");
	  continue;
	}
	  
	retractlexan();

#if 1
	switch(t) {
	  TYPEQUAL:
	  STORECLASS:
	  TYPESPEC:
	    ext_decl(&e1);
	    break;
	  default:
	    stmt(&e1);
	}
#else
	block(&e1);
#endif
	concode(&env->CODE, &e1.code);
    }
    concode(&env->CODE, &e1.code);

    if (e1.type)
	v.p.p = copytype(e1.type);
    else
	v.p.p = addtype(t_void, NULL);

    freetoken(&e1);
    generate(&env->CODE, halt, &v, 0);
}
/*
 * Globals used for
 * local aggregate data
 */

unsigned int ASPOT = 0, MASPOT = 0, ADDLOCALS = 0;
    
void updateLocals(void)
{    
    if (ASPOT > MASPOT)
	MASPOT = ASPOT;
    if (ENV->lsp > ADDLOCALS)
	ADDLOCALS = ENV->lsp;
}

void addinlocals(token_t * e1,int n)
{
    val_t v;
    token_t e2;
    
    inittoken(&e2);
    v.ival = n + (MASPOT>0);
    generate(&e2.code, checkar, &v, 0);
    setCodeLineNo(&e2.code,e2.code.nextinst-1,0);
    if (MASPOT) {
	v.ival = MASPOT;
	generate(&e2.code, pushint, &v, 0);
	setCodeLineNo(&e2.code,e2.code.nextinst-1,0);
	v.ival = n;

	{ /* correct for AAA location */
	    int i, N;
	    code_t *c = &e1->code;
	    N = nextinst(c);
	    for(i = 0; i < N; ++i)
		if(opcode(c,i) == lda && ivalcode(c,i) == -1)
		    ivalcode(c,i) = n;
	}
	generate(&e2.code, massign, &v, 1);
	setCodeLineNo(&e2.code,e2.code.nextinst-1,0);
	generate(&e1->code, fmem, &v, 1);
	setCodeLineNo(&e1->code,e1->code.nextinst-1,0);
	v.ival = n;
    }
    generate(&e1->code, reducear, &v, 0);
    setCodeLineNo(&e1->code,e1->code.nextinst-1,0);

    concode(&e2.code, &e1->code);
    e1->code = e2.code;
    ADDLOCALS = ASPOT = MASPOT = 0;
}

void block(token_t * e1)
{
    token_t e2;
    inittoken(&e2);
    while(ext_decl(&e2)) {
	contoken(e1,&e2);
	inittoken(&e2);
    }
    while (lexan() != '}' && token.tok != DONE) {
	retractlexan();
	stmt(e1);
    }
    if (token.tok == DONE)
	error("Expected }");
}

void stmt1(token_t * e1)
{
    void clearTempories();
    int bp;
    unsigned aspot;

    switch (lexan()) {
      case '{':			/* compound statement */
	S_LEVEL++;
	bp = ENV->lsp;

	aspot = ASPOT;

	block(e1);

	updateLocals();
	clearTempories();
	
	ASPOT = aspot;

	reset_env_pointers(e1, bp);
	remlevel(S_LEVEL);
	S_LEVEL--;
	break;
      case casesym:
      case defaultsym:
	label_stmt(e1);
	break;
      case breaksym:
      case continuesym:
      case returnsym:
	jump_stmt(e1);
	break;
      case ifsym:
      case switchsym:
	select_stmt(e1);
	break;
      case whilesym:
      case dosym:
      case forsym:
	iter_stmt(e1);
	break;
      TYPEQUAL:
      STORECLASS:
      TYPESPEC:	
        error("Unexpected declaration");
	retractlexan();
	ext_decl(e1);
	break;
      default:
	retractlexan();
	out_expr(e1);
	match(';', " ; ");
	return;
    }
    /* to get here a statement was executed, therefore */
    e1->type = addtype(t_void, e1->type);    
    clearTempories();
}

void stmt(token_t *e1)
{
    static int Stmts = 0;
    Stmts++;
    
    stmt1(e1);

    Stmts--;
    if (S_LEVEL == 1 && Stmts == 0) {
	updateLocals();
	if(ADDLOCALS > 0) 
	    addinlocals(e1,ADDLOCALS);
    }
}

void label_stmt(token_t * e1)
{
    if (!CASEON) {
	error("Illegal label");
	return;
    }
    if (token.tok == casesym) {
	token_t e2;
	int t;
	val_t v;

	inittoken(&e2);
	assign_expr(&e2);
	match(':', " :");
	if (isconst(e2.type)) {
	    t = gettype(e2.type);
	    if (t >= t_char && t <= t_uint) {
		for (t = 1; t < casestack->n; t += 2)
		    if (e2.val.ival == casestack->val[t].ival)
			error("Duplicate case in switch");
		eicpush(casestack, e2.val);
		v.ival = e1->code.nextinst;
		eicpush(casestack, v);
	    } else
		error("Must be integer type");
	} else
	    error("Must be const_expr type");
	stmt(e1);
	freetoken(&e2);
    } else {			/* default */
	if (casestack->val[0].ival != 0)
	    error("Duplicate default in switch");
	match(':', " :");
	casestack->val[0].ival = e1->code.nextinst;
	stmt(e1);
    }
}


void select_stmt(token_t * e1)
{
    int t;
    token_t e2;
    val_t u1;

    inittoken(&e2);
    if (token.tok == ifsym) {
	match('(', " ( ");
	out_expr(&e2);

	testfortype(&e2,1);
	t = gettype(e2.type);
	
	match(')', " ) ");
	contoken(e1, &e2);
	u1.ival = e1->code.nextinst;
	genJump(e1,&u1,t,0);
	stmt(e1);

	if (lexan() == elsesym) {
	    setInst(&e1->code,u1.ival,val.ival,e1->code.nextinst-u1.ival+1);
	    u1.ival = e1->code.nextinst;
	    generate(&e1->code, jmpu, &u1, 0);
	    setCodeLineNo(&e1->code,e1->code.nextinst-1,0);
	    stmt(e1);
	    setInst(&e1->code,u1.ival,val.ival,e1->code.nextinst-u1.ival);
	} else {
	    setInst(&e1->code,u1.ival,val.ival,e1->code.nextinst-u1.ival);
	    retractlexan();
	}

    } else {/* switchsym */
	int nxtbreak, loc;
	eicstack_t *hold;

	BREAK++;
	CASEON++;
	nxtbreak = breakstack.n;
	match('(', " (");
	out_expr(&e2);
	match(')', " )");
	hold = casestack;
	casestack = (eicstack_t *) xcalloc(1, sizeof(eicstack_t));
	u1.ival = 0;
	eicpush(casestack, u1);	/* push dummy for default */
	contoken(e1, &e2);
	u1.p.p = casestack;
	loc = e1->code.nextinst;
	generate(&e1->code, jmptab, &u1, 0);
	stmt(e1);
	fixcasestack(e1, nxtbreak, loc);
	casestack = hold;
	BREAK--;
	CASEON--;
    }
}

void testfortype(token_t *e1, int NoVoid)
{
    int t = gettype(e1->type);
    if(!isArithmetic(t) && t != t_pointer) {
	if(t == t_void) {
	    if(NoVoid)
		error("Void expression");
	}else 
	    warningerror("Possible non relational operation");
    }
}

void genJump(token_t *e1, val_t *v, int t,int T)
{
    /* 0 for false and 1 for True */
    static int tab[2][4] = {{ jmpFint,jmpFlng,jmpFdbl,jmpFptr},
			    { jmpTint,jmpTlng,jmpTdbl,jmpTptr}};
    switch(t) {
      case t_char: case t_uchar: case t_short:
      case t_int:  case t_uint:
	generate(&e1->code, tab[T][0], v, 0);
	break;
      case t_long: case t_ulong:
	generate(&e1->code, tab[T][1], v, 0);
	break;
      case t_float: case t_double:
	generate(&e1->code, tab[T][2], v, 0);
	break;
      case t_pointer:
	generate(&e1->code, tab[T][3], v, 0);
	break;
    }
}


void iter_stmt(token_t * e1)
{
    int t, rt,nxtbreak, nxtcont;
    val_t u1;
    token_t e2, e3;
    inittoken(&e2);
    BREAK++, CONT++;
    nxtbreak = breakstack.n;
    nxtcont = contstack.n;

    switch (token.tok) {
      case dosym:
	u1.ival = e1->code.nextinst;
	stmt(e1);
	if (lexan() != whilesym)
	    error("Missing while in do while statement");
	match('(', " (");
	correctit(&e1->code, &contstack, nxtcont, e1->code.nextinst);
	out_expr(&e2);
	match(')', ")");
	match(';', " ;");
	
	testfortype(&e2,1);
	rt = gettype(e2.type);

	contoken(e1, &e2);
	u1.ival = u1.ival - e1->code.nextinst;
	genJump(e1,&u1,rt,1);
	/*generate(&e1->code, jmpTint, &u1, 0);*/
	correctit(&e1->code, &breakstack, nxtbreak, e1->code.nextinst);
	break;
      case whilesym:
	u1.ival = e1->code.nextinst;
	generate(&e1->code, jmpu, &u1, 0);
	setCodeLineNo(&e1->code,e1->code.nextinst-1,0);
	match('(', " ( ");
	out_expr(&e2);		/* <expr> */

	testfortype(&e2,1);
	rt = gettype(e2.type);

	match(')', " ) ");
	stmt(e1);		/* <stmt> */
	setInst(&e1->code,u1.ival,val.ival,e1->code.nextinst-u1.ival);
	t = e1->code.nextinst;
	contoken(e1, &e2);
	u1.ival -= e1->code.nextinst - 1;
	genJump(e1,&u1,rt,1);
	/*generate(&e1->code, jmpTint, &u1, 0);*/
	setCodeLineNo(&e1->code,e1->code.nextinst-1,0);
	correctit(&e1->code, &breakstack, nxtbreak, e1->code.nextinst);
	correctit(&e1->code, &contstack, nxtcont, t);
	break;
      case forsym:
	inittoken(&e3);

	match('(', " ( ");
	out_expr(e1);		/* expr1 */
	match(';', " ; ");
	u1.ival = e1->code.nextinst;
	generate(&e1->code, jmpu, &u1, 0);
	out_expr(&e2);		/* expr2 */

	testfortype(&e2,0);
	rt = gettype(e2.type);
	
	match(';', " ; ");
	out_expr(&e3);		/* expr3 */
	match(')', " ; ");

	stmt(e1);
	correctit(&e1->code, &contstack,
		  nxtcont, e1->code.nextinst);
	
	contoken(e1, &e3);
	setInst(&e1->code,u1.ival,val.ival,e1->code.nextinst-u1.ival);
	contoken(e1, &e2);
	u1.ival -= (e1->code.nextinst - 1);
	if (rt ==  t_void)
	    generate(&e1->code, jmpu, &u1, 0);
	else
	    genJump(e1,&u1,rt,1);
	   /*generate(&e1->code, jmpTint, &u1, 0);*/
	
	setCodeLineNo(&e1->code,e1->code.nextinst -1, 0);
	
	correctit(&e1->code, &breakstack, nxtbreak,
		  e1->code.nextinst);
	break;
    }
    BREAK--, CONT--;
}

void jump_stmt(token_t * e1)
{
    extern int RETURNON;
    extern token_t *RETURNEXPR;
    val_t v;
    v.ival = e1->code.nextinst;
    switch (token.tok) {
      case breaksym:
	if (!BREAK)
	    error("Misplaced break statement");
	else
	    eicpush(&breakstack, v);
	generate(&e1->code, jmpu, &e1->val, 0);
	match(';', " ;");
	break;
      case continuesym:
	if (!CONT)
	    error("Misplaced continue statement");
	else
	    eicpush(&contstack, v);
	generate(&e1->code, jmpu, &e1->val, 0);
	match(';', " ;");
	break;
      case returnsym:
	if (!RETURNON)
	    error("Misplaced return statement");
	else {
	    token_t e2;
	    if (lexan() != ';') {
		retractlexan();
		inittoken(&e2);
		expr(&e2);
		/* catch dangling pointers */
		if(gettype(e2.type) == t_pointer && e2.sym) 
		    if((gettype(e2.sym->type) != t_pointer &&
		       gettype(e2.sym->type) != t_array)  && GI(&e2)) 
		       warningerror("Possible dangling pointer");
		match(';', " ;");
		if(gettype(nextType(RETURNEXPR->type)) == t_void &&
		   gettype(e2.type) != t_void)
		    error("Illegal return type, expected void");
		    
		if (isconst(e2.type)) {
		    castconst(&e2, RETURNEXPR, 1);
		    output(&e2);
		} else {
		    output(&e2);
		    castvar(&e2, RETURNEXPR, 0);
		}
		if(HasHiddenParm(RETURNEXPR->type)) {
		    val_t v;
		    v.ival = -1;
		    generate(&e1->code,rvalptr,&v,1);
		    v.ival = 1;
		    generate(&e1->code,bump,&v,0);
		    contoken(e1, &e2);
		    v.ival = get_sizeof(nextType(RETURNEXPR->type));
		    generate(&e1->code,refmem,&v,0);
		} else
		    contoken(e1, &e2);
	    } else
		if(gettype(nextType(RETURNEXPR->type)) != t_void)
		    warningerror("missing return value");
	    generate(&e1->code, eicreturn, &e1->val, 0);
	}
	break;
    }
}

void generatePtr(token_t * e1)
{
    if(!NoPTR && gettype(e1->type) == t_array && !e1->pflag) {
	exchtype(t_pointer, e1->type);
	if(!GI(e1)) { /* static of global variable */
	    setConst(e1->type);
	    e1->val.p = ENV->AR[e1->val.ival].v.p;
	}
    }
}

void expr(token_t * e1)  
{
    /* really a comma expression */
    token_t e2;
    int c = 0;
    do {
	inittoken(&e2);
	assign_expr(&e2);
	if(nextinst(&e1->code))
	    output(e1);
	concode(&e1->code,&e2.code);
	freetype(e1->type);	
	e1->type = copytype(e2.type);
	e1->pflag = e2.pflag;
	freetoken(&e2);
	c++;

    } while (lexan() == ',');
    retractlexan();
    e1->pflag = e2.pflag;
    e1->sym = e2.sym;
    e1->val = e2.val;
    if(c > 1)
	setConstp(e1->type);
}

void Outexpr(token_t * e1) 
{
    /* really a comma expression */
    token_t e2;
    int c = 0;
    do {
	inittoken(&e2);
	assign_expr(&e2);
	output(&e2);
	freetype(e1->type);
	concode(&e1->code,&e2.code);
	e1->type = copytype(e2.type);
	freetoken(&e2);
	c++;

    } while (lexan() == ',');
    retractlexan();
    e1->pflag = e2.pflag;
    e1->sym = e2.sym;
    e1->val = e2.val;
    if(c > 1)
	setConstp(e1->type);
}


void assign_expr(token_t * e1)
{
    cond_expr(e1);
    while (1)
	switch (lexan()) {
	  case ASS:		/* = */
	  case ADDEQ:		/* += */
	  case SUBEQ:		/* -= */
	  case DIVEQ:		/* /= */
	  case MULEQ:		/* *= */
	  case MODEQ:		/* %= */
	  case RSHTEQ:		/* >>= */
	  case LSHTEQ:		/* <<= */
	  case ANDEQ:		/* &= */
	  case BOREQ:		/* |= */
	  case XOREQ:		/* ^= */
	    assignop(assign_expr, e1, token.tok);
	    break;
	  default:
	    retractlexan();
	    generatePtr(e1);
	    return;
	}
}
void cond_expr(token_t * e1)
{
    log_or_expr(e1);
    if (lexan() == '?') {
	val_t v;
	int rt;
	token_t e2, e3;
	inittoken(&e2);
	out_expr(&e2);
	match(':', " :");
	inittoken(&e3);
	cond_expr(&e3);
	if(!isconst(e3.type))
	    output(&e3);
	cast2comm(&e2, &e3);
	output(&e3);
	output(e1);
	rt = gettype(e1->type);
	e1->type = freetype(e1->type);
	e1->type = copytype(e2.type);
	v.ival = e2.code.nextinst + 2;
#if 0
	  generate(&e1->code, jmpFint, &v, 0);
#else
	  genJump(e1,&v,rt,0);

#endif
	contoken(e1, &e2);

	v.ival = e1->code.nextinst;
	generate(&e1->code, jmpu, &v, 0);
	setInst(&e1->code,v.ival,val.ival,e3.code.nextinst+1);
	contoken(e1, &e3);
	 /* conditional's  can't form lvalues */
	setConstp(e1->type);
    } else
	retractlexan();
}

void fixit(code_t *C, int s)
{
    int i;
    for(i = s; i < C->nextinst;i++)
	switch(C->inst[i].opcode) {
	  case jmpFint:
	  case jmpFlng:
	  case jmpFdbl:
	  case jmpFptr:
	  case jmpTint:
	  case jmpTlng:
	  case jmpTdbl:
	  case jmpTptr:
	    if(C->inst[i].val.ival == INT_MAX)
		C->inst[i].val.ival = C->nextinst - i;
	}
}

void log_or_expr(token_t * e1)
{
    void do_lor(token_t *, int);
    log_and_expr(e1);
    if(lexan() == LOR) {
	token_t e2;
	int nxt = e1->code.nextinst;
	do_lor(e1, INT_MAX);
	do {
	    inittoken(&e2);
	    log_and_expr(&e2);
	    do_lor(&e2,INT_MAX);
	    contoken(e1,&e2);
	} while (lexan() == LOR);
	fixit(&e1->code,nxt);
    }
    retractlexan();
}

void log_and_expr(token_t * e1)
{
    void do_land(token_t *, int);
    inc_or_expr(e1);
    if(lexan() == LAND) {
	token_t e2;
	int nxt = e1->code.nextinst;
	do_land(e1, INT_MAX);
	do {
	    inittoken(&e2);
	    inc_or_expr(&e2);
	    do_land(&e2,INT_MAX);
	    contoken(e1,&e2);
	} while (lexan() == LAND);
	fixit(&e1->code,nxt);
    }
    retractlexan();
}

void inc_or_expr(token_t * e1)
{
    xor_expr(e1);
    while (lexan() == BOR)
	process_binary(xor_expr, e1, BOR);
    retractlexan();
}

void xor_expr(token_t * e1)
{
    and_expr(e1);
    while (lexan() == XOR)
	process_binary(and_expr, e1, XOR);
    retractlexan();
}

void and_expr(token_t * e1)
{
    equal_expr(e1);
    while (lexan() == AND)
	process_binary(equal_expr, e1, AND);
    retractlexan();
}

void equal_expr(token_t * e1)
{
    rel_expr(e1);
    while (1)
	switch (lexan()) {
	  case EQ:
	  case NE:
	    process_binary(rel_expr, e1, token.tok);
	    break;
	  default:
	    retractlexan();
	    return;
	}
}

void rel_expr(token_t * e1)
{
    shift_expr(e1);
    while (1)
	switch (lexan()) {
	  case LT:
	  case LE:
	  case GT:
	  case GE:
	    process_binary(shift_expr, e1, token.tok);
	    break;
	  default:
	    retractlexan();
	    return;
	}
}

void shift_expr(token_t * e1)
{
    add_expr(e1);
    while (1)
	switch (lexan()) {
	  case LSHT:
	  case RSHT:
	    process_binary(add_expr, e1, token.tok);
	    break;
	  default:
	    retractlexan();
	    return;
	}
}

void add_expr(token_t * e1)
{
    mult_expr(e1);
    while (1)
	switch (lexan()) {
	  case '+':
	  case '-':
	    process_binary(mult_expr, e1, token.tok);
	    break;
	  default:
	    retractlexan();
	    return;
	}
}

void mult_expr(token_t * e1)
{
    cast_expr(e1);
    while (1)
	switch (lexan()) {
	  case '*':
	  case '/':
	  case '%':
	    process_binary(cast_expr, e1, token.tok);
	    break;
	  default:
	    retractlexan();
	    return;
	}
}

void cast_expr(token_t * e1)
{
    if (lexan() != '(') {
	retractlexan();
	unary_expr(e1);
    } else {
	f_cast_expr(e1);
    }
}

void f_cast_expr(token_t * e1)
{
    token_t e2;
    switch (lexan()) {
      TYPESPEC:
      TYPEQUAL:
	inittoken(&e2);
	retractlexan();
	type_name(e1);
	match(')', " )");
	cast_expr(&e2);
	if (isconst(e2.type)) {
	    setConst(e1->type);
	    castconst(&e2, e1, 1);
	    e1->val = e2.val;
	} else {
	    output(&e2);
	    castvar(&e2, e1, 1);
	    e1->pflag = e2.pflag;
	    e1->sym = e2.sym;
	}
	contoken(e1, &e2);
	break;
	/* cast can't form lvalues */
	setConstp(e1->type);
      default:
	retractlexan();
	expr(e1);
	match(')', " )");
	r_postfix_expr(e1);
	break;
    }
}


void unary_expr(token_t * e1)
{
    int t;
    val_t v;
    
    t = lexan();
    switch (t) {
      case '+':			/* unary - */
      case '-':			/* unary + */
      case '*':			/* indirection */
      case '~':			/* ones complement */
      case NOT:			/* */
	cast_expr(e1);
	unaryop(e1, t);
	return;
      case INC:			/* ++ lval */
      case DEC:			/* -- lval */
	unary_expr(e1);
	unaryop(e1, t);
	return;
      case sizeofsym:
	NoPTR++;
	if(lexan() == '(') {
	    switch(lexan()) {
	      TYPESPEC:
	      TYPEQUAL:
		retractlexan();
		e1->type = freetype(e1->type);
		type_name(e1);
		break;
	      default: /* must be unary expr, i.e. ( expr ) */
		retractlexan();
		expr(e1);
	    }
	    match(')', " )");
	} else {
	    retractlexan();
	    unary_expr(e1);
	}
	freecode(&e1->code);
	if(isconst(e1->type)  && 
	   gettype(e1->type) == t_pointer &&
	   gettype(e1->type->nxt) == t_char) /* hack for char */
	                                          /* constants */
		e1->val.uival = strlen(e1->val.p.p) + 1;
	    else
		e1->val.uival = get_sizeof(e1->type);
   
	if(!e1->val.uival || !e1->type)
	    error("Invalid argument to sizeof");
	    
	e1->type = freetype(e1->type);
	
	e1->type = addtype(t_uint, e1->type);
	setConst(e1->type);
	e1->pflag = 0;
	NoPTR--;
	return;

    case AND:
#if 0
	/* this section of code is an attempt to
	 * to have constant addresses determined
	 * at compile time (not finished)
	 */
	NoPTR++;
	cast_expr(e1);
	t = gettype(e1->type);
	if(e1->sclass == c_register)
	    error("Cannot apply & to a register variable");
	/* check for global or static variable class */
	if(GI(e1) == 0 && !isconst(e1->type) && t != t_lval) {
	    void * getaddress(token_t * e1);
	    ptr_t *p;
	    if(t == t_union || t == t_struct)
		e1->type = addtype(t_pointer,e1->type);
	    p = getaddress(e1);
	    e1->val.p.ep = e1->val.p.sp = e1->val.p.p = p;
	    setConst(e1->type);
	}
	
	if(!isconst(e1->type)) {
	    switch (t) {
	    case t_char:
	    case t_uchar:
	    case t_short:
	    case t_ushort:
	    case t_int:
	    case t_uint:
	    CASE_LONG:
	    CASE_ULONG:
	    CASE_FLOAT:
	    case t_pointer:
		if (e1->pflag)
		    error("Must have lvalue");
		else
		    generate(&e1->code, lval, &e1->val, GI(e1));
		break;
	    case t_lval:
		e1->type = succType(e1->type);
		break;
	    case t_struct:
	    case t_union:
	    case t_array:
		generate(&e1->code, rvalptr, &e1->val, GI(e1));
		break;
	    case t_funcdec:
		exchtype(t_func, e1->type);
	    case t_func:
		v.p.p = e1->sym;
		generate(&e1->code, pushptr, &v, 0);
		break;
	    default:
		error("Must have lvalue");
	    }
	    e1->type = addtype(t_pointer, e1->type);
	    e1->pflag = 1;
	} else  /* nothing much to do */
	    if(gettype(e1->type) != t_pointer)
		error("Must have lvalue");
	NoPTR--;
	return;
    }

#else
	NoPTR++;
	cast_expr(e1);
	if(e1->sclass == c_register)
	    error("Cannot apply & to a register variable");
	switch (gettype(e1->type)) {
	  case t_char:
	  case t_uchar:
	  case t_short:
	  case t_ushort:
	  case t_int:
	  case t_uint:
	  CASE_LONG:
	  CASE_ULONG:
	  CASE_FLOAT:
	  case t_pointer:
	    if (e1->pflag)
		error("Must have lvalue");
	    else {
		e1->val.p.ep = (void*)(get_sizeof(e1->type)-1); 
		generate(&e1->code, lval, &e1->val, GI(e1));
	    }	
	    break;
	  case t_lval:
	    e1->type = succType(e1->type);
	    break;
	  case t_struct:
	  case t_union:
	  case t_array:
	    generate(&e1->code, rvalptr, &e1->val, GI(e1));
	    break;
	  case t_funcdec:
	    exchtype(t_func, e1->type);
	  case t_func:
	    v.p.p = e1->sym;
	    generate(&e1->code, pushptr, &v, 0);
	    break;
	  default:
	    error("Must have lvalue");
	}
	e1->type = addtype(t_pointer, e1->type);
	e1->pflag = 1;
	NoPTR--;
	return;
    }
#endif

    retractlexan();
    postfix_expr(e1);
}

void postfix_expr(token_t * e1)
{
    primary_expr(e1);
    r_postfix_expr(e1);
}

void r_postfix_expr(token_t * e1)
{
    void derefConst(token_t *);
    void binhlval(int, token_t *, token_t *);
    
    int t,c = 0;
    switch ((t = lexan())) {
      case INC:
      case DEC:
	unaryop(e1, t);
	if (t == INC)
	    do_inc_dec(e1, DEC);
	else
	    do_inc_dec(e1, INC);
	r_postfix_expr(e1);
	break;
      case '[':			/* handle array indexing */

	process_binary(expr,e1,'+');
	unaryop(e1, '*');
	
	match(']', " ]");
	r_postfix_expr(e1);

	break;

    case RARROW:
#if 0   /* this section of code was the start
	 * of get EiC to determine where possible
	 * addresses at compile time.
	 */
	if(!isconst(e1->type))
	    output(e1);
	else  /* else delay output */
	    ;
	    
	if (gettype(e1->type) != t_pointer) {
	    error("Pointer required");
	    break;
	}
	if(isconst(e1->type)) 
	    c = 1;
	else if(isconstp(e1->type))
	    c = 2;
	e1->type = succType(e1->type);
	if(c==1)
	    setConst(e1->type);
	else if(c==2)
	    setConstp(e1->type);
	
      case '.':
	if(isconst(e1->type)) 
	    c  = 1;
	else if( isconstp(e1->type))
	    c = 2;

	if(!isconst(e1->type))
	    output(e1);
	if (gettype(e1->type) == t_lval) {
	    e1->type = succType(e1->type);
	    e1->pflag = 1;
	}
	if (gettype(e1->type) == t_struct ||
	    gettype(e1->type) == t_union) {
	    if (lexan() == ID &&
		(t = findmem(e1->type, token.val.sym->id)) >= 0) {
		struct_t *S;
		val_t v;
		S = (void *) getInf(e1->type);
		if(!isconst(e1->type)) {
		    output(e1);
		    if (S->offset[t] > 0) {
			v.ival = 1;
			generate(&e1->code, bump, &v, 0);
			v.ival = S->offset[t];
			generate(&e1->code, pushint, &v, 0);
			generate(&e1->code, addptr2int, &v, 0);
		    }
		} else if (S->offset[t] > 0) {
		    /* handle constants */

		    e1->val.p.p = (char*) e1->val.p.p + S->offset[t];
		    e1->val.p.sp = e1->val.p.ep = e1->val.p.p;
			

		}
		    
		freetype(e1->type);
		e1->type = copytype(S->type[t]);
		if(c == 1)
		    e1->type = addtype(t_pointer,e1->type);		    
		else
		    e1->type = addtype(t_lval, e1->type);
		e1->pflag = 0;
		if (gettype(token.val.sym->type) == ID)
		    remsym(token.val.sym);
	    } else
		error("Illegal structure operation");
	} else
	    error("Illegal structure operation");
	if(c==1)
	    setConst(e1->type);
	else if(c==2)
	    setConstp(e1->type);

	r_postfix_expr(e1);
	break;
#else	
	output(e1);
	if (gettype(e1->type) != t_pointer) {
	    error("Pointer required");
	    break;
	}
	if(isconst(e1->type) || isconstp(e1->type)) 
	    c = 1;
	e1->type = succType(e1->type);
	if(c)
	    setConstp(e1->type);
      case '.':
	if(isconst(e1->type) || isconstp(e1->type))
	    c  = 1;
	output(e1);
	if (gettype(e1->type) == t_lval) {
	    e1->type = succType(e1->type);
	    e1->pflag = 1;
	}
	if (gettype(e1->type) == t_struct ||
	    gettype(e1->type) == t_union) {
	    if (lexan() == ID &&
		(t = findmem(e1->type, token.val.sym->id)) >= 0) {
		struct_t *S;
		val_t v;
		S = (void *) getInf(e1->type);
		output(e1);
		if (S->offset[t] > 0) {
		    v.ival = 1;
		    generate(&e1->code, bump, &v, 0);
		    v.ival = S->offset[t];
		    generate(&e1->code, pushint, &v, 0);
		    generate(&e1->code, addptr2int, &v, 0);
		}
		freetype(e1->type);
		e1->type = copytype(S->type[t]);
		e1->type = addtype(t_lval, e1->type);
		e1->pflag = 0;
		if (gettype(token.val.sym->type) == ID)
		    remsym(token.val.sym);
	    } else
		error("Illegal structure operation");
	} else
	    error("Illegal structure operation");
	if(c)
	    setConstp(e1->type);
	r_postfix_expr(e1);
	break;
#endif
      case '(':		/* handle function calls */
	
	if (gettype(e1->type) != t_lval)
	    output(e1);
	
	if(nextType(e1->type)) 
	    e1->type = succType(e1->type);
	t=gettype(e1->type);
	
	if(t == t_func || t == t_funcdec || t == t_builtin) {
	      
	    if(getInf(e1->type)) {
		val_t v;
		token_t e2;
		int c;

		inittoken(&e2);

		c = arg_expr_list(e1,&e2);

		v.ival = 1;
		generate(&e1->code, bump, &v, 0);
		if (c) {
		    v.ival = c;
		    generate(&e1->code, checkar, &v, 1);
		}
		contoken(e1, &e2);
		
		e1->type = addtype(t_lval, e1->type);
		e1->pflag = 0;
	    } else
		error("Incorrect function usage: %s",
		      e1->sym->id);
	} else
	    error("Incorrect function usage: %s",
		  e1->sym->id);
	
	match(')', " )");
	setConstp(e1->type);
	r_postfix_expr(e1);
	break;
      default:
	retractlexan();
    }    
}

int arg_expr_list(token_t * E1, token_t * e1)
{
    token_t * genTemporay(type_expr *type, int level);
    token_t e2;
    int t, t2, Svar = 0, aggparm = 0;
    val_t v;
    int ext = 0,count = 0;
    func_t *f;
    token_t *e3;


    if(HasHiddenParm(E1->type)) {
	/* need to now generate temporary variable */
	e3 =  genTemporay(nextType(E1->type),S_LEVEL);
	output(e3);
	/* concatenate code */
	concode(&e1->code,&e3->code);
	xfree(e3);
	generate(&e1->code, stoval, &e1->val, t_hidden);
	count++;
	ext = -1;
    }

    f = getInf(E1->type);
    
    do {
	inittoken(&e2);
	assign_expr(&e2);
	if ((t = gettype(e2.type)) != t_void) {

	    e1->type = getFPty(f,Svar);

	    t2 = gettype(e1->type);
	    if(t2 == t_void)
		error("Illegal parameter no. [%d]",count+1);	    

	    if(t == t_struct || t== t_union) {
		if(!isconst(e2.type) && !IsTemp(e2.type)) {
		    e3 =  genTemporay(e2.type,S_LEVEL);
		    output(e3);
		    /* concatenate code */
		    concode(&e2.code,&e3->code);
		    xfree(e3);
		    aggparm = 1;
		    v.ival = 1;
		    generate(&e2.code, bump, &v, t_hidden);
		}
		if(t2 == t_var) /* EiC doesn't not allow this */
		    error("passing a struct/union to variadic"
				 " function `%s'",E1->sym->id);
	    }
		
	    if (!isconst(e2.type)) {
		output(&e2);
		if (t2 != t_var)
		    castvar(&e2, e1, 0);
	    } else {		/* small bit of optimisation */
		if (t2 != t_var)
		    castconst(&e2, e1, 0);
		output(&e2);
	    }
	    if(aggparm) { /* passing a structure or union */
		v.ival = get_sizeof(e2.type);
		generate(&e2.code,refmem,&v,0);
		aggparm = 0;
	    }
	    v.ival = count;
	    generate(&e2.code, stoval, &v, gettype(e1->type));
	    

	    /* collect parameters in reverse order */
	    concode(&e2.code, &e1->code);
	    e1->code = e2.code;
	    freetype(e2.type);
	    e1->type = NULL;
	    count++;
	    if ((t2 != t_var) && (Svar < getFNp(f)-1))
		Svar++;
	} else {
	    freetype(e2.type);
	    if (count + ext != 0) {
		error("Illegal void parameter no [%d]",count+1);
	    } else
		ext++;
	}
	if (lexan() != ',') {
	    v.ival = count;
	    generate(&e1->code, pushint, &v, 0);
	}
    } while (token.tok == ',');
    retractlexan();

    t = gettype(getFPty(f,0));


    if((count == 0 && t != t_void && t != t_var) ||
       (count+ext != getFNp(f) && !IsVariadic(f) && t != t_var))
	error("Wrong number of arguments for %s",
	      E1->sym->id);		      
    
    return count;
}


void primary_expr(token_t * e1)
{
    extern int RETURNON;
    extern token_t *RETURNEXPR;
    
    switch (lexan()) {
      case '(':
	/*assign_expr(e1);*/
	expr(e1);
	match(')', " ) ");
	return;
      case CHAR:
	*e1 = token;
	e1->type = addtype(t_char, NULL);
	setConst(e1->type);
	break;
      case INT:
	*e1 = token;
	e1->type = addtype(t_int, NULL);
	setConst(e1->type);
	break;
      case UINT:
	*e1 = token;
	e1->type = addtype(t_uint, NULL);
	setConst(e1->type);
	break;
      case LONG:
	*e1 = token;
	e1->type = addtype(t_long, NULL);
	setConst(e1->type);
	break;
      case ULONG:
	*e1 = token;
	e1->type = addtype(t_ulong, NULL);
	setConst(e1->type);
	break;
      case FLOAT:
	*e1 = token;
	e1->type = addtype(t_double, NULL);
	setConst(e1->type);
	break;
      case STR:
	*e1 = token;
	e1->type = addtype(t_pointer, addtype(t_char,NULL));
	setConst(e1->type);
	if (!RETURNON)
	    xmark(e1->val.p.p, eicgstring); /* garbage */
	else	/* store string */
	    add_func_str(getInf(RETURNEXPR->type),
			 e1->val.p.p);
	break;
      case ID:
	if (gettype(token.val.sym->type) == ID) {
	    /*remsym(token.val.sym); */
	    error("Undeclared identifier %s",token.val.sym->id);
	}
	e1->type = copytype(token.val.sym->type);
	e1->val = token.val.sym->val;
	e1->sym = token.val.sym;
	e1->sclass = token.val.sym->sclass;
	/*generatePtr(e1);*/
	
	break;
      default:
	retractlexan();
	e1->pflag = 0;
	e1->type = addtype(t_void, e1->type);
	break;
    }
}

int do_binaryop(token_t * e1, token_t * e2, int op);

int GI(token_t * e1)
{
    /* returns -1 on error,
     *         1 if the type is automatic
     *         0  otherwise
     */
    if(e1 && e1->sym)
	return (e1->sym->level > 1 && !(e1->sym->sclass & c_static));
    else
	return -1;
}


void process_binary(void (*func) (token_t * e),
		    token_t * e1, int op)
{
    extern int checkPeepHole(token_t *e1,int op);
    token_t e2;
    val_t v;


    inittoken(&e2);

    (*func) (&e2);

    generatePtr(&e2);
    generatePtr(e1);
    
    if (!isconst(e1->type))
	output(e1);
    if (!isconst(e2.type))
	output(&e2);

    if(checkPeepHole(e1,op)) {
	if(gettype(e2.type) < gettype(e1->type)) {
	    if(isconst(e2.type))
		castconst(&e2,e1,1);
	    else
		castvar(&e2,e1,1);
	}	
	freetype(e1->type);
	*e1 = e2;
	return;
    }
    if(checkPeepHole(&e2,op)) {
	if(gettype(e2.type) > gettype(e1->type)) {
	    if(isconst(e1->type))
		castconst(e1,&e2,1);
	    else
		castvar(e1,&e2,1);
	}
	freetoken(&e2);
	return;
    }

    bin_validate(op, e1, &e2);

    if (isconst(e1->type) && isconst(e2.type)) {
	contoken(e1, &e2);
	e1->pflag = 0;
	return;
    }
    if (/*isconst(e1->type) && */ !e1->pflag)
	output(e1);

    if(op != LAND) {
	v.ival = 1;
	generate(&e1->code, bump, &v, 0);
    }
    contoken(e1, &e2);
}


void assignop(void (*func) (token_t * e),
	      token_t * e1, int op)
{
    int t;
    token_t e2;
    val_t v;
    int op2;

    if (e1->pflag || isconst(e1->type) || isconstp(e1->type))
	error("Illegal assignment operation");

    inittoken(&e2);
    
    generatePtr(&e2);

    switch (op) {
      case ADDEQ: op2 = '+'; break;
      case SUBEQ: op2 = '-'; break;
      case DIVEQ: op2 = '/'; break;
      case MULEQ: op2 = '*'; break;
      case MODEQ: op2 = '%'; break;
      case RSHTEQ: op2 = RSHT; break;
      case LSHTEQ: op2 = LSHT; break;
      case ANDEQ: op2 = AND; break;
      case BOREQ: op2 = BOR; break;
      case XOREQ: op2 = XOR; break;
      default:			/* do equals */
	op2 = 0;
	(*func) (&e2);
    }

    t = gettype(e1->type);
    if((t == t_struct || t == t_union) && check4constmem(e1->type))
	error("Illegal assignment operation");
    
    if (op2) {
	e2.type = copytype(e1->type);
	e2.val = e1->val;
	e2.sym = e1->sym;
	if (t == t_lval || t == t_struct || t == t_union)  {
	    v.ival = 1;
	    generate(&e1->code, dupval, &v, 0);
	}
	concode(&e2.code, &e1->code);
	process_binary(func, &e2, op2);
    } 	

    /*
     * check 4 assignment of pointer to const X to
     * pointer to X: this is illegal.
     */
    if(t == t_pointer && gettype(e2.type) == t_pointer)
	if(ConstIntegrity(nextType(e1->type),nextType(e2.type)))
	    error("Assignment loses a const qualifier");
    
    if (isconst(e2.type)) {
	void SaveGlobalString(ptr_t *s);
	extern int RETURNON;
	castconst(&e2, e1, 0);
	if(!RETURNON && gettype(e2.type) == t_pointer
	   && gettype(nextType(e2.type)) == t_char) {
	    /* got string */
	    /* Note: string is markged for garbage collector */
	    SaveGlobalString(&e2.val.p);
	}
	output(&e2);
    } else {
	output(&e2);
	castvar(&e2, e1, 0);
    }
    if (!e2.pflag)
	error("Invalid assignment");

    if (t == t_lval || t == t_struct || t == t_union) {
	if(!op2) {
	    v.ival = 1;
	    if(!e1->code.nextinst) {
		output(e1);
	    }
	    if(t==t_struct || t == t_union) {
		generate(&e1->code, bump, &v, 0);
		v.ival = get_sizeof(e2.type);
		generate(&e2.code,refmem,&v,0);
	    } else
		generate(&e1->code, bump, &v, 0);
	    

	} else 
	    concode(&e2.code, &e1->code);
    }
    contoken(e1, &e2);
    do_stooutput(e1);
    /* assignments can't form lvalues */
    setConstp(e1->type);
}

int findmem(type_expr * t, char *id)
{
    int i;
    struct_t *S = (struct_t *)getInf(t);
    for (i = 0; i < S->n; i++)
	if (strcmp(id, S->id[i]) == 0)
	    return i;

    return -1;
}

int check4constmem(type_expr *t)
{
    struct_t *S = (struct_t *)getInf(t);
    int i;
    for(i=0;i<S->n;i++)
	if(isconstp(S->type[i]))
	    return 1;
    return 0;
}
	

