/* starteic.c
 *
 *	(C) Copyright Dec 31 1995, Edmond J. Breen.
 *		   ALL RIGHTS RESERVED.
 * This code may be copied for personal, non-profit use only.
 *
 */

/* In accordance with the practice of fair use, I hereby acknowledge
 * that:
 */

static char *PlaugerStr_=
"Portions of this work are derived from the Standard C library, (C), 1992 by P.J. Plauger, published by Prentice-Hall and are used with permission.";

#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include "ctype.h"
#include "signal.h"
#include <setjmp.h>

#include "MachSet.h"
#include "lexer.h"
#include "global.h"
#include "typemod.h"
#include "func.h"
#include "xalloc.h"
#include "preproc.h"
#include "emitter.h"
#include "parser.h"
#include "symbol.h"

extern void interpret(environ_t * env);

#define HERE printf("%s %d\n",__FILE__,__LINE__)

char *Version =
"*************************************************************************\n"
"EiC V3.6 - Copyright (c) 1995 to 1998,"
" by Edmond J. Breen\n"
"EiC comes `as is' and with ABSOLUTELY NO WARRANTY OF MERCHANTIBILITY AND\n"
" FITNESS OF PURPOSE\n"
"*************************************************************************\n"
;
    

environ_t *ENV;

size_t ENV_NextEntryNum()
{
    return stab_NextEntryNum(&ENV->stab);
}

void ENV_CleanUp(size_t bot)
{
    stab_CleanUp(&ENV->stab,bot);
}

/*END STRING TABLE ROUTINES*/

void reset_env_pointers(token_t * e1, int bp)
{
    ENV->lsp = bp;
}


void markENV(char mark)
{
    stab_Mark(&ENV->stab,mark);
}	


void free_env_code(void)
{
    unsigned int i;
    inst_t *inst;
     inst = ENV->CODE.inst;
     /* rem free up other info also */
    for (i = 0; i < ENV->CODE.nextinst; i++, inst++)
	if (inst->opcode == jmptab) {
	    eicstack_t *s;
	    s = inst->val.p.p;
	    xfree(s->val);
	    xfree(s);
	} else if (inst->opcode == assigntype) {
	    freetype(inst->val.p.p);

	} else if(inst->opcode == halt) {
	    freetype(inst->val.p.p);
	}

    freecode(&ENV->CODE);
}



void SaveGlobalString(ptr_t *s)
{
    int len = (char*)s->ep - (char*)s->sp;
    s->sp = s->p = stab_SaveString(&ENV->stab,s->p);
    s->ep = (char*)s->p + len;
}

/* signal handling  routines*/

jmp_buf env;
static int jmpOn = 0;
#define _Err_   1
#define _Exit_  2

char * CurrentFile =NULL;
unsigned int  CurrentLine = -1;

#define errStr(x)   if(CurrentFile) fprintf(stderr,"\n" x ", file %s, line %d\n",CurrentFile,CurrentLine)
#define eic_disaster() fprintf(stderr,"exit EiC\n");exit(1)



void exit_call(int i)
{
#ifndef _SEIC_
    size_t fopen_entry = fopen_NextEntryNum(); 
    signal(SIGUSR1, exit_call);
    _ffexit(fopen_entry);
    if(jmpOn && jmpOn++ <= 1)
      longjmp(env,_Exit_);
    else
	eic_disaster();
#else
    eic_disaster();
#endif
 
}

void  bus_err(int i)
{
    signal(SIGBUS, bus_err);
    errStr("EiC bus error trapped");
    if(jmpOn && jmpOn++ <= 1)
      longjmp(env,_Err_);
    else
      eic_disaster();
}
void  ill_oper(int i)
{
    signal(SIGILL, ill_oper);
    errStr("EiC illegal operation attempted\n");
    if(jmpOn && jmpOn++ <= 1)
      longjmp(env,_Err_);
    else
      eic_disaster();
}
void stor_access(int i)
{
    signal(SIGSEGV, stor_access);
    errStr("EiC illegal storage access");
    if(jmpOn && jmpOn++ <= 1)  {
	longjmp(env,_Err_);
    } else
	eic_disaster();
}

void float_err(int i)
{
    signal(SIGFPE, float_err);
    errStr("EiC maths  exception");
    if(jmpOn && jmpOn++ <= 1)
      longjmp(env,_Err_);
    else
      eic_disaster();
}

void term_int(int i)
{
    signal(SIGINT, term_int);
    errStr("EiC interrupted");
    if(jmpOn && jmpOn++ <= 1)
      longjmp(env,_Err_);
    else
      eic_disaster();
}

FILE *EiChist = NULL;
extern int verboseON, SHOWLINE, traceON, TIMER, interActive, showIncludes;
int silent = 0, HistoryFile = 1;
int StartUpH = 1;
int reStart = 0, prompt = 0;
int EiC_Interact = 1;
int ScriptMode = 0;

void usage(void)
{
    puts("\n\t\tEiC\n"
	 " An Extensible Interactive C interpreter\n"
	 " To start eic, type eic.\n"
	 " To exit eic, type :exit.\n"
	 "\n Usage:\n"
	 "\teic  [-Ipath] [-Dname[=var]] -[hHvVcCrR]  [file] [fileargs]\n"
	 " Options:\n"
	 "   C preprocessor directives:\n"
	 "\t-Ipath      search for include files in path\n"
	 "\t-Dname      define a symbolic name to the value 1\n"
	 "\t-Dname=var  define a symbolic name to the value var\n"
	 "\t            Note, there is no spaces allowed\n"
	 "   EiC directives:\n"
	 "\t-h -H       causes this usage to be displayed\n"
	 "\t-v -V       print EiC's Log information\n"
	 "\t-p          showline\n"
	 "\t-P          show path of include files\n"
	 "\t-t -T       turns trace on\n"
	 "\t-c -C       turns timer on\n"
	 "\t-e          echo HTML mode\n" 
	 "\t-r          restart EiC. Causes EiC to be re initiated\n"
         "\t               from the contents of EiChist.lst file\n"
	 "\t-R          same as `r', but prompts the user to accept\n"
	 "\t               or reject each input line first\n" 
	 "\t-s -S       run silently\n"
	 "\t-f          run in script mode\n"
	 "\t-n          no history file\n"
	 "\t-N          don't use any startup.h files\n"
	 "\t-A          Non-interactive-mode\n"
	 "\tfile        EiC will execute `file' and then stop; for example:\n"
	 "\t              % eic foo.c \n"
	 "\tfileargs    command line arguments, which get passed onto file\n"
	 );
    exit(1);
}


int do_sw_commands(char *cp)
{
    int insertpath(const char *path);
    void dodefine(char * def);

    while(*cp) {
	switch(*cp++) {
	case 'D': dodefine(cp);  return 1;
	case 'I': insertpath(cp); return 1;
	case '\?':
	case 'h':
	case 'H': usage();
	case 'c':
	case 'C': TIMER = 1; break;
	case 'e':
	    /* connect stderr to stdout */
	    dup2(1,2); 
	    /* setup stdout to behave like stderr */
	    if(setvbuf(stdout,NULL,_IONBF,0) != 0)
		fprintf(stderr,"failed to setup stdout\n");
	    /* inform the browser */
	    puts("Content-type: text/plain\n\n");
	    break;
	case 'v':
	case 'V': puts(Version); exit(1);

	case 'R': prompt = 1;
	case 'r': reStart = 1;  break;

	case 's': 
	case 'S': silent = 1; break; 

	case 'f': ScriptMode = 1; break;
	case 'p': SHOWLINE = 1; break;
	case 'P': showIncludes = 1; break;
	case 't':
	case 'T': traceON = 1; break;
	case 'n': HistoryFile = 0; break;
	case 'N': StartUpH = 0; break;
	case 'A': EiC_Interact = 0; break;
	default:
	    while(isspace(*cp)) cp++;
	    if(*cp == '-')  /* assume script mode */
		while(isspace(*++cp));
	    else if(*cp)  /* catch for lines ending with whitespace */
		return 0;
	}
    }
    return 1;
}

void save_history(FILE *fp, int from);

extern int EiC_errs;
extern int verboseON;
size_t xalloc_bot, pp_bot, lut_bot, ENV_bot;
void clear_err_msgs(void);

void  pp_CleanUp(size_t bot);
size_t pp_NextEntryNum(void);
int setinfile(const char *);
int Include_file(char *, int);
void marksyms(char);
    
char *inbuf;
extern char *nextproline();
extern int listcodeON, interpON,memdumpON;
extern unsigned long tot_memory;
extern size_t tot_alloc;

void do_displays()
{
    if(verboseON)
	printf("\ninstr = %d sARsize = %d  lsp = %d aARsize  = %d\n"
	       "tot_alloc = %lu tot_memory = %lu\n",
	       ENV->CODE.nextinst,
	       ENV->ARsize,
	       ENV->lsp,
	       ENV->LARsize,
	       (unsigned long)tot_alloc,
	       tot_memory);
    listcode(&ENV->CODE);
}

void runEiC()
{
    void remTempories(void);
    void peephole(code_t *c);
    int getHistLineNo(), hfrom;
    static int E, errR =0;

    if((E = setjmp(env)) == 0) {
	jmpOn = 1;
	errR = 0;
	ENV_bot = ENV_NextEntryNum();
	lut_bot = lut_NextEntryNum();
	pp_bot = pp_NextEntryNum();
	xalloc_bot = xalloc_NextEntryNum();
	hfrom = getHistLineNo();
	inbuf = nextproline();

	initlex(inbuf);
	initparser();
	parse(ENV);
	peephole(&getenvcode(ENV));
	
	if(listcodeON)
	    do_displays();
	if (ENV->CODE.nextinst && interpON && !ParseError) {
	    interpret(ENV);
	}
    } else if(E == _Err_)
	ParseError = 1;

    if (EiC_errs) 
	clear_err_msgs();
    

    xfreemark(eicgstring);
    free_env_code();
    remTempories();
    
    if(ParseError) {
	size_t xalloc_top = xalloc_NextEntryNum();
	if(ENV->lsp != 0) {
	    fputs("EiC::Reset Local Stack Pointer\n",stdout);
	    ENV->lsp = 0;
	}	
	if(ENV->ARgar.n)
	    xmark(ENV->ARgar.val,eicstay);

	fprintf(stderr,"EiC: error clean up entry pt ");
	/* The while loop is a precaution against
	 * errors that occur during cleanup.
	 */
	while(errR < 10) {
	    fprintf(stderr,"%d,",errR);
	    switch(errR++) {
	      case 0: lut_CleanUp(lut_bot); break;
	      case 1: ENV_CleanUp(ENV_bot); break;
	      case 2: pp_CleanUp(pp_bot); break;
	      case 3: xalloc_CleanUp(xalloc_bot,xalloc_top); break;
	    default:
		printf("\n");
		errR = 10;
	    }
	}
	ParseError = 0;
    } else 
	if(EiChist)
	    save_history(EiChist,hfrom);
    if(memdumpON) {
	marksyms(NON_LEAK);
	if(verboseON)
	    printf("--- XMEM DUMP\n");
	xdumpnonmark("eicxdump", NON_LEAK);
    }
}

#define FORMAT1 "\n#ifdef _STDLIBH\nexit(main());\n"\
                 "#else\nmain();\n#endif\n"

#define FORMAT2   "\n"\
	  "#ifdef _STDLIBH\n"\
	  "exit(main(_Argc,_Argv));\n"\
	  "#else\n"\
	  "main(_Argc,_Argv);\n"\
	  "#endif\n"

int doEntry(char *name)
{
    func_t *F;
    symentry_t *sym;
    FILE *fp;
    sym = lookup(stand_tab,"main");
    if(!sym) {
	if(!ScriptMode) 
	    fprintf(stderr,"Error: Missing main\n");
	return 0;
    }
    if(gettype(sym->type) != t_func) {
	fprintf(stderr,"Error: main NOT DECLARED as a function\n");
	return 0;
    }
    /* get number of arguments */

    fp = fopen(name,"w");
    if(!fp) {
	fprintf(stderr,"Error: Failed to initiate properly\n");
	return 0;
    }
    F = getInf(sym->type);

    if(getFNp(F) <= 1)  /* allow for void argument */
	fputs(FORMAT1,fp);
    else
	fputs(FORMAT2,fp);
    fclose(fp);
    return 1;
}


char * doargs(int argc, char **argv)
{
    int i,sz;
    char buf[51];
    char *p;
    
    sprintf(buf,"int _Argc = %d;char *_Argv[] = {",argc);

    /* get size of string needed */
    sz = strlen(buf);
    for(i=0;i<argc;++i)
	sz += strlen(argv[i]) + 3; /* 3  comma + quotes */
    /* needs to be NULL terminated */
    sz += strlen("(void*)0};") + 2; /* 2 for NULL */

    p = xmalloc(sz);

    *p ='\0';
    strcat(p,buf);
    for(i=0;i<argc;++i) {
	strcat(p,"\"");
        strcat(p,argv[i]);
	strcat(p,"\",");
    }
    strcat(p,"(void*)0};");
    return p;
}
    


void startEiC(int argc, char **argv)
{
    extern load_history(char *fname,int prompt);
    if(argc-- > 0)
	while(argv[1]) {
	    if(argv[1][0] == '-') {
		if(!do_sw_commands(&argv[1][1])) {
		    fprintf(stderr,"Unknown switch command [%s]\n",
			    argv[1]);
		    usage();
		} 
	    } else 
		break;
	    argv++;
	    argc--;
	}



    
    signal(SIGBUS, bus_err);
    signal(SIGILL, ill_oper);
    signal(SIGSEGV, stor_access);
    signal(SIGFPE, float_err);
    signal(SIGINT, term_int);
    signal(SIGUSR1, exit_call);

    while(argc-- >0) {
	char * n1 = tmpnam(NULL);
	char * n2;
	char *prolineString(char *str);

	
	n2 = doargs(argc+1,argv+1);


	prolineString(n2);

	interActive = 0;  /* turn off interactive mode */

	if(!Include_file(argv[1],1)) 
	    fprintf(stderr,"Error: Failed to load file %s\n",argv[1]);
	else 
	    runEiC();

	if(!ScriptMode && doEntry(n1)) {
	    if(!Include_file(n1,1)) 
		fprintf(stderr,"Error: Failed to load file %s\n",n1);
	    else
		runEiC();
	} 
	remove(n1);
	xfree(n2);
	exit(0);
    }

    if(!silent) {
      fputs(Version,stdout);
    }
    if(StartUpH)
	setinfile("starteic.h");

    if(HistoryFile) {
	char * name = "EiChist.lst";
	if(reStart) {
	  puts("Re Initiating EiC -- please wait.");
	  if(load_history(name,prompt)) {
	    runEiC();  /* ensure that startup files have
			  been loaded and compiled */
	    EiChist = fopen(name,"w");
	    save_history(EiChist,0);
	    fclose(EiChist);
		
	    Include_file(name,1);
	    runEiC();
	    EiChist = fopen(name,"a");
	  }else {
	    fprintf(stderr,"Error: unable to load history file\n");
	    HistoryFile = 0;
	    EiChist = NULL;
	  }
	  puts("Done.");
	} else {
	  EiChist = fopen(name,"w");
	  if(!EiChist) {
	    fprintf(stderr,"Failed to create EiChist.lst\n"
		    "Start No history file mode switch\n");
	    HistoryFile = 0;
	  }
	}
	if(HistoryFile)	    
	  setvbuf(EiChist,NULL,_IOLBF,0);
      }
    
      while(1) 
	runEiC();
}






