/* symbol.c
 *
 *	(C) Copyright Apr 15 1995, Edmond J. Breen.
 *		   ALL RIGHTS RESERVED.
 * This code may be copied for personal, non-profit use only.
 *
 */

#include "stdio.h"
#include "stdlib.h"
#include "string.h"

#include "typemod.h"
#include "MachSet.h"
#include  "global.h"
#include "xalloc.h"
#include "preproc.h"
#include "symbol.h"


int iskeyword(keyword_t *keywords,char*id,int n)
{
    int i;
    for(i=0;i<n;i++)
	if(strcmp(keywords[i].id,id) == 0)
	    return keywords[i].token;
    return 0;
}

#define MoD 2
void eicpush(eicstack_t *s, val_t v)
{
    if(!(s->n%MoD)) {
	if(!s->n)
	    s->val = (val_t*)xcalloc(sizeof(val_t),MoD);
	else
	    s->val = (val_t*)xrealloc(s->val,(s->n+MoD)*sizeof(val_t));
    }
    s->val[s->n] = v;
    s->n++;
}

int eicpop(eicstack_t *s, val_t *pop)
{
    if(s->n == 0)
	return 0;
    s->n--;
    *pop = s->val[s->n];
    if(!(s->n%MoD)) {		
	if(!s->n)
	    xfree(s->val);
	else
	    s->val = (val_t*)xrealloc(s->val,s->n*sizeof(val_t));
    }
    return 1;
}


/* LOOK UP TABLE ROUTINES
   --------------------------*/

static size_t _EnTrY_No = 0;
int hashsmc(char * s,int mod);
symentry_t *HTAB[HSIZE];


#if 1
/* code for generating tempories */
unsigned int NumTemps=0, CurTemp =0;

symentry_t * nxtTemp(int obj, int level)
{
    symentry_t *sym;
    char tmpName[50];
    sprintf(tmpName,"%s%d","__TeMp",NumTemps);

    /* There should be no need to watch out for change of level !!!!
     * It will be assumed that the compound statement routine will
     * handle it.
     */

#if 0

    if(CurTemp >= NumTemps) { /* generate new entry */
	sym = insertLUT(work_tab,tmpName,obj);
	CurTemp = NumTemps;
	NumTemps++;
    } else {
	sym = lookup(work_tab,tmpName);
	freetype(sym->type);
	sym->type = addtype(obj,NULL);
	sym->val.ival = -1;	 /* indicates  unused */
    }
#else

    sym = insertLUT(work_tab,tmpName,obj);
    NumTemps++;

#endif

    CurTemp++;

    if(sym->val.ival == -1) /* needs a home */ 
	stackit(sym,level);

    /* Setting up the token information is left
     * to the caller of this routine.
     */

    sym->level = level;
    return sym;
}    

#endif


size_t lut_NextEntryNum(void)
{
    return _EnTrY_No;
}
                             /*CUT EiChashFunc*/
int hashsmc(char * s,int mod)
{
    register unsigned int h, c;
    h = 0;
    while(  (c = (int) *s++) > 0)
	h =   (h << 1) + c;
    return ( h % mod);
}
                           /*END CUT*/

                            /*CUT lutLookUp*/

int Pclash;

char *getClashedfname(char nspace,char *id)
{
    symentry_t *sym;
    for(sym = HTAB[hashsmc(id,HSIZE)]; sym != NULL; sym = sym->next)
	if(sym->nspace == nspace && strcmp(id,sym->id) == 0)
	    if((sym->sclass & c_private) &&
	       sym->fname != CurrentFileName()) {
		return sym->fname;
	    }
    return NULL;
}
symentry_t * lookup(char nspace, char *id)
{
    symentry_t *sym;
    Pclash = 0;
    for(sym = HTAB[hashsmc(id,HSIZE)]; sym != NULL; sym = sym->next)
	if(sym->nspace == nspace && strcmp(id,sym->id) == 0)
	    if((sym->sclass & c_private) &&
	       sym->fname != CurrentFileName()) {
		Pclash = 1;
		continue;
	    }
	    else
		break;
    return(sym);
}
                             /*END CUT*/

                           /*CUT lutInsert*/
symentry_t * insertLUT(char nspace,char *id,int type)
{
    symentry_t *sym;
    auto int  hashval;

    sym = (symentry_t *) xcalloc(1,sizeof(symentry_t));
    if(sym == NULL)
	return(NULL);
    if( (sym->id = strsave(id)) == NULL) {
	xfree(sym);
	return(NULL);
    }
    sym->entry = _EnTrY_No++;
    hashval = hashsmc(sym->id,HSIZE);
    sym->next = HTAB[hashval];
    HTAB[hashval] = sym;
    sym->nspace = nspace;
    sym->val.ival = -1;		/* indicates  unused */
    sym->type = addtype(type,NULL);
    sym->fname =  CurrentFileName();
    return(sym);
}
                           /*END CUT*/

void delete(symentry_t *sym)
{
    auto symentry_t * this;
    auto int idx;
    
    idx = hashsmc(sym->id,HSIZE);
    this = HTAB[idx];
    if(this == sym)
	HTAB[idx] = sym->next;
    else {			/* find and unlink  */
	while(this && this->next != sym)
	    this = this->next;
	this->next = sym->next;
    }
}

void UpdateSymPos(symentry_t * sym)
{

    if(sym->next && sym->next->level > sym->level) {
	symentry_t * p = HTAB[hashsmc(sym->id,HSIZE)];
	delete(sym);
	while(p->next != sym)
	    if(p->next) {
		if(p->next->level > sym->level)
		    p=p->next;
		else {
		    sym->next = p->next;
		    p->next = sym;
		}
	    } else {
		p->next = sym;
		sym->next = NULL;
	    }
    }
}
                           /*CUT lutRemSym*/
void remsym(symentry_t *sym)
{
    delete(sym);
    free_sym(sym);
}
                       /*END CUT*/
                       /*CUT lutRemLevel*/
void remlevel(int level)
{
    int i;
    symentry_t * sym, *symh;

    for(i=0;i<HSIZE;i++) {
	sym = HTAB[i];
	while(sym && sym->level >= level) {
	    symh = sym->next;
	    free_sym(sym);
	    sym = symh;
	}
	HTAB[i] = sym;
    }
}
                     /*END CUT*/
                    /*CUT lutRemTempories*/
void remTempories(void)
{
    int i;
    symentry_t * sym, *symh;

    for(i=0;i<HSIZE;i++) {
	sym = HTAB[i];
	while(sym && IsTemp(sym->type)) {
	    symh = sym->next;
	    free_sym(sym);
	    sym = symh;
	}
	HTAB[i] = sym;
    }
}
                    /*END CUT*/

                    /*CUT lutClear*/
void lutClearFileEntries(char *FileName)
{
    int i;
    symentry_t * sym, *p, *t;

    for(i=0;i<HSIZE;i++) {
	t = sym = HTAB[i];
	p = NULL;
	while(sym){
	    if(strcmp(sym->fname,FileName) == 0) {
		if(p)
		    p->next = sym->next;
		else
		    t = sym->next;
		free_sym(sym);
		if(!p) {
		    sym = t;
		    continue;
		}
	    } else
		p = sym;
	    sym=p->next;
	}
	HTAB[i] = t;
    }
}
                    /*END CUT*/
                     /*CUT lutCleanUp*/
#if 0
void lut_CleanUp(size_t bot)
{
    int i;
    symentry_t * sym, *symh;

    for(i=0;i<HSIZE;i++) {
	sym = HTAB[i];
	while(sym && sym->entry >= bot) {
	   /* printf("%s\n",sym->id);*/
	    symh = sym->next;
	    free_sym(sym);
	    sym = symh;
	}
	HTAB[i] = sym;
    }
}
#else
void lut_CleanUp(size_t bot)
{
    int i;
    symentry_t * sym, *p, *t;

    for(i=0;i<HSIZE;i++) {
	t = sym = HTAB[i];
	p = NULL;
	while(sym){
	    if(sym->entry >= bot) {
		if(p)
		    p->next = sym->next;
		else
		    t = sym->next;
		free_sym(sym);
		if(!p) {
		    sym = t;
		    continue;
		}
	    } else
		p = sym;
	    sym=p->next;
	}
	HTAB[i] = t;
    }
}
#endif
                     /*END CUT*/
                     /*CUT lutFreeSym*/

#define freeAg(X)   do {\
			    symentry_t *x = X; \
			    if(!isconst(x->type)) \
			         xfree(ENV->AR[x->val.ival].v.p.p);\
			    else xfree(x->val.p.p);\
		    } while (0)


#define freeAg1(X)  xfree(ENV->AR[X->val.ival].v.p.p)


void free_sym(symentry_t *sym)
{
    /*printf("[%s] %d  %d [%d]\n",
      sym->id, sym->entry, sym->val.ival,gettype(sym->type));
      */
    
    xfree(sym->id);
    
    if(gettype(sym->type) == t_func) {
	killcode(sym);
    } else if(sym->level == 1)  { /* global value */
	int t;
	if((t=gettype(sym->type)) == t_array) 
	    freeAg(sym);
	else if ((t== t_union || t == t_struct) && sym->val.ival >=0 ) 
	    freeAg(sym);
		
    }
    /*
     * the conditions for pushing  onto the ARgar stack
     * must be the same as those for stacking as found in
     * function establish_id
     */
    if(sym->val.ival >=0 && sym->level == 1) 
	    if(! isconst(sym->type) &&
	       sym->nspace == stand_tab && sym->sclass != c_typedef) {
		eicpush(&ENV->ARgar,sym->val);
	    }
    freetype(sym->type);
    xfree(sym);
}
                        /*END CUT*/

void UpdateEntry(symentry_t *sym)
{
    int t = gettype(sym->type);
    if(CurrentFileName() != sym->fname &&
       t != t_func &&
       t != t_funcdec &&
       t != t_builtin)
	return;
    sym->entry =  _EnTrY_No++;
    sym->pname = sym->fname;
    sym->fname = CurrentFileName();
}

void addoffsettolevel(char nspace,int level,int off)
{
    int i;
    symentry_t * sym;

    for(i=0;i<HSIZE;i++) {
	sym = HTAB[i];
	while(sym && sym->level == level && sym->nspace == nspace) {
	    sym->val.ival = -(sym->val.ival + off);
	    sym = sym->next;
	}
    }
}

void marktype(type_expr *type, char mark)
{
    int i;
    struct_t *s;
    void markFunc(type_expr *t,int mark);
    
    while(type) {
	xmark(type,mark);
	switch(gettype(type)) {
	  case t_builtin:
	    if(getInf(type) == NULL)
		break;
	  case t_funcdec:
	  case t_func:
	    markFunc(type,mark);
	    break;
	  case t_union:
	  case t_struct:
	    if(type->alias)
		break;
	    s = getInf(type);
	    /*
	     * REM must allow for incomplete
	     * types.
	     */
	    if(!s)
		break;
	    xmark(s,mark);
	    if(!s->cl)
		break;
	    xmark(s->offset,mark);
	    xmark(s->id,mark);
	    xmark(s->type,mark);
	    for(i=0;i<s->n;i++) {
		xmark(s->id[i],mark);
		marktype(s->type[i],mark);
	    }
	    if(s->ntags) {
		xmark(s->tag,mark);
		for(i=0;i<s->ntags;++i)
		    marktype(s->tag[i],mark);
	    }
	    break;
	}
	type = nextType(type);
    }
}

void markcode(symentry_t *sym,char mark)
{
    int i;
    inst_t *inst;
    code_t * code;
    code = ENV->AR[sym->val.ival].v.p.p;
    xmark(code,mark);
    xmark(code->inst,mark);
    inst = code->inst;
    for(i=0;i<code->nextinst;i++,inst++)
	if(inst->opcode == jmptab) {
	    eicstack_t * s;
	    s = inst->val.p.p;
	    xmark(s->val,mark);
	    xmark(s,mark);
	} else if(inst->opcode == assigntype)
	    marktype(inst->val.p.p,mark);
}


void marksyms(char mark)
{
    void markmacros(char);
    void markENV(char);
    
    int i;
    symentry_t * sym;

    markmacros(mark);
    
    if(ENV->AR)
	xmark(ENV->AR,mark);
    if(ENV->ARgar.n) 
	xmark(ENV->ARgar.val,mark);
    if(ENV->LAR)
	xmark(ENV->LAR,mark);
    if(ENV->CODE.nextinst)
	xmark(ENV->CODE.inst,mark);
    xmark(ENV,mark);

    markENV(mark);
    
    for(i=0;i<HSIZE;i++) {
	sym = HTAB[i];
	while(sym) {
	    /*printf("marking %s\n",sym->id);*/
	    xmark(sym,mark);
	    xmark(sym->id,mark);
	    marktype(sym->type,mark);
	    if(sym->nspace != tag_tab)
		switch(gettype(sym->type)) {
		  case t_func: markcode(sym,mark); break;
		  case t_array:
		  case t_union:
		  case t_struct:
		    if(isconst(sym->type))
		       xmark(sym->val.p.p,mark);
		    else
			if(sym->sclass != c_typedef && sym->val.ival >= 0)
			    xmark(ENV->AR[sym->val.ival].v.p.p,mark);
		    break;
		}
	    sym = sym->next;
	}
    }
}

char * strsave(char *s)
{
    char *p;
    int n;

    for(n = 0,p =s; *p != '\0';++p,++n)
	;
    n++;
    p = xcalloc(n,sizeof(char));
    if(p) 
	while((*p++ = *s++));
    return p - n;
}

void newsymtype(symentry_t *sym, type_expr *t)
{
    if(sym) {
	if(sym->type && sym->type != t)
	    freetype(sym->type);
	sym->type = t;
    }
}

int nextstackitem(int level)
{
    if(level == 1) {		/* static variables */
	val_t v;
	if(!eicpop(&ENV->ARgar,&v))  { /* check for spare slots */
	    if(ENV->sp == ENV->ARsize) {
		if(!ENV->ARsize)
		    ENV->AR = (AR_t*)xcalloc(sizeof(AR_t),1);
		else
		    ENV->AR = (AR_t*)xrealloc(ENV->AR,
					      (ENV->sp+1)*sizeof(AR_t));
		ENV->ARsize++;
	    }
	    v.ival = ENV->sp;
	    ENV->sp++;
	}
	return v.ival;
    } else {			/* automatic variables */
	if(ENV->lsp == ENV->LARsize) {
	    if(!ENV->LARsize)
		ENV->LAR = (AR_t*)xcalloc(sizeof(AR_t),1);
	    else
		ENV->LAR = (AR_t*)xrealloc(ENV->LAR,(ENV->lsp+1)*
					   sizeof(AR_t));
	    ENV->LARsize++;
	}
	ENV->lsp++;
	return ENV->lsp - 1;
    }
}

int stackit(symentry_t * sym,int level)
{
    int i;
    AR_t * ar;
    
    i = nextstackitem(level);
    if(level == 1)		/* static variables */
	ar = ENV->AR;
    else 			/* local variable */	
	ar = ENV->LAR;

    sym->val.ival = i;

    ar[i].v.dval = 0; /* NULL it */
    ar[i].type = sym->type;
    return i;
}


/*------------------------------------------------*/
void inittoken(token_t * e1)
{
    e1->pflag = 0;
    e1->code.binst = e1->code.nextinst = 0;
    e1->typequal = e1->sclass = 0;
    e1->sym = NULL;
    e1->val.sym = NULL;
    e1->type = NULL;
}
void freetoken(token_t * e1)
{
    freetype(e1->type);
    e1->type = NULL;
}
void initcode(code_t * code)
{
    code->binst = code->nextinst = 0;
}


void cleancode(code_t * code)
{
    unsigned int i;
    inst_t *inst;

    if(!code)
	return;
    inst = code->inst;
    /* printf("Next instr: %d\n", code->nextinst); */
    /* rem free up other info also */
    for(i=0;i<code->nextinst;i++,inst++)
	if(inst->opcode == jmptab) {
	    eicstack_t *s;
	    s = inst->val.p.p;
	    xfree(s->val);
	    xfree(s);
	} else if(inst->opcode == assigntype)
	    freetype(inst->val.p.p);

}    

void killcode(symentry_t *sym)
{
    code_t * code;
    code = ENV->AR[sym->val.ival].v.p.p;
    if(!code)
	return;
    cleancode(code);
    xfree(code->inst);
    xfree(code);
}
void freecode(code_t * code)
{
    if(code && code->binst > 0) {
	xfree(code->inst);
	code->nextinst = code->binst = 0;
    }
}
void generate(code_t * code, int opcode,val_t *val,int ext)
{
    unsigned mod = 5;
    inst_t * inst;
    inst = code->inst;
    if(code->nextinst == code->binst)
	if(!(code->binst%mod)) {
	    if(!code->binst)
		inst = (inst_t*)xcalloc(1,sizeof(inst_t)*mod);
	    else
		inst = (inst_t*)
		    xrealloc(inst,(code->binst+mod)*sizeof(inst_t));
	    code->binst += mod;
	}
    code->inst = inst;
    inst[code->nextinst].opcode = opcode;
    inst[code->nextinst].val = *val;
    inst[code->nextinst].ext = ext;
    inst[code->nextinst].line = CurrentLineNo();
    code->nextinst++;
}
void copycode(code_t * c1, code_t * c2)
{
    unsigned int i;
    inst_t *inst;
    inst = c2->inst;
    for(i=0;i<c2->nextinst;++i,++inst) {
	generate(c1,inst->opcode,&inst->val,inst->ext);
	c1->inst[c1->nextinst-1].line = inst->line;
    }
}
void concode(code_t * c1, code_t * c2)
{
    copycode(c1,c2);
    freecode(c2);
}
void contoken(token_t * e1, token_t * e2)
{
    concode(&e1->code,&e2->code);
    e1->type = getcontype(e1->type,e2->type);
    freetoken(e2);
}

void swaptokens(token_t *e1, token_t * e2)
{
    token_t e3;
    e3 = *e2;
    *e2 = *e1;
    *e1 = e3;
}
    


