/* typesets.c
 *
 *	(C) Copyright Apr 15 1995, Edmond J. Breen.
 *		   ALL RIGHTS RESERVED.
 * This code may be copied for personal, non-profit use only.
 *
 */

#include "stdio.h"
#include "stdlib.h"

#include "MachSet.h"
#include "lexer.h"
#include "global.h"
#include "typemod.h"
#include "xalloc.h"
#include "typesets.h"
#include "symbol.h"
#include "error.h"

extern void binhint(), binhuint();
extern void binhlong(), binhulong(), binhdouble(), binhlval(), binhconst();

void (*BINFUN[]) () = {
    NULL, NULL,		/* t_error,   t_bool, */
    binhint, binhint,	/* t_char,    t_uchar, */
    binhint, binhint,	/* t_short,   t_ushort */
    binhint, binhint,	/* t_int,     t_enum, */
    binhuint,		/* t_uint, */
    
    binhlong, binhulong,  /* t_long,    t_ulong, */
    binhdouble, binhdouble, /* t_float,   t_double, */
    binhlval, NULL,	  /* t_pointer, t_void */
    NULL,NULL,            /* t_struct, t_union */
    NULL,binhlval,        /* t_lval,  t_array */       
    
};

typedef unsigned long set_t;
#define ismem(S,m) (((long)1 << (m)) & (S))
#define I(x)       ((long)1 << x)

#define SETNUM   I(t_char)+I(t_short)+I(t_int)+I(t_enum)+I(t_long)
#define SETUNUM  I(t_uchar)+I(t_ushort)+I(t_uint)+I(t_ulong)
#define SETFLOAT I(t_float)+I(t_double)

typedef struct {
    set_t Lset;
    set_t Rset;
} LRset_t;

LRset_t modLR[] =
{
{SETNUM + SETUNUM,
     SETNUM + SETUNUM,}
};

/*
 * two pointers cannot be added together
 */
LRset_t addLR[] =
{
{SETNUM + SETUNUM + SETFLOAT,
     SETNUM + SETUNUM + SETFLOAT},
{I(t_pointer) + I(t_array),
     SETNUM + SETUNUM},
{SETNUM + SETUNUM,
     I(t_pointer) + I(t_array)},
};

/*
 * integrals can be subtracted from a pointer
 * but not visa a versa
 */
LRset_t subLR[] =
{
{SETNUM + SETUNUM + SETFLOAT,
     SETNUM + SETUNUM + SETFLOAT},
{I(t_pointer) + I(t_array),
     I(t_pointer) + I(t_array) + SETNUM + SETUNUM},
};

LRset_t multLR[] =
{
{SETNUM + SETUNUM + SETFLOAT ,
     SETNUM + SETUNUM + SETFLOAT }};

LRset_t divLR[] =
{
{SETNUM + SETUNUM + SETFLOAT ,
     SETNUM + SETUNUM + SETFLOAT}};

LRset_t LTLR[] =
{
{I(t_pointer) + I(t_array) + SETNUM + SETUNUM + SETFLOAT ,
     I(t_pointer) + I(t_array) + SETNUM + SETUNUM + SETFLOAT}};

val_t VAL;
typedef struct {
unsigned oper;
LRset_t *LRset;
unsigned N;
} Binset;

Binset BINSET[] =
{
    {'+', addLR, sizeof(addLR) / sizeof(LRset_t),},
    {'-', subLR, sizeof(subLR) / sizeof(LRset_t),},
    {'*', multLR, sizeof(multLR) / sizeof(LRset_t),},
    {'/', divLR, sizeof(divLR) / sizeof(LRset_t),},
    {'%', modLR, sizeof(modLR) / sizeof(LRset_t),},
    {LSHT, modLR, sizeof(modLR) / sizeof(LRset_t),},
    {RSHT, modLR, sizeof(modLR) / sizeof(LRset_t),},
    {BOR, modLR, sizeof(modLR) / sizeof(LRset_t),},
    {XOR, modLR, sizeof(modLR) / sizeof(LRset_t),},
    {AND, modLR, sizeof(modLR) / sizeof(LRset_t),},
    {LT, LTLR, sizeof(LTLR) / sizeof(LRset_t),},
    {LE, LTLR, sizeof(LTLR) / sizeof(LRset_t),},
    {EQ, LTLR, sizeof(LTLR) / sizeof(LRset_t),},
    {NE, LTLR, sizeof(LTLR) / sizeof(LRset_t),},
    {GT, LTLR, sizeof(LTLR) / sizeof(LRset_t),},
    {GE, LTLR, sizeof(LTLR) / sizeof(LRset_t),},
    {LOR, LTLR, sizeof(LTLR) / sizeof(LRset_t),},
    {LAND, LTLR, sizeof(LTLR) / sizeof(LRset_t),},
};


void castvar(token_t * e1,
	     token_t * e2, int explicit)
{
    /* cast e1 into e2 */
    type_expr *t;
    if (gettype(e2->type) == t_func) {
	token_t e3;
	e3.type = copytype(nextType(e2->type));
	castvar(e1, &e3, explicit);
	freetype(e3.type);
	return;
    }
    if (gettype(e2->type) == t_lval) {
	t = e2->type;
	e2->type = nextType(e2->type);
	castvar(e1, e2, explicit);
	e2->type = t;
	return;
    }
    if(gettype(e2->type) == t_void) {
	e1->type = addtype(t_void,freetype(e1->type));
	return;
    }
        
    switch (gettype(e1->type)) {
      CASE_INT:
	switch (gettype(e2->type)) {
	  case t_uchar:
	    generate(&e1->code,int2uchar,&e1->val,0);
	    break;
	  case t_ushort:
	    generate(&e1->code,int2ushort,&e1->val,0);
	    break;
	  case t_char: case t_short: case t_int: case t_uint:
	    break;
	  CASE_LONG: CASE_ULONG:
	    if(sizeof(int) != sizeof(long))
		generate(&e1->code, int2long, &e1->val, 0);
	    break;
	  CASE_FLOAT:
	    generate(&e1->code, int2double, &e1->val, 0);
	    break;
	  case t_pointer:
	    generate(&e1->code, int2ptr, &e1->val, 0);
	    e1->type = addtype(t_pointer,e1->type);
	    if (!explicit) 
		warningerror("Suspicious pointer conversion");
	    return;
	    
	  default:
	    error("Illegal cast operation");
	}
	break;
      CASE_UINT:
	switch (gettype(e2->type)) {
	  CASE_INT: CASE_UINT:
	    break;
	  CASE_LONG: CASE_ULONG:
	    if(sizeof(int) != sizeof(long))
		generate(&e1->code, uint2long, &e1->val, 0);
	    break;
	  CASE_FLOAT:
	    generate(&e1->code, uint2double, &e1->val, 0);
	    break;
	  default:
	    error("Illegal cast operation");
	}
	break;
      CASE_LONG: CASE_ULONG:
	switch (gettype(e2->type)) {
	  CASE_INT: CASE_UINT:
	    if(sizeof(int) != sizeof(long))
		generate(&e1->code, long2int, &e1->val, 0);
	    break;
	  CASE_LONG: CASE_ULONG:
	    break;
	  CASE_FLOAT:
	    generate(&e1->code, long2double, &e1->val, 0);
	    break;
	  default:
	    error("Illegal cast operation");
	}
	break;
      CASE_FLOAT:
	switch (gettype(e2->type)) {
	  CASE_INT: CASE_UINT:
	    generate(&e1->code, double2int, &e1->val, 0);
	    break;
	  CASE_LONG: CASE_ULONG:
	    generate(&e1->code, double2long, &e1->val, 0);
	    break;
	  CASE_FLOAT:
	    /*
	    if(gettype(e1->type) == t_double
	       && gettype(e2->type) == t_float)
		generate(&e1->code, double2float, &e1->val, 0);
	    else
		return;
		break;
		*/
	    return;
	  default:
	    error("Illegal cast operation");
	}
	break;
	
      case t_pointer:
	switch (gettype(e2->type)) {
	  case t_uchar:
	  case t_char:
	    error("Illegal cast operation");
	    return;
	  case t_short:
	  case t_int:
	  case t_ushort:
	  case t_uint:
	    generate(&e1->code, ptr2int, &e1->val, 0);
	    break;
	  CASE_LONG:
	  case t_ulong:
	    generate(&e1->code, ptr2long, &e1->val, 0);
	    break;
	  CASE_FLOAT:error("Illegal cast operation");
	    return;
	  case t_pointer:
	    if (!sametypes(e1->type, e2->type)) {
		if (gettype(e2->type) != t_pointer)
		    error("Illegal cast operation");
		else if (!explicit) 
		    warningerror("Suspicious pointer conversion");
	    } else if(!explicit) 
		if(ConstIntegrity(nextType(e2->type),nextType(e1->type)))
		    error("Cast loses const qualifier");
	    break;
	  default:
	    error("Illegal cast operation");
	    return;
	}
	freetype(e1->type);
	e1->type = copytype(e2->type);
	return;
      case t_array:
	if (gettype(e2->type) == t_pointer) {
	    exchtype(t_pointer, e1->type);
	    if (!sametypes(e1->type, e2->type) && !explicit)
		warningerror("Suspicious pointer conversion");
	} else if(gettype(e2->type) == t_array) {
	    if (!sametypes(e1->type, e2->type))
		error("Illegal cast operation");
	} else
	    error("Illegal cast operation");
	return;
      case t_lval:
	e1->type = succType(e1->type);
	castvar(e1, e2, explicit);
	setConst(e1->type);
	return;
      case t_union:
      case t_struct:
	if (gettype(e2->type) == t_lval) {
	    if (!sametypes(e1->type, nextType(e2->type)))
		error("Illegal cast operation");
	} else if (!sametypes(e1->type, e2->type))
	    error("Illegal cast operation");
	return;
      default:
	error("Illegal cast operation");
    }
    set_bastype(bastype(e2->type), e1->type);
}

void castconst(token_t * e1, token_t * e2, int explicit)
{
    /*
     *  cast e1 type into an e2 type and
     *  where e1 is a constant
     */
    unsigned t;
    if (gettype(e2->type) == t_func) {
	token_t e3;
	e3.type = copytype(nextType(e2->type));
	castconst(e1, &e3, explicit);
	freetype(e3.type);
	return;
    }
    if (e1->pflag) {
	castvar(e1, e2, explicit);
	return;
    }
    if(gettype(e2->type) == t_void) {
	e1->type = addtype(t_void,freetype(e1->type));
	setConst(e1->type);
	return;
    }
	
    switch ((t = gettype(e2->type))) {
      CASE_INT: CASE_UINT:
	switch (gettype(e1->type)) {
	  CASE_INT: CASE_UINT: break;
	  CASE_LONG:e1->val.ival = (int) e1->val.lval;   break;
	  CASE_ULONG:e1->val.ival = (int) e1->val.ulval; break;
	  CASE_FLOAT:e1->val.ival = (int) e1->val.dval;
	    break;
	  default:
	    error("Illegal cast operation");
	}
	break;
      CASE_LONG:
	switch (gettype(e1->type)) {
	  CASE_INT:e1->val.lval = (long) e1->val.ival;
	    break;
	  CASE_UINT:e1->val.lval = (long) e1->val.uival;
	    break;
	  CASE_LONG:return;
	  case t_ulong:
	    break;
	  CASE_FLOAT:e1->val.lval = (long) e1->val.dval;
	    break;
	  default:
	    error("Illegal cast operation");
	}
	break;
      case t_ulong:
	switch (gettype(e1->type)) {
	  CASE_INT:e1->val.ulval = (unsigned long) e1->val.ival;
	    break;
	  CASE_UINT:e1->val.ulval = (unsigned long) e1->val.uival;
	    break;
	  CASE_LONG:break;
	  case t_ulong:
	    return;
	  CASE_FLOAT:e1->val.ulval = (unsigned long) e1->val.dval;
	    break;
	  default:
	    error("Illegal cast operation");
	}
	break;
      CASE_FLOAT:
	switch (gettype(e1->type)) {
	  CASE_INT:e1->val.dval = (double) e1->val.ival;
	    break;
	  CASE_UINT:e1->val.dval = (double) e1->val.uival;
	    break;
	  CASE_LONG:e1->val.dval = (double) e1->val.lval;
	    break;
	  case t_ulong:
	    e1->val.dval = (double) e1->val.ulval;
	    break;
	  case t_pointer:
	    error("Illegal floating point operation");
	    break;
	  CASE_FLOAT:return;
	}
	break;
      case t_lval:
	e2->type = succType(e2->type);
	castconst(e1, e2, explicit);
	e2->type = addtype(t, e2->type);
	return;
      case t_pointer:
	if (gettype(e1->type) != t_pointer && !explicit
	    && e1->val.ival != 0)
	    warningerror("Suspicious pointer conversion");
	switch (gettype(e1->type)) {
	  CASE_INT:e1->val.p.p =  (void *) e1->val.ival;
	    break;
	  CASE_UINT:e1->val.p.p = (void *) e1->val.uival;
	    break;
	  CASE_LONG:
	  case t_ulong:
	    e1->val.p.p = (void *) e1->val.lval;
	    break;
	  case t_pointer:
	    if (!sametypes(e2->type,e1->type))
		warningerror("Suspicious pointer conversion");
	    e1->type = freetype(e1->type);
	    e1->type = copytype(e2->type);
	    setConst(e1->type);
	    return;
	  default:
	    error("Illegal cast operation");
	    break;
	}
	exchtype(t_pointer, e1->type);
	/*if(!explicit) */
	setConst(e1->type);
	return;
      case t_union:
      case t_struct:
	if (gettype(e2->type) == t_lval) {
	    if (!sametypes(e1->type, nextType(e2->type)))
		error("Illegal cast operation");
	} else if (!sametypes(e1->type, e2->type))
	    error("Illegal cast operation");
	return;
      default:
	error("Illegal cast operation");
    }
    set_bastype(bastype(e2->type), e1->type);
}


void do_lor(token_t *e1, int n)
{
    int inst;
    val_t u1; 
    output(e1);
    switch(gettype(e1->type)) {
      CASE_INT:
      CASE_UINT: inst = jmpTint;  break;
      CASE_LONG:
      CASE_ULONG: inst = jmpTlng; break;
      CASE_FLOAT: inst = jmpTdbl; break;
      case t_pointer: inst = jmpTptr;break;
      default:
	error("syntax error near '||'");
	return;
    }
    exchtype(t_int,e1->type);
    u1.ival = n;
    generate(&e1->code, inst, &u1, 0);
}
void do_land(token_t *e1, int n)
{
    int inst;
    val_t u1; 
    output(e1);
    switch(gettype(e1->type)) {
      CASE_INT:
      CASE_UINT: inst = jmpFint;  break;
      CASE_LONG:
      CASE_ULONG: inst = jmpFlng; break;
      CASE_FLOAT: inst = jmpFdbl; break;
      case t_pointer: inst = jmpFptr;break;
      default:
	error("syntax error associated near '&&'");
	return;
    }
    exchtype(t_int,e1->type);
    u1.ival = n;
    generate(&e1->code, inst, &u1, 0);
}

static void checkint(unsigned obj, token_t * e1, token_t * e2)
{
    if (gettype(e1->type) != obj) {
	if(!isconst(e1->type))
	    exchtype(obj, e1->type);
	else
	    castconst(e1,e2,0);
    }
    if (gettype(e2->type) != obj) {
	if(!isconst(e1->type))
	    exchtype(obj, e2->type);
	else
	    castconst(e2, e1, 0);
    }
    if (!e2->pflag)
	output(e2);
}


void binhint(unsigned oper, token_t * e1, token_t * e2)
{
    checkint(t_int, e1, e2);
    switch (oper) {
      case '+':
	generate(&e2->code, addint, &e2->val, 0);
	break;
      case '-':
	generate(&e2->code, subint, &e2->val, 0);
	break;
      case '*':
	generate(&e2->code, multint, &e2->val, 0);
	break;
      case '/':
	generate(&e2->code, divint, &e2->val, 0);
	break;
      case '%':
	generate(&e2->code, modint, &e2->val, 0);
	break;
      case LSHT:
	generate(&e2->code, lshtint, &e2->val, 0);
	break;
      case RSHT:
	generate(&e2->code, rshtint, &e2->val, 0);
	break;
      case LT:
	generate(&e2->code, ltint, &e2->val, 0);
	break;
      case LE:
	generate(&e2->code, leint, &e2->val, 0);
	break;
      case EQ:
	generate(&e2->code, eqint, &e2->val, 0);
	break;
      case NE:
	generate(&e2->code, neint, &e2->val, 0);
	break;
      case GT:
	generate(&e2->code, gtint, &e2->val, 0);
	break;
      case GE:
	generate(&e2->code, geint, &e2->val, 0);
	break;
      case BOR:
	generate(&e2->code, borint, &e2->val, 0);
	break;
      case XOR:
	generate(&e2->code, xorint, &e2->val, 0);
	break;
      case AND:
	generate(&e2->code, andint, &e2->val, 0);
	break;
    }
}

void binhuint(unsigned oper, token_t * e1, token_t * e2)
{
    checkint(t_uint, e1, e2);
    switch (oper) {
      case '+':
	generate(&e2->code, adduint, &e2->val, 0);
	break;
      case '-':
	generate(&e2->code, subuint, &e2->val, 0);
	break;
      case '*':
	generate(&e2->code, multuint, &e2->val, 0);
	break;
      case '/':
	generate(&e2->code, divuint, &e2->val, 0);
	break;
      case '%':
	generate(&e2->code, moduint, &e2->val, 0);
	break;
      case LSHT:
	generate(&e2->code, lshtuint, &e2->val, 0);
	break;
      case RSHT:
	generate(&e2->code, rshtuint, &e2->val, 0);
	break;
      case BOR:
	generate(&e2->code, boruint, &e2->val, 0);
	break;
      case XOR:
	generate(&e2->code, xoruint, &e2->val, 0);
	break;
      case AND:
	generate(&e2->code, anduint, &e2->val, 0);
	break;
      default:
	output(e1);
	exchtype(t_int, e1->type);	/* should really be boolean */
	switch (oper) {
	  case LT:
	    generate(&e2->code, ltuint, &e2->val, 0);
	    break;
	  case LE:
	    generate(&e2->code, leuint, &e2->val, 0);
	    break;
	  case EQ:
	    generate(&e2->code, equint, &e2->val, 0);
	    break;
	  case NE:
	    generate(&e2->code, neuint, &e2->val, 0);
	    break;
	  case GT:
	    generate(&e2->code, gtuint, &e2->val, 0);
	    break;
	  case GE:
	    generate(&e2->code, geuint, &e2->val, 0);
	    break;
	}
    }
}

static void checklong(unsigned obj, token_t * e1, token_t * e2)
{
    int t;
    if ((t=gettype(e1->type)) != obj) {
	if (!isconst(e1->type)) {
	    if (!e1->pflag)
		output(e1);
	    if(sizeof(int) != sizeof(long))
		switch (t) {
		CASE_INT:generate(&e1->code, int2long, &e1->val, 0);
		break;
		CASE_UINT:generate(&e1->code, uint2long, &e1->val, 0);
		break;
		}
	    exchtype(obj, e1->type);
	} else 
	    castconst(e1, e2, 0);
    }
    if ((t=gettype(e2->type)) != obj) {
	if (!isconst(e2->type)) {
	    if (!e2->pflag)
		output(e2);
	    if(sizeof(int) != sizeof(long))
		switch (t) {
		CASE_INT:generate(&e2->code, int2long, &e1->val, 0);
		break;
		CASE_UINT:generate(&e2->code, uint2long, &e1->val, 0);
		break;
		}
	    exchtype(obj, e2->type);
	} else 
	    castconst(e2, e1, 0);
    }
    if (!e2->pflag)
	output(e2);
}

void binhlong(unsigned oper, token_t * e1, token_t * e2)
{
    checklong(t_long, e1, e2);
    
    switch (oper) {
      case '+':
	generate(&e2->code, addlong, &e2->val, 0);
	break;
      case '-':
	generate(&e2->code, sublong, &e2->val, 0);
	break;
      case '*':
	generate(&e2->code, multlong, &e2->val, 0);
	break;
      case '/':
	generate(&e2->code, divlong, &e2->val, 0);
	break;
      case '%':
	generate(&e2->code, modlong, &e2->val, 0);
	break;
      case LSHT:
	generate(&e2->code, lshtlong, &e2->val, 0);
	break;
      case RSHT:
	generate(&e2->code, rshtlong, &e2->val, 0);
	break;
      case BOR:
	generate(&e2->code, borlong, &e2->val, 0);
	break;
      case XOR:
	generate(&e2->code, xorlong, &e2->val, 0);
	break;
      case AND:
	generate(&e2->code, andlong, &e2->val, 0);
	break;
      default:
	output(e1);
	exchtype(t_int, e1->type);	/* should really be boolean */
	switch (oper) {
	  case LT:
	    generate(&e2->code, ltlong, &e2->val, 0);
	    break;
	  case LE:
	    generate(&e2->code, lelong, &e2->val, 0);
	    break;
	  case EQ:
	    generate(&e2->code, eqlong, &e2->val, 0);
	    break;
	  case NE:
	    generate(&e2->code, nelong, &e2->val, 0);
	    break;
	  case GT:
	    generate(&e2->code, gtlong, &e2->val, 0);
	    break;
	  case GE:
	    generate(&e2->code, gelong, &e2->val, 0);
	    break;
	}
	break;
    }
}

void binhulong(unsigned oper, token_t * e1, token_t * e2)
{
    checklong(t_ulong, e1, e2);
    switch (oper) {
      case '+':
	generate(&e2->code, addulong, &e2->val, 0);
	break;
      case '-':
	generate(&e2->code, subulong, &e2->val, 0);
	break;
      case '*':
	generate(&e2->code, multulong, &e2->val, 0);
	break;
      case '/':
	generate(&e2->code, divulong, &e2->val, 0);
	break;
      case '%':
	generate(&e2->code, modulong, &e2->val, 0);
	break;
      case LSHT:
	generate(&e2->code, lshtulong, &e2->val, 0);
	break;
      case RSHT:
	generate(&e2->code, rshtulong, &e2->val, 0);
	break;
      case BOR:
	generate(&e2->code, borulong, &e2->val, 0);
	break;
      case XOR:
	generate(&e2->code, xorulong, &e2->val, 0);
	break;
      case AND:
	generate(&e2->code, andulong, &e2->val, 0);
	break;
      default:
	output(e1);
	exchtype(t_int, e1->type);	/* should really be boolean */
	switch (oper) {
	  case LT:
	    generate(&e2->code, ltulong, &e2->val, 0);
	    break;
	  case LE:
	    generate(&e2->code, leulong, &e2->val, 0);
	    break;
	  case EQ:
	    generate(&e2->code, equlong, &e2->val, 0);
	    break;
	  case NE:
	    generate(&e2->code, neulong, &e2->val, 0);
	    break;
	  case GT:
	    generate(&e2->code, gtulong, &e2->val, 0);
	    break;
	  case GE:
	    generate(&e2->code, geulong, &e2->val, 0);
	    break;
	}
	break;
    }
}

void binhdouble(unsigned oper, token_t * e1, token_t * e2)
{
    int t;
    if ((t=gettype(e1->type)) != t_double) {
	if (!isconst(e1->type)) {
	    if (!e1->pflag)
		output(e1);
	    switch (t) {
	      CASE_INT:generate(&e1->code, int2double, &e1->val, 0);
		break;
	      CASE_UINT:generate(&e1->code, uint2double, &e1->val, 0);
		break;
	      CASE_LONG:generate(&e1->code, long2double, &e1->val, 0);
		break;
	      CASE_ULONG:generate(&e1->code, ulong2double, &e1->val, 0);
		break;
	    }
	    exchtype(t_double, e1->type);
	} else 
	    castconst(e1, e2, 0);
    }
    if ((t=gettype(e2->type)) != t_double) {
	if (!isconst(e2->type)) {
	    if (!e2->pflag)
		output(e2);
	    switch (t) {
	      CASE_INT:generate(&e2->code, int2double, &e1->val, 0);
		break;
	      CASE_UINT:generate(&e2->code, uint2double, &e1->val, 0);
		break;
	      CASE_LONG:generate(&e2->code, long2double, &e1->val, 0);
		break;
	      CASE_ULONG:generate(&e2->code, ulong2double, &e1->val, 0);
		break;
	    }
	    exchtype(t_double, e2->type);
	} else
	    castconst(e2, e1, 0);
    }
    if (!e2->pflag)
	output(e2);
    
    switch (oper) {
      case '+':
	generate(&e2->code, adddouble, &e2->val, 0);
	break;
      case '-':
	generate(&e2->code, subdouble, &e2->val, 0);
	break;
      case '*':
	generate(&e2->code, multdouble, &e2->val, 0);
	break;
      case '/':
	generate(&e2->code, divdouble, &e2->val, 0);
	break;
      default:
	output(e1);
	exchtype(t_int, e1->type);	/* should really be boolean */
	switch (oper) {
	  case LT:
	    generate(&e2->code, ltdouble, &e2->val, 0);
	    break;
	  case LE:
	    generate(&e2->code, ledouble, &e2->val, 0);
	    break;
	  case EQ:
	    generate(&e2->code, eqdouble, &e2->val, 0);
	    break;
	  case NE:
	    generate(&e2->code, nedouble, &e2->val, 0);
	    break;
	  case GT:
	    generate(&e2->code, gtdouble, &e2->val, 0);
	    break;
	  case GE:
	    generate(&e2->code, gedouble, &e2->val, 0);
	    break;
	  default:error("illegal binary operation to `%c'", oper);
	}
    }
}

#define arrayLimit(type)    /* replaced by runtime checks */

/*if(limit && (e2->val.type >= limit || \
				      e2->val.type < 0))\
                               error("Array bound violation");
			       */
#define ConstPval(y) { if(oper == '+')\
			    e1->val.p.p = (char*)e1->val.p.p + e2->val.y;\
			else\
			    e1->val.p.p = (char*)e1->val.p.p - e2->val.y;\
		    }

void binhlval(unsigned oper, token_t * e1, token_t * e2)
{
    unsigned limit = 0;
    if (gettype(e1->type) != t_pointer) {
	if(gettype(e1->type) != t_array) {
	    binhlval(oper,e2,e1);
	    swaptokens(e1,e2);
	    return;
	} else { /* convert array to pointer */
	    limit = getNumElems(e1->type);
	    output(e1);
	}
    }
    if(gettype(e2->type) == t_array)
	output(e2);
       
    if (gettype(e2->type) != t_pointer) {
	if(oper == '+' || oper == '-') {
	    int c = 0;
	    VAL.ival = get_sizeof(nextType(e1->type));
	    if(!VAL.ival)
		error("Illegal pointer operation");
	    if(isconst(e1->type) && isconst(e2->type))
		c = 1;
	    if(isconst(e2->type)) {
		switch (gettype(e2->type)) {
		  CASE_INT:
		    arrayLimit(ival);
		    e2->val.ival *= VAL.ival;
		    if(c)
			ConstPval(ival);
		    break;
		  CASE_UINT:
		    arrayLimit(uival);
		    e2->val.uival *= VAL.ival;
		    if(c)
			ConstPval(uival);
		    break;
		  CASE_LONG:
		    arrayLimit(lval);
		    e2->val.lval *= VAL.ival;
		    if(c)
			ConstPval(lval);
		    break;
		  CASE_ULONG:
		    arrayLimit(ulval);
		    e2->val.ulval *= VAL.ival;
		    if(c)
			ConstPval(ulval);
		    break;
		  default:
		    error("Illegal pointer operation");
		}
		VAL.ival = 1; 
	    }
	    if(!c) {
		if(oper == '-')
		    VAL.ival = -VAL.ival;
		output(e2);
		if(gettype(e2->type) > t_uint && sizeof(int) != sizeof(long)) {
		    token_t e3;
		    inittoken(&e3);
		    e3.type = addtype(t_int,NULL);
		    castvar(e2,&e3,0);
		    freetype(e3.type);
		}
		if(VAL.ival > 1 || VAL.ival < 0)
		    generate(&e2->code, ixa, &VAL, limit);
		else
		    generate(&e2->code, addptr2int, &e2->val, 0);
	    }
	    return;
	}
	if (isconst(e2->type))
	    castconst(e2, e1, 0);
	else
	    castvar(e2, e1, 0);
	output(e2);
	    
    } else if (!compatibletypes(e1->type, e2->type))
	error("Mixed pointer operation");

    output(e2);
    
    switch (oper) {
      case '-':
	generate(&e2->code, subptr, &e2->val, 0);
	VAL.ival = get_sizeof(nextType(e1->type));
	if (VAL.ival > 1) {
	    int size = VAL.ival;
	    VAL.ival = 1;
	    generate(&e2->code, bump, &VAL, 0);
	    VAL.ival = size;
	    generate(&e2->code, pushint, &VAL, 0);
	    generate(&e2->code, divint, &VAL, 0);
	}
	break;
      case LT:
	generate(&e2->code, ltptr, &e2->val, 0);
	break;
      case LE:
	generate(&e2->code, leptr, &e2->val, 0);
	break;
      case EQ:
	generate(&e2->code, eqptr, &e2->val, 0);
	break;
      case NE:
	generate(&e2->code, neptr, &e2->val, 0);
	break;
      case GT:
	generate(&e2->code, gtptr, &e2->val, 0);
	break;
      case GE:
	generate(&e2->code, geptr, &e2->val, 0);
	break;
    }
    output(e1);
    freetype(e1->type);
    e1->type = addtype(t_int, NULL);
} 
#undef ConstPval

void (*compatable(LRset_t S[],
		  unsigned n, unsigned left,
		  unsigned right)) ()
{
    static int i, m;

    if(left > 32 || right > 32)
	return NULL;
    
    for (i = 0; i < n; i++)
	if (ismem(S[i].Lset, left) && ismem(S[i].Rset, right)) {
	    m = left > right ? left : right;
	    return  BINFUN[m];
	}
    return NULL;
}

int bin_validate(unsigned oper, token_t * e1, token_t * e2)
{
    int i, left, right;
    void (*f) (int op, token_t * e1, token_t * e2);

    if(sizeof(int) == sizeof(long)) {
	if(gettype(e1->type) == t_uint)
	    exchtype(t_ulong,e1->type);
	if(gettype(e2->type) == t_uint)
	    exchtype(t_ulong,e2->type);
    }
	
    
    if(isconst(e1->type) || isconst(e2->type)) {
	binhconst(oper,e1,e2);
	return 1;
    }
    
    left = gettype(e1->type);
    right = gettype(e2->type);
    
    for (i = 0; i < sizeof(BINSET) / sizeof(Binset); i++)
	if (oper == BINSET[i].oper) {
	    f = compatable(BINSET[i].LRset,
			   BINSET[i].N,
			   left, right);
	    if (f != NULL) {
		(*f) (oper, e1, e2);
		return 1;
	    } else
		error("Incompatible types");
	}
    return 0;
}

void cast2comm(token_t * e1, token_t * e2)
{
    /* cast e1 and e2 to a common type */
    
    unsigned t1, t2;
    t1 = gettype(e1->type);
    t2 = gettype(e2->type);


    if(isconst(e1->type)) {
	if(t1 > t2) {
	    if(isconst(e2->type))
	       castconst(e2, e1, 0);
	    else
		castvar(e2, e1, 0);
	} else
	    castconst(e1, e2, 0);
	return;
    }	
    
    switch (t1) {
      CASE_NUM:
	if(t2 == t_array || t2 == t_pointer ||
	   t2 == t_pointer)
	    error("Illegal cast operation");
	else {
	    if(t1 > t2) {
		if(isconst(e2->type))
		    castconst(e2,e1,0);
		else 
		    castvar(e2,e1,0);
	    } else
		castvar(e1,e2,0);
	}
	break;
      case t_pointer:
	if (!sametypes(e1->type, e2->type)) {
	    if (isconst(e2->type))
		castconst(e2, e1, 0);
	}
	break;
      case t_array:
	if (gettype(e2->type) == t_pointer) {
	    if (!sametypes(nextType(e1->type), nextType(e2->type)))
		warningerror("Suspicious pointer conversion");
	    exchtype(t_pointer, e1->type);
	} else
	    error("Illegal cast operation");
	break;
      case t_union:
      case t_struct:
	if(sametypes(e1->type,e2->type))
	    break;
      default:
	error("Illegal cast operation");
    }
}


int unaryop(token_t * e1, int op)
{
    void derefConst(token_t * e1);
    static int t;
    
    if (!isconst(e1->type) && op != INC && op != DEC)
	output(e1);

    if(isconst(e1->type)) {
	switch(op) {
	  case '-':
	    switch (gettype(e1->type)) {
	      CASE_INT:
		e1->val.ival = -e1->val.ival;
		break;
	      CASE_UINT:
		e1->val.uival = -e1->val.uival;
		break;
	      CASE_LONG:
		e1->val.lval = -e1->val.lval;
		break;
	      case t_ulong:
		e1->val.lval = -e1->val.ulval;
		break;
	      CASE_FLOAT:
		e1->val.dval = -e1->val.dval;
		break;
	      default:
		error("Illegal operand");
	    }
	    break;
	  case '~':
	    switch (gettype(e1->type)) {
	      CASE_INT: CASE_UINT:
		e1->val.ival = ~e1->val.ival;
		break;
	      CASE_LONG:
		e1->val.lval = ~e1->val.lval;
		break;
	      case t_ulong:
		e1->val.lval = ~e1->val.ulval;
		break;
	      default:
		error("Illegal operand");
	    }
	    break;
	  case NOT:
	    switch (gettype(e1->type)) {
	      CASE_INT: CASE_UINT:
		e1->val.ival = !e1->val.ival;
		break;
	      CASE_LONG:
	      case t_ulong:
		e1->val.ival = !e1->val.lval;
		break;
	      CASE_FLOAT:e1->val.ival = !e1->val.dval;
		break;
	      case t_pointer:
		e1->val.ival = !e1->val.p.p;
		break;
	      default:
		error("Illegal operand");
	    }
	    e1->type = addtype(t_int, freetype(e1->type));
	    setConst(e1->type);
	    return 1;
	  case '*':
	    derefConst(e1);
	    break;
	  default:
	    error("Illegal unary opertion");
	}
	return 1;
    }

    switch (op) {
      case '+':
	break;
      case '-':
	switch ((t =gettype(e1->type))) {
	  case t_char:
	  case t_uchar:
	  case t_short:
	  case t_ushort:
	  case t_int:
	    if(t != t_int)
		exchtype(t_int,e1->type);
	    generate(&e1->code, negint, &e1->val, 0);
	    break;
	  case t_uint:
	    generate(&e1->code, neguint, &e1->val, 0);
	    break;
	  CASE_LONG:
	    generate(&e1->code, neglong, &e1->val, 0);
	    break;
	  case t_ulong:
	    generate(&e1->code, negulong, &e1->val, 0);
	    break;
	  CASE_FLOAT:
	    generate(&e1->code, negdouble, &e1->val, 0);
	    break;
	  default:
	    error("Illegal operand");
	    
	}
	break;
      case '~':
	switch ((t=gettype(e1->type))) {
	  CASE_INT: CASE_UINT:
	    if(t < t_int)
		exchtype(t_int,e1->type);
	    generate(&e1->code, compint, &e1->val, 0);
	    break;
	  CASE_LONG:
	  case t_ulong:
	    generate(&e1->code, complong, &e1->val, 0);
	    break;
	  default:
	    error("Illegal operand");
	}
	break;
      case NOT:
	switch (gettype(e1->type)) {
	  CASE_INT: CASE_UINT:
	    generate(&e1->code, notint, &e1->val, 0);
	    break;
	  CASE_LONG:
	  case t_ulong:
	    generate(&e1->code, notlong, &e1->val, 0);
	    break;
	  CASE_FLOAT:generate(&e1->code, notdouble, &e1->val, 0);
	    break;
	  case t_lval:
	    t = gettype(e1->type);
	    e1->type = succType(e1->type);
	    unaryop(e1, op);
	    e1->type = addtype(t, e1->type);
	    break;
	  case t_pointer:
	    generate(&e1->code, notptr, &e1->val, 0);
	    break;
	  default:
	    error("Illegal operand");
	}
	e1->type = addtype(t_int, freetype(e1->type));
	break;
      case INC:
      case DEC:
	do_inc_dec(e1, op);
	e1->pflag = 0;
	do_stooutput(e1);
	break;
      case '*':
	if((t = gettype(e1->type)) == t_pointer || t == t_array) {
	    e1->pflag = 0;
	    exchtype(t_lval, e1->type);
	    if(nextType(e1->type) && (!isconstp(nextType(e1->type)) ||
			 !isconst(nextType(e1->type))))
		unsetConst(e1->type);
	} else if(t != t_lval)
	    error("Must have pointer");
	return 1;
      default:
	error(" Invalid unary assignment");
	break;
    }
    return 1;
}

int do_binaryop(token_t * e1, token_t * e2, int op)
{
    int t1 = gettype(e1->type);
    int t2 = gettype(e2->type);
    if(t1 == t_pointer || t2 == t_pointer) {
	if(!(op >= LT && op <= GE)) {
	    binhlval(op,e1,e2);
	    return 1;
	}
    }

    if (t1 > t2)
	castconst(e2, e1, 0);
    else 
	castconst(e1, e2, 0);
    
    switch (gettype(e1->type)) {
      CASE_FLOAT:
	switch (op) {
	  case '*':
	    e1->val.dval *= e2->val.dval;
	    break;
	  case '/':
	    e1->val.dval /= e2->val.dval;
	    break;
	  case '+':
	    e1->val.dval += e2->val.dval;
	    break;
	  case '-':
	    e1->val.dval -= e2->val.dval;
	    break;
	  default:
	    set_bastype(t_int, e1->type);
	    switch (op) {
	      case LT:
		e1->val.ival = e1->val.dval < e2->val.dval;
		break;
	      case LE:
		e1->val.ival = e1->val.dval <= e2->val.dval;
		break;
	      case EQ:
		e1->val.ival = e1->val.dval == e2->val.dval;
		break;
	      case NE:
		e1->val.ival = e1->val.dval != e2->val.dval;
		break;
	      case GT:
		e1->val.ival = e1->val.dval > e2->val.dval;
		break;
	      case GE:
		e1->val.ival = e1->val.dval >= e2->val.dval;
		break;
	      default: error("illegal binary operation to `%c'", op);
	    }
	    break;
	}
	break;
      CASE_INT:
	switch (op) {
	  case '*':
	    e1->val.ival *= e2->val.ival;
	    break;
	  case '/':
	    e1->val.ival /= e2->val.ival;
	    break;
	  case '+':
	    e1->val.ival += e2->val.ival;
	    break;
	  case '-':
	    e1->val.ival -= e2->val.ival;
	    break;
	  case '%':
	    e1->val.ival %= e2->val.ival;
	    break;
	  case LSHT:
	    e1->val.ival <<= e2->val.ival;
	    break;
	  case RSHT:
	    e1->val.ival >>= e2->val.ival;
	    break;
	  case AND:
	    e1->val.ival = e1->val.ival & e2->val.ival;
	    break;
	  case BOR:
	    e1->val.ival = e1->val.ival | e2->val.ival;
	    break;
	  case XOR:
	    e1->val.ival = e1->val.ival ^ e2->val.ival;
	    break;
	  case LT:
	    e1->val.ival = e1->val.ival < e2->val.ival;
	    break;
	  case LE:
	    e1->val.ival = e1->val.ival <= e2->val.ival;
	    break;
	  case EQ:
	    e1->val.ival = e1->val.ival == e2->val.ival;
	    break;
	  case NE:
	    e1->val.ival = e1->val.ival != e2->val.ival;
	    break;
	  case GT:
	    e1->val.ival = e1->val.ival > e2->val.ival;
	    break;
	  case GE:
	    e1->val.ival = e1->val.ival >= e2->val.ival;
	    break;
	}
	break;
      CASE_UINT:
	switch (op) {
	  case '*':
	    e1->val.uival *= e2->val.uival;
	    break;
	  case '/':
	    e1->val.uival /= e2->val.uival;
	    break;
	  case '+':
	    e1->val.uival += e2->val.uival;
	    break;
	  case '-':
	    e1->val.uival -= e2->val.uival;
	    break;
	  case '%':
	    e1->val.uival %= e2->val.uival;
	    break;
	  case LSHT:
	    e1->val.uival <<= e2->val.uival;
	    break;
	  case RSHT:
	    e1->val.uival >>= e2->val.uival;
	    break;
	  case AND:
	    e1->val.uival = e1->val.uival & e2->val.uival;
	    break;
	  case BOR:
	    e1->val.uival = e1->val.uival | e2->val.uival;
	    break;
	  case XOR:
	    e1->val.uival = e1->val.uival ^ e2->val.uival;
	    break;
	  default:
	    set_bastype(t_int, e1->type);
	    switch (op) {
	      case LT:
		e1->val.ival = e1->val.uival < e2->val.uival;
		break;
	      case LE:
		e1->val.ival = e1->val.uival <= e2->val.uival;
		break;
	      case EQ:
		e1->val.ival = e1->val.uival == e2->val.uival;
		break;
	      case NE:
		e1->val.ival = e1->val.uival != e2->val.uival;
		break;
	      case GT:
		e1->val.ival = e1->val.uival > e2->val.uival;
		break;
	      case GE:
		e1->val.ival = e1->val.uival >= e2->val.uival;
		break;
	    }
	    break;
	}
	break;
      CASE_LONG:
	switch (op) {
	  case '*':
	    e1->val.lval *= e2->val.lval;
	    break;
	  case '/':
	    e1->val.lval /= e2->val.lval;
	    break;
	  case '+':
	    e1->val.lval += e2->val.lval;
	    break;
	  case '-':
	    e1->val.lval -= e2->val.lval;
	    break;
	  case '%':
	    e1->val.lval %= e2->val.lval;
	    break;
	  case LSHT:
	    e1->val.lval <<= e2->val.lval;
	    break;
	  case RSHT:
	    e1->val.lval >>= e2->val.lval;
	    break;
	  case AND:
	    e1->val.lval = e1->val.lval & e2->val.lval;
	    break;
	  case BOR:
	    e1->val.lval = e1->val.lval | e2->val.lval;
	    break;
	  case XOR:
	    e1->val.lval = e1->val.lval ^ e2->val.lval;
	    break;
	  default:
	    set_bastype(t_int, e1->type);
	    switch (op) {
	      case LT:
		e1->val.ival = e1->val.lval < e2->val.lval;
		break;
	      case LE:
		e1->val.ival = e1->val.lval <= e2->val.lval;
		break;
	      case EQ:
		e1->val.ival = e1->val.lval == e2->val.lval;
		break;
	      case NE:
		e1->val.ival = e1->val.lval != e2->val.lval;
		break;
	      case GT:
		e1->val.ival = e1->val.lval > e2->val.lval;
		break;
	      case GE:
		e1->val.ival = e1->val.lval >= e2->val.lval;
		break;
	    }
	    break;
	}
	break;
      case t_ulong:
	switch (op) {
	  case '*':
	    e1->val.ulval *= e2->val.ulval;
	    break;
	  case '/':
	    e1->val.ulval /= e2->val.ulval;
	    break;
	  case '+':
	    e1->val.ulval += e2->val.ulval;
	    break;
	  case '-':
	    e1->val.ulval -= e2->val.ulval;
	    break;
	  case '%':
	    e1->val.ulval %= e2->val.ulval;
	    break;
	  case LSHT:
	    e1->val.ulval <<= e2->val.ulval;
	    break;
	  case RSHT:
	    e1->val.ulval >>= e2->val.ulval;
	    break;
	  case AND:
	    e1->val.ulval = e1->val.ulval & e2->val.ulval;
	    break;
	  case BOR:
	    e1->val.ulval = e1->val.ulval | e2->val.ulval;
	    break;
	  case XOR:
	    e1->val.ulval = e1->val.ulval ^ e2->val.ulval;
	    break;
	  default:
	    set_bastype(t_int, e1->type);
	    switch (op) {
	      case LT:
		e1->val.ival = e1->val.ulval < e2->val.ulval;
		break;
	      case LE:
		e1->val.ival = e1->val.ulval <= e2->val.ulval;
		break;
	      case EQ:
		e1->val.ival = e1->val.ulval == e2->val.ulval;
		break;
	      case NE:
		e1->val.ival = e1->val.ulval != e2->val.ulval;
		break;
	      case GT:
		e1->val.ival = e1->val.ulval > e2->val.ulval;
		break;
	      case GE:
		e1->val.ival = e1->val.ulval >= e2->val.ulval;
		break;
	    }
	    break;
	}
	break;
      case t_pointer:
	switch (op) {
	  case LT:
	    e1->val.ival = e1->val.p.p < e2->val.p.p;
	    break;
	  case LE:
	    e1->val.ival = e1->val.p.p <= e2->val.p.p;
	    break;
	  case EQ:
	    e1->val.ival = e1->val.p.p == e2->val.p.p;
	    break;
	  case NE:
	    e1->val.ival = e1->val.p.p != e2->val.p.p;
	    break;
	  case GT:
	    e1->val.ival = e1->val.p.p > e2->val.p.p;
	    break;
	  case GE:
	    e1->val.ival = e1->val.p.p >= e2->val.p.p;
	    break;
	  default:
	    error("Illegal constant pointer operation");
	}
	e1->type = addtype(t_int, freetype(e1->type));
	setConst(e1->type);
	break;	
      default:
	error("Illegal constants");
    }
    return 1;
}


void * getaddress(token_t * e1)
{
    void *p = NULL;
    int t = gettype(e1->type);
    switch (t) {
    case t_builtin:
	if (e1->sym->val.vfunc != 0)
	    p = (void*)e1->sym->val.vfunc;
	else
	    error("No link to builtin function");
	e1->type = addtype(t_pointer, e1->type);
	break;
    case t_funcdec:
    case t_func:
	p = e1->sym;
	e1->type = addtype(t_pointer, e1->type);
	break;
    case t_union:
    case t_struct:
	    error("Illegal use of structure or uion: need an lvalue");
    case t_array:
	if (t == t_array)
	    exchtype(t_pointer, e1->type);
    case t_pointer:
	p = &ENV->AR[e1->sym->val.ival].v;
	break;
    default:
	error("Unknown type: need an lvalue");
    }
    return p;
}

void derefConst(token_t * e1)
{
    /* assumes incoming type is a ptr or array */
    
    e1->type = succType(e1->type);
    
    switch(gettype(e1->type)) {
      case t_char:  case t_uchar:
      case t_short: case t_ushort:
      case t_int:   case t_uint:
      case t_long:  case t_ulong:
      case t_float: case t_double:
      case t_pointer:
      case t_union:
      case t_struct:
	generate(&e1->code,pushptr,&e1->val, 0);
	e1->type = addtype(t_lval,e1->type);
	 break;
      case t_void:
	return;
      case t_array:
        exchtype(t_pointer,e1->type);
	setConst(e1->type);
        
        break;
      default:
	error("Unknown const in derefConst");
    }

}

void output(token_t * e1)
{
    static val_t v;
    if (e1->pflag == 0) {
	int t = gettype(e1->type);

	if(isconst(e1->type)) {
	    switch (t) {
	      CASE_INT:
	      CASE_UINT:generate(&e1->code, pushint, &e1->val, 0);
		break;
	      CASE_LONG:
	      CASE_ULONG:generate(&e1->code, pushlong, &e1->val, 0);
		break;
	      CASE_FLOAT:generate(&e1->code, pushdouble, &e1->val, 0);
		break;
	      case t_union:
	      case t_struct:
	      case t_array:
		if (t == t_array)
		    exchtype(t_pointer, e1->type);
	      case t_pointer:
		generate(&e1->code, pushptr, &e1->val, 0);
		break;
	      case t_void:
		return;
	      default:
		error("Unknown const type through output");
	    }
	    e1->pflag =1;
	    unsetConst(e1->type);
	    return;
	} 
	
	switch (t) {
	  case t_char:
	    generate(&e1->code, rvalchar, &e1->val, GI(e1));
	    exchtype(t_int,e1->type);
	    break;
	  case t_uchar:
	    generate(&e1->code, rvaluchar, &e1->val, GI(e1));
	    exchtype(t_int,e1->type);
	    break;
	  case t_short:
	    generate(&e1->code, rvalshort, &e1->val, GI(e1));
	    exchtype(t_int,e1->type);
	    break; 
	  case t_ushort:
	    generate(&e1->code, rvalushort, &e1->val, GI(e1));
	    exchtype(t_int,e1->type);
	    break;
	  case t_int:
	  case t_uint:
	    generate(&e1->code, rvalint, &e1->val, GI(e1));
	    break;
	  CASE_LONG:
	  CASE_ULONG:generate(&e1->code, rvallong, &e1->val, GI(e1));
	    break;
	  case t_float:
	    generate(&e1->code, rvalfloat, &e1->val,GI(e1));
	    exchtype(t_double,e1->type);
	    break;
	    
	  case t_double:
	    generate(&e1->code, rvaldouble, &e1->val,GI(e1));
	    break;

	  case t_builtin:
	    if (e1->sym->val.vfunc != 0) {
		v.vfunc = e1->sym->val.vfunc;
		generate(&e1->code, pushptr, &v, 0);
		e1->type = addtype(t_pointer, e1->type);
	    } else
		error("No link to builtin function");
	    break;
	  case t_funcdec:
	    exchtype(t_func, e1->type);
	  case t_func:
	    v.p.p = e1->sym;
	    generate(&e1->code, pushptr, &v, 0);
	    e1->type = addtype(t_pointer, e1->type);
	    break;
	  case t_union:
	  case t_struct:
	  case t_array:
	    if (t == t_array)
		exchtype(t_pointer, e1->type);
	  case t_pointer:
	    generate(&e1->code, rvalptr, &e1->val, GI(e1));
	    break;

	  case t_lval:
	    e1->type = succType(e1->type);
	    switch (gettype(e1->type)) {
	      case t_uchar:
		generate(&e1->code, drefuchar, &e1->val, 0);
		break;
	      case t_char:
		generate(&e1->code, drefchar, &e1->val, 0);
		break;
	      case t_short:
		generate(&e1->code, drefshort, &e1->val, 0);
		break;
	      case t_ushort:
		generate(&e1->code, drefushort, &e1->val, 0);
		break;
	      case t_int:
		generate(&e1->code, drefint, &e1->val, 0);
		break;
	      case t_uint:
		generate(&e1->code, drefuint, &e1->val, 0);
		break;
	      CASE_LONG:generate(&e1->code, dreflong, &e1->val, 0);
		break;
	      CASE_ULONG:generate(&e1->code, drefulong, &e1->val, 0);
		break;
	      case t_float:
		generate(&e1->code, dreffloat, &e1->val, 0);
		break;
	      case t_double:
		generate(&e1->code, drefdouble, &e1->val, 0);
		break;
	      case t_array:
	      case t_struct:
	      case t_union:
		break;
	      case t_pointer:
		generate(&e1->code, drefptr, &e1->val, GI(e1));
		break;
		
	      case t_funcdec:
	      case t_func:
		e1->type = succType(e1->type);
		generate(&e1->code, eiccall, &e1->val, 0);
		break;
	      case t_builtin:
		generate(&e1->code, call, &e1->val, 0);
		e1->type = succType(e1->type);
		break;
	      default:
		error("Illegal lvalue");
		break;
	    }
	    break;
	  case t_void:
	    return;
	  default:
	    error("Unknown type through output");
	    return;
	}
	e1->pflag = 1;
    }
}

void do_inc_dec(token_t * e1, int op)
{
    val_t v;
    int t;
    
    switch (gettype(e1->type)) {
      CASE_INT:
	t = gettype(e1->type);
	output(e1);
	v.ival = 1;
	if (op == INC)
	    op = incint;
	else
	    op = decint;
	exchtype(t,e1->type);
	break;
      CASE_UINT:
	t = gettype(e1->type);
	output(e1);
	v.ival = 1;
	if (op == INC)
	    op = incuint;
	else
	    op = decuint;
	exchtype(t,e1->type);
	break;
      CASE_LONG:
	output(e1);
	v.ival = 1;
	if (op == INC)
	    op = inclong;
	else
	    op = declong;
	break;
      CASE_ULONG:
	output(e1);
	v.ival = 1;
	if (op == INC)
	    op = inculong;
	else
	    op = deculong;
	break;
      case t_pointer:
	output(e1);
	v.ival = get_sizeof(nextType(e1->type));
	if (op == INC)
	    op = incptr;
	else
	    op = decptr;
	break;
      case t_lval:
	v.ival = 1;
	generate(&e1->code, dupval, &v, 0);
	output(e1);
	do_inc_dec(e1, op);
	v.ival = 1;
	/*generate(&e1->code, bump, &v, 0);*/
	e1->type = addtype(t_lval, e1->type);
	return;
      case t_float:
      case t_double:
	t = gettype(e1->type);
	output(e1);
	v.ival = 1;
	if (op == INC)
	    op = incdouble;
	else
	    op = decdouble;
	exchtype(t,e1->type);
	break;
	
      default:
	error("Cannot apply inc\\dec");
	return;
    }
    generate(&e1->code, op, &v, 0);
}

void do_stooutput(token_t * e1)
{
    
    if (e1->pflag == 0) {
	switch (gettype(e1->type)) {
	  case t_char:
	  case t_uchar:
	    generate(&e1->code, stochar, &e1->val, GI(e1));
	    break;
	  case t_short:
	  case t_ushort:
	    generate(&e1->code, stoshort, &e1->val, GI(e1));
	    break;
	  case t_int:
	  case t_uint:
	    generate(&e1->code, stoint, &e1->val, GI(e1));
	    break;
	  CASE_LONG:
	  CASE_ULONG:generate(&e1->code, stolong, &e1->val, GI(e1));
	    break;
	  case t_float:
	    generate(&e1->code, stofloat, &e1->val, GI(e1));
	    break;
	  case t_double:
	    generate(&e1->code, stodouble, &e1->val, GI(e1));
	    break;
	  case t_pointer:
	    generate(&e1->code, stoptr, &e1->val, GI(e1));
	    break;
	  case t_struct:
	  case t_union:{
	      val_t v;
	      v.ival = get_sizeof(e1->type);
	      output(e1);
	      generate(&e1->code, refmem, &v, 0);
	  }
	    break;
	  case t_lval:
	    e1->type = succType(e1->type);
	    if(isconst(e1->type) || isconstp(e1->type))
		error("Illegal assignment operation");
	    switch (gettype(e1->type)) {
	      case t_uchar:
	      case t_char:
		generate(&e1->code, refchar, &e1->val, 0);
		break;
	      case t_short:
	      case t_ushort:
		generate(&e1->code, refshort, &e1->val, 0);
		break;
	      case t_int:
	      case t_uint:
		generate(&e1->code, refint, &e1->val, 0);
		break;
	      CASE_LONG:generate(&e1->code, reflong, &e1->val, 0);
		break;
	      CASE_ULONG:generate(&e1->code, refulong, &e1->val, 0);
		break;
	      case t_float:
		generate(&e1->code, reffloat, &e1->val, 0);
		break;
	      case t_double:
		generate(&e1->code, refdouble, &e1->val, 0);
		break;
	      case t_pointer:
		generate(&e1->code, refptr, &e1->val, GI(e1));
		break;
	      case t_struct:
	      case t_union:{
		  val_t v;
		  v.ival = get_sizeof(e1->type);
		  generate(&e1->code, refmem, &v, 0);
	      }
		break;
	      default:
		error("Illegal lvalue");
		break;
	    }
	    break;
	  default:
	    error(" Invalid assignment");
	    break;
	}
	e1->pflag = 1;
    }
}

void binhconst(unsigned opr, token_t * e1, token_t * e2)
{
    void (*f) (unsigned, token_t *, token_t *);
    int t;

    if(isconst(e1->type) && isconst(e2->type)) {
	do_binaryop(e1, e2, opr);
	return;
    
    }
    /*
     * Now perform usual arithmetic conversion
     * i.e promote to larger type.
     */

    if(gettype(e1->type) < gettype(e2->type))
	t = gettype(e2->type);
    else
	t = gettype(e1->type);
    
    switch (t) {
      CASE_INT:f = binhint;
	break;
      CASE_UINT:f = binhuint;
	break;
      CASE_LONG:f = binhlong;
	break;
      CASE_ULONG:f = binhulong;
	break;
      CASE_FLOAT:f = binhdouble;
	break;
      case t_array:
      case t_pointer:
	f = binhlval;
	break;
      default:
	error("Unknown type in binhconst");
	return;
    }
    (*f) (opr, e1, e2);
}
